// SPDX-FileCopyrightText: Nheko Contributors
//
// SPDX-License-Identifier: GPL-3.0-or-later

// DO NOT EDIT PROVIDER.CPP DIRECTLY! EDIT resources/provider-head.txt AND RUN scripts/codegen.sh!

#include "emoji/Provider.h"

using namespace emoji;

constexpr

  std::array<Emoji, 3799>
    emoji::Provider::emoji = {
      // People
      Emoji{std::u16string_view(u"\U0001F600"),
            std::u16string_view(u"grinning"),
            std::u16string_view(u"grinning face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F603"),
            std::u16string_view(u"grinning_face_with_big_eyes"),
            std::u16string_view(u"grinning face with big eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F604"),
            std::u16string_view(u"smile"),
            std::u16string_view(u"grinning face with smiling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F601"),
            std::u16string_view(u"beaming_face_with_smiling_eyes"),
            std::u16string_view(u"beaming face with smiling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F606"),
            std::u16string_view(u"grinning_squinting"),
            std::u16string_view(u"grinning squinting face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F605"),
            std::u16string_view(u"sweat_smile"),
            std::u16string_view(u"grinning face with sweat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F923"),
            std::u16string_view(u"rofl"),
            std::u16string_view(u"rolling on the floor laughing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F602"),
            std::u16string_view(u"face_with_tears_of_joy"),
            std::u16string_view(u"face with tears of joy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F642"),
            std::u16string_view(u"slightly_smiling"),
            std::u16string_view(u"slightly smiling face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F643"),
            std::u16string_view(u"upside_down"),
            std::u16string_view(u"upside-down face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE0"),
            std::u16string_view(u"melting"),
            std::u16string_view(u"melting face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F609"),
            std::u16string_view(u"winking"),
            std::u16string_view(u"winking face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60A"),
            std::u16string_view(u"smiling_face_with_smiling_eyes"),
            std::u16string_view(u"smiling face with smiling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F607"),
            std::u16string_view(u"innocent"),
            std::u16string_view(u"smiling face with halo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F970"),
            std::u16string_view(u"smiling_face_with_hearts"),
            std::u16string_view(u"smiling face with hearts"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60D"),
            std::u16string_view(u"heart_eyes"),
            std::u16string_view(u"smiling face with heart-eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F929"),
            std::u16string_view(u"star_struck"),
            std::u16string_view(u"star-struck"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F618"),
            std::u16string_view(u"face_blowing_a_kiss"),
            std::u16string_view(u"face blowing a kiss"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F617"),
            std::u16string_view(u"kissing"),
            std::u16string_view(u"kissing face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000263A\U0000FE0F"),
            std::u16string_view(u"smiling"),
            std::u16string_view(u"smiling face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61A"),
            std::u16string_view(u"kissing_face_with_closed_eyes"),
            std::u16string_view(u"kissing face with closed eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F619"),
            std::u16string_view(u"kissing_face_with_smiling_eyes"),
            std::u16string_view(u"kissing face with smiling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F972"),
            std::u16string_view(u"cry_smile"),
            std::u16string_view(u"smiling face with tear"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60B"),
            std::u16string_view(u"yum"),
            std::u16string_view(u"face savoring food"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61B"),
            std::u16string_view(u"face_with_tongue"),
            std::u16string_view(u"face with tongue"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61C"),
            std::u16string_view(u"winking_face_with_tongue"),
            std::u16string_view(u"winking face with tongue"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92A"),
            std::u16string_view(u"zany"),
            std::u16string_view(u"zany face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61D"),
            std::u16string_view(u"squinting_face_with_tongue"),
            std::u16string_view(u"squinting face with tongue"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F911"),
            std::u16string_view(u"money_mouth"),
            std::u16string_view(u"money-mouth face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F917"),
            std::u16string_view(u"hugging_face"),
            std::u16string_view(u"smiling face with open hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92D"),
            std::u16string_view(u"face_with_hand_over_mouth"),
            std::u16string_view(u"face with hand over mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE2"),
            std::u16string_view(u"face_with_open_eyes_and_hand_over_mouth"),
            std::u16string_view(u"face with open eyes and hand over mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE3"),
            std::u16string_view(u"face_with_peeking_eye"),
            std::u16string_view(u"face with peeking eye"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92B"),
            std::u16string_view(u"shushing"),
            std::u16string_view(u"shushing face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F914"),
            std::u16string_view(u"thinking"),
            std::u16string_view(u"thinking face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE1"),
            std::u16string_view(u"saluting"),
            std::u16string_view(u"saluting face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F910"),
            std::u16string_view(u"zipper_mouth"),
            std::u16string_view(u"zipper-mouth face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F928"),
            std::u16string_view(u"face_with_raised_eyebrow"),
            std::u16string_view(u"face with raised eyebrow"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F610"),
            std::u16string_view(u"neutral"),
            std::u16string_view(u"neutral face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F611"),
            std::u16string_view(u"expressionless"),
            std::u16string_view(u"expressionless face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F636"),
            std::u16string_view(u"face_without_mouth"),
            std::u16string_view(u"face without mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE5"),
            std::u16string_view(u"dotted_line"),
            std::u16string_view(u"dotted line face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F636\U0000200D\U0001F32B\U0000FE0F"),
            std::u16string_view(u"face_in_clouds"),
            std::u16string_view(u"face in clouds"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60F"),
            std::u16string_view(u"smirking"),
            std::u16string_view(u"smirking face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F612"),
            std::u16string_view(u"unamused"),
            std::u16string_view(u"unamused face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F644"),
            std::u16string_view(u"face_with_rolling_eyes"),
            std::u16string_view(u"face with rolling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62C"),
            std::u16string_view(u"grimacing"),
            std::u16string_view(u"grimacing face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62E\U0000200D\U0001F4A8"),
            std::u16string_view(u"face_exhaling"),
            std::u16string_view(u"face exhaling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F925"),
            std::u16string_view(u"lying"),
            std::u16string_view(u"lying face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE8"),
            std::u16string_view(u"shaking"),
            std::u16string_view(u"shaking face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F642\U0000200D\U00002194\U0000FE0F"),
            std::u16string_view(u"head_shaking_horizontally"),
            std::u16string_view(u"head shaking horizontally"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F642\U0000200D\U00002195\U0000FE0F"),
            std::u16string_view(u"head_shaking_vertically"),
            std::u16string_view(u"head shaking vertically"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60C"),
            std::u16string_view(u"relieved"),
            std::u16string_view(u"relieved face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F614"),
            std::u16string_view(u"pensive"),
            std::u16string_view(u"pensive face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62A"),
            std::u16string_view(u"sleepy"),
            std::u16string_view(u"sleepy face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F924"),
            std::u16string_view(u"drooling"),
            std::u16string_view(u"drooling face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F634"),
            std::u16string_view(u"sleeping"),
            std::u16string_view(u"sleeping face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F637"),
            std::u16string_view(u"face_with_medical_mask"),
            std::u16string_view(u"face with medical mask"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F912"),
            std::u16string_view(u"face_with_thermometer"),
            std::u16string_view(u"face with thermometer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F915"),
            std::u16string_view(u"face_with_head_bandage"),
            std::u16string_view(u"face with head-bandage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F922"),
            std::u16string_view(u"nauseated"),
            std::u16string_view(u"nauseated face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92E"),
            std::u16string_view(u"face_vomiting"),
            std::u16string_view(u"face vomiting"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F927"),
            std::u16string_view(u"sneezing"),
            std::u16string_view(u"sneezing face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F975"),
            std::u16string_view(u"hot"),
            std::u16string_view(u"hot face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F976"),
            std::u16string_view(u"cold"),
            std::u16string_view(u"cold face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F974"),
            std::u16string_view(u"woozy"),
            std::u16string_view(u"woozy face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F635"),
            std::u16string_view(u"face_with_crossed_out_eyes"),
            std::u16string_view(u"face with crossed-out eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F635\U0000200D\U0001F4AB"),
            std::u16string_view(u"face_with_spiral_eyes"),
            std::u16string_view(u"face with spiral eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92F"),
            std::u16string_view(u"exploding_head"),
            std::u16string_view(u"exploding head"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F920"),
            std::u16string_view(u"cowboy_hat"),
            std::u16string_view(u"cowboy hat face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F973"),
            std::u16string_view(u"partying"),
            std::u16string_view(u"partying face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F978"),
            std::u16string_view(u"disguised"),
            std::u16string_view(u"disguised face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F60E"),
            std::u16string_view(u"smiling_face_with_sunglasses"),
            std::u16string_view(u"smiling face with sunglasses"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F913"),
            std::u16string_view(u"nerd"),
            std::u16string_view(u"nerd face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D0"),
            std::u16string_view(u"face_with_monocle"),
            std::u16string_view(u"face with monocle"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F615"),
            std::u16string_view(u"confused"),
            std::u16string_view(u"confused face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE4"),
            std::u16string_view(u"face_with_diagonal_mouth"),
            std::u16string_view(u"face with diagonal mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61F"),
            std::u16string_view(u"worried"),
            std::u16string_view(u"worried face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F641"),
            std::u16string_view(u"slightly_frowning"),
            std::u16string_view(u"slightly frowning face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002639\U0000FE0F"),
            std::u16string_view(u"frowning"),
            std::u16string_view(u"frowning face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62E"),
            std::u16string_view(u"face_with_open_mouth"),
            std::u16string_view(u"face with open mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62F"),
            std::u16string_view(u"hushed"),
            std::u16string_view(u"hushed face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F632"),
            std::u16string_view(u"astonished"),
            std::u16string_view(u"astonished face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F633"),
            std::u16string_view(u"flushed"),
            std::u16string_view(u"flushed face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F97A"),
            std::u16string_view(u"pleading"),
            std::u16string_view(u"pleading face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F979"),
            std::u16string_view(u"face_holding_back_tears"),
            std::u16string_view(u"face holding back tears"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F626"),
            std::u16string_view(u"frowning_face_with_open_mouth"),
            std::u16string_view(u"frowning face with open mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F627"),
            std::u16string_view(u"anguished"),
            std::u16string_view(u"anguished face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F628"),
            std::u16string_view(u"fearful"),
            std::u16string_view(u"fearful face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F630"),
            std::u16string_view(u"anxious_face_with_sweat"),
            std::u16string_view(u"anxious face with sweat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F625"),
            std::u16string_view(u"sad_but_relieved"),
            std::u16string_view(u"sad but relieved face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F622"),
            std::u16string_view(u"crying"),
            std::u16string_view(u"crying face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62D"),
            std::u16string_view(u"sob"),
            std::u16string_view(u"loudly crying face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F631"),
            std::u16string_view(u"scream"),
            std::u16string_view(u"face screaming in fear"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F616"),
            std::u16string_view(u"confounded"),
            std::u16string_view(u"confounded face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F623"),
            std::u16string_view(u"persevering"),
            std::u16string_view(u"persevering face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F61E"),
            std::u16string_view(u"disappointed"),
            std::u16string_view(u"disappointed face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F613"),
            std::u16string_view(u"sweat"),
            std::u16string_view(u"downcast face with sweat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F629"),
            std::u16string_view(u"weary"),
            std::u16string_view(u"weary face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F62B"),
            std::u16string_view(u"tired"),
            std::u16string_view(u"tired face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F971"),
            std::u16string_view(u"yawning"),
            std::u16string_view(u"yawning face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F624"),
            std::u16string_view(u"triumph"),
            std::u16string_view(u"face with steam from nose"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F621"),
            std::u16string_view(u"enraged"),
            std::u16string_view(u"enraged face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F620"),
            std::u16string_view(u"angry"),
            std::u16string_view(u"angry face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F92C"),
            std::u16string_view(u"face_with_symbols_on_mouth"),
            std::u16string_view(u"face with symbols on mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F608"),
            std::u16string_view(u"smiling_imp"),
            std::u16string_view(u"smiling face with horns"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47F"),
            std::u16string_view(u"imp"),
            std::u16string_view(u"angry face with horns"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F480"),
            std::u16string_view(u"skull"),
            std::u16string_view(u"skull"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002620\U0000FE0F"),
            std::u16string_view(u"skull_and_crossbones"),
            std::u16string_view(u"skull and crossbones"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A9"),
            std::u16string_view(u"pile_of_poo"),
            std::u16string_view(u"pile of poo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F921"),
            std::u16string_view(u"clown"),
            std::u16string_view(u"clown face"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F479"),
            std::u16string_view(u"ogre"),
            std::u16string_view(u"ogre"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47A"),
            std::u16string_view(u"goblin"),
            std::u16string_view(u"goblin"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47B"),
            std::u16string_view(u"ghost"),
            std::u16string_view(u"ghost"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47D"),
            std::u16string_view(u"alien"),
            std::u16string_view(u"alien"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47E"),
            std::u16string_view(u"alien_monster"),
            std::u16string_view(u"alien monster"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F916"),
            std::u16string_view(u"robot"),
            std::u16string_view(u"robot"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63A"),
            std::u16string_view(u"smiley_cat"),
            std::u16string_view(u"grinning cat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F638"),
            std::u16string_view(u"smile_cat"),
            std::u16string_view(u"grinning cat with smiling eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F639"),
            std::u16string_view(u"joy_cat"),
            std::u16string_view(u"cat with tears of joy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63B"),
            std::u16string_view(u"heart_eyes_cat"),
            std::u16string_view(u"smiling cat with heart-eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63C"),
            std::u16string_view(u"smirk_cat"),
            std::u16string_view(u"cat with wry smile"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63D"),
            std::u16string_view(u"kissing_cat"),
            std::u16string_view(u"kissing cat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F640"),
            std::u16string_view(u"scream_cat"),
            std::u16string_view(u"weary cat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63F"),
            std::u16string_view(u"crying_cat"),
            std::u16string_view(u"crying cat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F63E"),
            std::u16string_view(u"pouting_cat"),
            std::u16string_view(u"pouting cat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F648"),
            std::u16string_view(u"see_no_evil_monkey"),
            std::u16string_view(u"see-no-evil monkey"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F649"),
            std::u16string_view(u"hear_no_evil_monkey"),
            std::u16string_view(u"hear-no-evil monkey"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64A"),
            std::u16string_view(u"speak_no_evil_monkey"),
            std::u16string_view(u"speak-no-evil monkey"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48C"),
            std::u16string_view(u"love_letter"),
            std::u16string_view(u"love letter"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F498"),
            std::u16string_view(u"heart_with_arrow"),
            std::u16string_view(u"heart with arrow"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49D"),
            std::u16string_view(u"heart_with_ribbon"),
            std::u16string_view(u"heart with ribbon"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F496"),
            std::u16string_view(u"sparkling_heart"),
            std::u16string_view(u"sparkling heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F497"),
            std::u16string_view(u"growing_heart"),
            std::u16string_view(u"growing heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F493"),
            std::u16string_view(u"beating_heart"),
            std::u16string_view(u"beating heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49E"),
            std::u16string_view(u"revolving_hearts"),
            std::u16string_view(u"revolving hearts"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F495"),
            std::u16string_view(u"two_hearts"),
            std::u16string_view(u"two hearts"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49F"),
            std::u16string_view(u"heart_decoration"),
            std::u16string_view(u"heart decoration"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002763\U0000FE0F"),
            std::u16string_view(u"heart_exclamation"),
            std::u16string_view(u"heart exclamation"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F494"),
            std::u16string_view(u"broken_heart"),
            std::u16string_view(u"broken heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002764\U0000FE0F\U0000200D\U0001F525"),
            std::u16string_view(u"heart_on_fire"),
            std::u16string_view(u"heart on fire"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002764\U0000FE0F\U0000200D\U0001FA79"),
            std::u16string_view(u"mending_heart"),
            std::u16string_view(u"mending heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U00002764\U0000FE0F"),
            std::u16string_view(u"red_heart"),
            std::u16string_view(u"red heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FA77"),
            std::u16string_view(u"pink_heart"),
            std::u16string_view(u"pink heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9E1"),
            std::u16string_view(u"orange_heart"),
            std::u16string_view(u"orange heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49B"),
            std::u16string_view(u"yellow_heart"),
            std::u16string_view(u"yellow heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49A"),
            std::u16string_view(u"green_heart"),
            std::u16string_view(u"green heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F499"),
            std::u16string_view(u"blue_heart"),
            std::u16string_view(u"blue heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FA75"),
            std::u16string_view(u"light_blue_heart"),
            std::u16string_view(u"light blue heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F49C"),
            std::u16string_view(u"purple_heart"),
            std::u16string_view(u"purple heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90E"),
            std::u16string_view(u"brown_heart"),
            std::u16string_view(u"brown heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F5A4"),
            std::u16string_view(u"black_heart"),
            std::u16string_view(u"black heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FA76"),
            std::u16string_view(u"grey_heart"),
            std::u16string_view(u"grey heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90D"),
            std::u16string_view(u"white_heart"),
            std::u16string_view(u"white heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48B"),
            std::u16string_view(u"kiss_mark"),
            std::u16string_view(u"kiss mark"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AF"),
            std::u16string_view(u"100"),
            std::u16string_view(u"hundred points"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A2"),
            std::u16string_view(u"anger_symbol"),
            std::u16string_view(u"anger symbol"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A5"),
            std::u16string_view(u"boom"),
            std::u16string_view(u"collision"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AB"),
            std::u16string_view(u"dizzy"),
            std::u16string_view(u"dizzy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A6"),
            std::u16string_view(u"sweat_droplets"),
            std::u16string_view(u"sweat droplets"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A8"),
            std::u16string_view(u"dashing_away"),
            std::u16string_view(u"dashing away"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F573\U0000FE0F"),
            std::u16string_view(u"hole"),
            std::u16string_view(u"hole"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AC"),
            std::u16string_view(u"speech_balloon"),
            std::u16string_view(u"speech balloon"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F441\U0000FE0F\U0000200D\U0001F5E8\U0000FE0F"),
            std::u16string_view(u"eye_in_speech_bubble"),
            std::u16string_view(u"eye in speech bubble"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F5E8\U0000FE0F"),
            std::u16string_view(u"left_speech_bubble"),
            std::u16string_view(u"left speech bubble"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F5EF\U0000FE0F"),
            std::u16string_view(u"right_anger_bubble"),
            std::u16string_view(u"right anger bubble"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AD"),
            std::u16string_view(u"thought_balloon"),
            std::u16string_view(u"thought balloon"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4A4"),
            std::u16string_view(u"zzz"),
            std::u16string_view(u"ZZZ"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B"),
            std::u16string_view(u"wave"),
            std::u16string_view(u"waving hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B\U0001F3FB"),
            std::u16string_view(u"waving_hand_light_skin_tone"),
            std::u16string_view(u"waving hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B\U0001F3FC"),
            std::u16string_view(u"waving_hand_medium_light_skin_tone"),
            std::u16string_view(u"waving hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B\U0001F3FD"),
            std::u16string_view(u"waving_hand_medium_skin_tone"),
            std::u16string_view(u"waving hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B\U0001F3FE"),
            std::u16string_view(u"waving_hand_medium_dark_skin_tone"),
            std::u16string_view(u"waving hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44B\U0001F3FF"),
            std::u16string_view(u"waving_hand_dark_skin_tone"),
            std::u16string_view(u"waving hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A"),
            std::u16string_view(u"raised_back_of_hand"),
            std::u16string_view(u"raised back of hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A\U0001F3FB"),
            std::u16string_view(u"raised_back_of_hand_light_skin_tone"),
            std::u16string_view(u"raised back of hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A\U0001F3FC"),
            std::u16string_view(u"raised_back_of_hand_medium_light_skin_tone"),
            std::u16string_view(u"raised back of hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A\U0001F3FD"),
            std::u16string_view(u"raised_back_of_hand_medium_skin_tone"),
            std::u16string_view(u"raised back of hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A\U0001F3FE"),
            std::u16string_view(u"raised_back_of_hand_medium_dark_skin_tone"),
            std::u16string_view(u"raised back of hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91A\U0001F3FF"),
            std::u16string_view(u"raised_back_of_hand_dark_skin_tone"),
            std::u16string_view(u"raised back of hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0000FE0F"),
            std::u16string_view(u"hand_with_fingers_splayed"),
            std::u16string_view(u"hand with fingers splayed"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0001F3FB"),
            std::u16string_view(u"hand_with_fingers_splayed_light_skin_tone"),
            std::u16string_view(u"hand with fingers splayed: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0001F3FC"),
            std::u16string_view(u"hand_with_fingers_splayed_medium_light_skin_tone"),
            std::u16string_view(u"hand with fingers splayed: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0001F3FD"),
            std::u16string_view(u"hand_with_fingers_splayed_medium_skin_tone"),
            std::u16string_view(u"hand with fingers splayed: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0001F3FE"),
            std::u16string_view(u"hand_with_fingers_splayed_medium_dark_skin_tone"),
            std::u16string_view(u"hand with fingers splayed: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F590\U0001F3FF"),
            std::u16string_view(u"hand_with_fingers_splayed_dark_skin_tone"),
            std::u16string_view(u"hand with fingers splayed: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B"),
            std::u16string_view(u"raised_hand"),
            std::u16string_view(u"raised hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B\U0001F3FB"),
            std::u16string_view(u"raised_hand_light_skin_tone"),
            std::u16string_view(u"raised hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B\U0001F3FC"),
            std::u16string_view(u"raised_hand_medium_light_skin_tone"),
            std::u16string_view(u"raised hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B\U0001F3FD"),
            std::u16string_view(u"raised_hand_medium_skin_tone"),
            std::u16string_view(u"raised hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B\U0001F3FE"),
            std::u16string_view(u"raised_hand_medium_dark_skin_tone"),
            std::u16string_view(u"raised hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270B\U0001F3FF"),
            std::u16string_view(u"raised_hand_dark_skin_tone"),
            std::u16string_view(u"raised hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596"),
            std::u16string_view(u"vulcan_salute"),
            std::u16string_view(u"vulcan salute"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596\U0001F3FB"),
            std::u16string_view(u"vulcan_salute_light_skin_tone"),
            std::u16string_view(u"vulcan salute: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596\U0001F3FC"),
            std::u16string_view(u"vulcan_salute_medium_light_skin_tone"),
            std::u16string_view(u"vulcan salute: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596\U0001F3FD"),
            std::u16string_view(u"vulcan_salute_medium_skin_tone"),
            std::u16string_view(u"vulcan salute: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596\U0001F3FE"),
            std::u16string_view(u"vulcan_salute_medium_dark_skin_tone"),
            std::u16string_view(u"vulcan salute: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F596\U0001F3FF"),
            std::u16string_view(u"vulcan_salute_dark_skin_tone"),
            std::u16string_view(u"vulcan salute: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1"),
            std::u16string_view(u"rightwards_hand"),
            std::u16string_view(u"rightwards hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FB"),
            std::u16string_view(u"rightwards_hand_light_skin_tone"),
            std::u16string_view(u"rightwards hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FC"),
            std::u16string_view(u"rightwards_hand_medium_light_skin_tone"),
            std::u16string_view(u"rightwards hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FD"),
            std::u16string_view(u"rightwards_hand_medium_skin_tone"),
            std::u16string_view(u"rightwards hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FE"),
            std::u16string_view(u"rightwards_hand_medium_dark_skin_tone"),
            std::u16string_view(u"rightwards hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FF"),
            std::u16string_view(u"rightwards_hand_dark_skin_tone"),
            std::u16string_view(u"rightwards hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2"),
            std::u16string_view(u"leftwards_hand"),
            std::u16string_view(u"leftwards hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2\U0001F3FB"),
            std::u16string_view(u"leftwards_hand_light_skin_tone"),
            std::u16string_view(u"leftwards hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2\U0001F3FC"),
            std::u16string_view(u"leftwards_hand_medium_light_skin_tone"),
            std::u16string_view(u"leftwards hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2\U0001F3FD"),
            std::u16string_view(u"leftwards_hand_medium_skin_tone"),
            std::u16string_view(u"leftwards hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2\U0001F3FE"),
            std::u16string_view(u"leftwards_hand_medium_dark_skin_tone"),
            std::u16string_view(u"leftwards hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF2\U0001F3FF"),
            std::u16string_view(u"leftwards_hand_dark_skin_tone"),
            std::u16string_view(u"leftwards hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3"),
            std::u16string_view(u"palm_down_hand"),
            std::u16string_view(u"palm down hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3\U0001F3FB"),
            std::u16string_view(u"palm_down_hand_light_skin_tone"),
            std::u16string_view(u"palm down hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3\U0001F3FC"),
            std::u16string_view(u"palm_down_hand_medium_light_skin_tone"),
            std::u16string_view(u"palm down hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3\U0001F3FD"),
            std::u16string_view(u"palm_down_hand_medium_skin_tone"),
            std::u16string_view(u"palm down hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3\U0001F3FE"),
            std::u16string_view(u"palm_down_hand_medium_dark_skin_tone"),
            std::u16string_view(u"palm down hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF3\U0001F3FF"),
            std::u16string_view(u"palm_down_hand_dark_skin_tone"),
            std::u16string_view(u"palm down hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4"),
            std::u16string_view(u"palm_up_hand"),
            std::u16string_view(u"palm up hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4\U0001F3FB"),
            std::u16string_view(u"palm_up_hand_light_skin_tone"),
            std::u16string_view(u"palm up hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4\U0001F3FC"),
            std::u16string_view(u"palm_up_hand_medium_light_skin_tone"),
            std::u16string_view(u"palm up hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4\U0001F3FD"),
            std::u16string_view(u"palm_up_hand_medium_skin_tone"),
            std::u16string_view(u"palm up hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4\U0001F3FE"),
            std::u16string_view(u"palm_up_hand_medium_dark_skin_tone"),
            std::u16string_view(u"palm up hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF4\U0001F3FF"),
            std::u16string_view(u"palm_up_hand_dark_skin_tone"),
            std::u16string_view(u"palm up hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7"),
            std::u16string_view(u"leftwards_pushing_hand"),
            std::u16string_view(u"leftwards pushing hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7\U0001F3FB"),
            std::u16string_view(u"leftwards_pushing_hand_light_skin_tone"),
            std::u16string_view(u"leftwards pushing hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7\U0001F3FC"),
            std::u16string_view(u"leftwards_pushing_hand_medium_light_skin_tone"),
            std::u16string_view(u"leftwards pushing hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7\U0001F3FD"),
            std::u16string_view(u"leftwards_pushing_hand_medium_skin_tone"),
            std::u16string_view(u"leftwards pushing hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7\U0001F3FE"),
            std::u16string_view(u"leftwards_pushing_hand_medium_dark_skin_tone"),
            std::u16string_view(u"leftwards pushing hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF7\U0001F3FF"),
            std::u16string_view(u"leftwards_pushing_hand_dark_skin_tone"),
            std::u16string_view(u"leftwards pushing hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8"),
            std::u16string_view(u"rightwards_pushing_hand"),
            std::u16string_view(u"rightwards pushing hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8\U0001F3FB"),
            std::u16string_view(u"rightwards_pushing_hand_light_skin_tone"),
            std::u16string_view(u"rightwards pushing hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8\U0001F3FC"),
            std::u16string_view(u"rightwards_pushing_hand_medium_light_skin_tone"),
            std::u16string_view(u"rightwards pushing hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8\U0001F3FD"),
            std::u16string_view(u"rightwards_pushing_hand_medium_skin_tone"),
            std::u16string_view(u"rightwards pushing hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8\U0001F3FE"),
            std::u16string_view(u"rightwards_pushing_hand_medium_dark_skin_tone"),
            std::u16string_view(u"rightwards pushing hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF8\U0001F3FF"),
            std::u16string_view(u"rightwards_pushing_hand_dark_skin_tone"),
            std::u16string_view(u"rightwards pushing hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C"),
            std::u16string_view(u"ok_hand"),
            std::u16string_view(u"OK hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C\U0001F3FB"),
            std::u16string_view(u"ok_hand_light_skin_tone"),
            std::u16string_view(u"OK hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C\U0001F3FC"),
            std::u16string_view(u"ok_hand_medium_light_skin_tone"),
            std::u16string_view(u"OK hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C\U0001F3FD"),
            std::u16string_view(u"ok_hand_medium_skin_tone"),
            std::u16string_view(u"OK hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C\U0001F3FE"),
            std::u16string_view(u"ok_hand_medium_dark_skin_tone"),
            std::u16string_view(u"OK hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44C\U0001F3FF"),
            std::u16string_view(u"ok_hand_dark_skin_tone"),
            std::u16string_view(u"OK hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C"),
            std::u16string_view(u"pinched_fingers"),
            std::u16string_view(u"pinched fingers"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C\U0001F3FB"),
            std::u16string_view(u"pinched_fingers_light_skin_tone"),
            std::u16string_view(u"pinched fingers: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C\U0001F3FC"),
            std::u16string_view(u"pinched_fingers_medium_light_skin_tone"),
            std::u16string_view(u"pinched fingers: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C\U0001F3FD"),
            std::u16string_view(u"pinched_fingers_medium_skin_tone"),
            std::u16string_view(u"pinched fingers: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C\U0001F3FE"),
            std::u16string_view(u"pinched_fingers_medium_dark_skin_tone"),
            std::u16string_view(u"pinched fingers: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90C\U0001F3FF"),
            std::u16string_view(u"pinched_fingers_dark_skin_tone"),
            std::u16string_view(u"pinched fingers: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F"),
            std::u16string_view(u"pinching_hand"),
            std::u16string_view(u"pinching hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F\U0001F3FB"),
            std::u16string_view(u"pinching_hand_light_skin_tone"),
            std::u16string_view(u"pinching hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F\U0001F3FC"),
            std::u16string_view(u"pinching_hand_medium_light_skin_tone"),
            std::u16string_view(u"pinching hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F\U0001F3FD"),
            std::u16string_view(u"pinching_hand_medium_skin_tone"),
            std::u16string_view(u"pinching hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F\U0001F3FE"),
            std::u16string_view(u"pinching_hand_medium_dark_skin_tone"),
            std::u16string_view(u"pinching hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F90F\U0001F3FF"),
            std::u16string_view(u"pinching_hand_dark_skin_tone"),
            std::u16string_view(u"pinching hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0000FE0F"),
            std::u16string_view(u"victory_hand"),
            std::u16string_view(u"victory hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0001F3FB"),
            std::u16string_view(u"victory_hand_light_skin_tone"),
            std::u16string_view(u"victory hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0001F3FC"),
            std::u16string_view(u"victory_hand_medium_light_skin_tone"),
            std::u16string_view(u"victory hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0001F3FD"),
            std::u16string_view(u"victory_hand_medium_skin_tone"),
            std::u16string_view(u"victory hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0001F3FE"),
            std::u16string_view(u"victory_hand_medium_dark_skin_tone"),
            std::u16string_view(u"victory hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270C\U0001F3FF"),
            std::u16string_view(u"victory_hand_dark_skin_tone"),
            std::u16string_view(u"victory hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E"),
            std::u16string_view(u"crossed_fingers"),
            std::u16string_view(u"crossed fingers"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E\U0001F3FB"),
            std::u16string_view(u"crossed_fingers_light_skin_tone"),
            std::u16string_view(u"crossed fingers: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E\U0001F3FC"),
            std::u16string_view(u"crossed_fingers_medium_light_skin_tone"),
            std::u16string_view(u"crossed fingers: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E\U0001F3FD"),
            std::u16string_view(u"crossed_fingers_medium_skin_tone"),
            std::u16string_view(u"crossed fingers: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E\U0001F3FE"),
            std::u16string_view(u"crossed_fingers_medium_dark_skin_tone"),
            std::u16string_view(u"crossed fingers: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91E\U0001F3FF"),
            std::u16string_view(u"crossed_fingers_dark_skin_tone"),
            std::u16string_view(u"crossed fingers: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF0"),
            std::u16string_view(u"hand_with_index_finger_and_thumb_crossed"),
            std::u16string_view(u"hand with index finger and thumb crossed"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF0\U0001F3FB"),
            std::u16string_view(u"hand_with_index_finger_and_thumb_crossed_light_skin_tone"),
            std::u16string_view(u"hand with index finger and thumb crossed: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(u"\U0001FAF0\U0001F3FC"),
        std::u16string_view(u"hand_with_index_finger_and_thumb_crossed_medium_light_skin_tone"),
        std::u16string_view(u"hand with index finger and thumb crossed: medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF0\U0001F3FD"),
            std::u16string_view(u"hand_with_index_finger_and_thumb_crossed_medium_skin_tone"),
            std::u16string_view(u"hand with index finger and thumb crossed: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF0\U0001F3FE"),
            std::u16string_view(u"hand_with_index_finger_and_thumb_crossed_medium_dark_skin_tone"),
            std::u16string_view(u"hand with index finger and thumb crossed: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF0\U0001F3FF"),
            std::u16string_view(u"hand_with_index_finger_and_thumb_crossed_dark_skin_tone"),
            std::u16string_view(u"hand with index finger and thumb crossed: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F"),
            std::u16string_view(u"love_you_gesture"),
            std::u16string_view(u"love-you gesture"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F\U0001F3FB"),
            std::u16string_view(u"love_you_gesture_light_skin_tone"),
            std::u16string_view(u"love-you gesture: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F\U0001F3FC"),
            std::u16string_view(u"love_you_gesture_medium_light_skin_tone"),
            std::u16string_view(u"love-you gesture: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F\U0001F3FD"),
            std::u16string_view(u"love_you_gesture_medium_skin_tone"),
            std::u16string_view(u"love-you gesture: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F\U0001F3FE"),
            std::u16string_view(u"love_you_gesture_medium_dark_skin_tone"),
            std::u16string_view(u"love-you gesture: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91F\U0001F3FF"),
            std::u16string_view(u"love_you_gesture_dark_skin_tone"),
            std::u16string_view(u"love-you gesture: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918"),
            std::u16string_view(u"sign_of_the_horns"),
            std::u16string_view(u"sign of the horns"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918\U0001F3FB"),
            std::u16string_view(u"sign_of_the_horns_light_skin_tone"),
            std::u16string_view(u"sign of the horns: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918\U0001F3FC"),
            std::u16string_view(u"sign_of_the_horns_medium_light_skin_tone"),
            std::u16string_view(u"sign of the horns: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918\U0001F3FD"),
            std::u16string_view(u"sign_of_the_horns_medium_skin_tone"),
            std::u16string_view(u"sign of the horns: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918\U0001F3FE"),
            std::u16string_view(u"sign_of_the_horns_medium_dark_skin_tone"),
            std::u16string_view(u"sign of the horns: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F918\U0001F3FF"),
            std::u16string_view(u"sign_of_the_horns_dark_skin_tone"),
            std::u16string_view(u"sign of the horns: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919"),
            std::u16string_view(u"call_me_hand"),
            std::u16string_view(u"call me hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919\U0001F3FB"),
            std::u16string_view(u"call_me_hand_light_skin_tone"),
            std::u16string_view(u"call me hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919\U0001F3FC"),
            std::u16string_view(u"call_me_hand_medium_light_skin_tone"),
            std::u16string_view(u"call me hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919\U0001F3FD"),
            std::u16string_view(u"call_me_hand_medium_skin_tone"),
            std::u16string_view(u"call me hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919\U0001F3FE"),
            std::u16string_view(u"call_me_hand_medium_dark_skin_tone"),
            std::u16string_view(u"call me hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F919\U0001F3FF"),
            std::u16string_view(u"call_me_hand_dark_skin_tone"),
            std::u16string_view(u"call me hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448"),
            std::u16string_view(u"point_left"),
            std::u16string_view(u"backhand index pointing left"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448\U0001F3FB"),
            std::u16string_view(u"backhand_index_pointing_left_light_skin_tone"),
            std::u16string_view(u"backhand index pointing left: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448\U0001F3FC"),
            std::u16string_view(u"backhand_index_pointing_left_medium_light_skin_tone"),
            std::u16string_view(u"backhand index pointing left: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448\U0001F3FD"),
            std::u16string_view(u"backhand_index_pointing_left_medium_skin_tone"),
            std::u16string_view(u"backhand index pointing left: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448\U0001F3FE"),
            std::u16string_view(u"backhand_index_pointing_left_medium_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing left: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F448\U0001F3FF"),
            std::u16string_view(u"backhand_index_pointing_left_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing left: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449"),
            std::u16string_view(u"point_right"),
            std::u16string_view(u"backhand index pointing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449\U0001F3FB"),
            std::u16string_view(u"backhand_index_pointing_right_light_skin_tone"),
            std::u16string_view(u"backhand index pointing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449\U0001F3FC"),
            std::u16string_view(u"backhand_index_pointing_right_medium_light_skin_tone"),
            std::u16string_view(u"backhand index pointing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449\U0001F3FD"),
            std::u16string_view(u"backhand_index_pointing_right_medium_skin_tone"),
            std::u16string_view(u"backhand index pointing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449\U0001F3FE"),
            std::u16string_view(u"backhand_index_pointing_right_medium_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F449\U0001F3FF"),
            std::u16string_view(u"backhand_index_pointing_right_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446"),
            std::u16string_view(u"point_up_2"),
            std::u16string_view(u"backhand index pointing up"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446\U0001F3FB"),
            std::u16string_view(u"backhand_index_pointing_up_light_skin_tone"),
            std::u16string_view(u"backhand index pointing up: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446\U0001F3FC"),
            std::u16string_view(u"backhand_index_pointing_up_medium_light_skin_tone"),
            std::u16string_view(u"backhand index pointing up: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446\U0001F3FD"),
            std::u16string_view(u"backhand_index_pointing_up_medium_skin_tone"),
            std::u16string_view(u"backhand index pointing up: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446\U0001F3FE"),
            std::u16string_view(u"backhand_index_pointing_up_medium_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing up: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F446\U0001F3FF"),
            std::u16string_view(u"backhand_index_pointing_up_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing up: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595"),
            std::u16string_view(u"middle_finger"),
            std::u16string_view(u"middle finger"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595\U0001F3FB"),
            std::u16string_view(u"middle_finger_light_skin_tone"),
            std::u16string_view(u"middle finger: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595\U0001F3FC"),
            std::u16string_view(u"middle_finger_medium_light_skin_tone"),
            std::u16string_view(u"middle finger: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595\U0001F3FD"),
            std::u16string_view(u"middle_finger_medium_skin_tone"),
            std::u16string_view(u"middle finger: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595\U0001F3FE"),
            std::u16string_view(u"middle_finger_medium_dark_skin_tone"),
            std::u16string_view(u"middle finger: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F595\U0001F3FF"),
            std::u16string_view(u"middle_finger_dark_skin_tone"),
            std::u16string_view(u"middle finger: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447"),
            std::u16string_view(u"point_down"),
            std::u16string_view(u"backhand index pointing down"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447\U0001F3FB"),
            std::u16string_view(u"backhand_index_pointing_down_light_skin_tone"),
            std::u16string_view(u"backhand index pointing down: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447\U0001F3FC"),
            std::u16string_view(u"backhand_index_pointing_down_medium_light_skin_tone"),
            std::u16string_view(u"backhand index pointing down: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447\U0001F3FD"),
            std::u16string_view(u"backhand_index_pointing_down_medium_skin_tone"),
            std::u16string_view(u"backhand index pointing down: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447\U0001F3FE"),
            std::u16string_view(u"backhand_index_pointing_down_medium_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing down: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F447\U0001F3FF"),
            std::u16string_view(u"backhand_index_pointing_down_dark_skin_tone"),
            std::u16string_view(u"backhand index pointing down: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0000FE0F"),
            std::u16string_view(u"point_up"),
            std::u16string_view(u"index pointing up"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0001F3FB"),
            std::u16string_view(u"index_pointing_up_light_skin_tone"),
            std::u16string_view(u"index pointing up: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0001F3FC"),
            std::u16string_view(u"index_pointing_up_medium_light_skin_tone"),
            std::u16string_view(u"index pointing up: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0001F3FD"),
            std::u16string_view(u"index_pointing_up_medium_skin_tone"),
            std::u16string_view(u"index pointing up: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0001F3FE"),
            std::u16string_view(u"index_pointing_up_medium_dark_skin_tone"),
            std::u16string_view(u"index pointing up: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000261D\U0001F3FF"),
            std::u16string_view(u"index_pointing_up_dark_skin_tone"),
            std::u16string_view(u"index pointing up: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5"),
            std::u16string_view(u"point_you"),
            std::u16string_view(u"index pointing at the viewer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5\U0001F3FB"),
            std::u16string_view(u"index_pointing_at_the_viewer_light_skin_tone"),
            std::u16string_view(u"index pointing at the viewer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5\U0001F3FC"),
            std::u16string_view(u"index_pointing_at_the_viewer_medium_light_skin_tone"),
            std::u16string_view(u"index pointing at the viewer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5\U0001F3FD"),
            std::u16string_view(u"index_pointing_at_the_viewer_medium_skin_tone"),
            std::u16string_view(u"index pointing at the viewer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5\U0001F3FE"),
            std::u16string_view(u"index_pointing_at_the_viewer_medium_dark_skin_tone"),
            std::u16string_view(u"index pointing at the viewer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF5\U0001F3FF"),
            std::u16string_view(u"index_pointing_at_the_viewer_dark_skin_tone"),
            std::u16string_view(u"index pointing at the viewer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D"),
            std::u16string_view(u"thumbs_up"),
            std::u16string_view(u"thumbs up"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D\U0001F3FB"),
            std::u16string_view(u"thumbs_up_light_skin_tone"),
            std::u16string_view(u"thumbs up: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D\U0001F3FC"),
            std::u16string_view(u"thumbs_up_medium_light_skin_tone"),
            std::u16string_view(u"thumbs up: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D\U0001F3FD"),
            std::u16string_view(u"thumbs_up_medium_skin_tone"),
            std::u16string_view(u"thumbs up: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D\U0001F3FE"),
            std::u16string_view(u"thumbs_up_medium_dark_skin_tone"),
            std::u16string_view(u"thumbs up: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44D\U0001F3FF"),
            std::u16string_view(u"thumbs_up_dark_skin_tone"),
            std::u16string_view(u"thumbs up: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E"),
            std::u16string_view(u"thumbs_down"),
            std::u16string_view(u"thumbs down"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E\U0001F3FB"),
            std::u16string_view(u"thumbs_down_light_skin_tone"),
            std::u16string_view(u"thumbs down: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E\U0001F3FC"),
            std::u16string_view(u"thumbs_down_medium_light_skin_tone"),
            std::u16string_view(u"thumbs down: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E\U0001F3FD"),
            std::u16string_view(u"thumbs_down_medium_skin_tone"),
            std::u16string_view(u"thumbs down: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E\U0001F3FE"),
            std::u16string_view(u"thumbs_down_medium_dark_skin_tone"),
            std::u16string_view(u"thumbs down: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44E\U0001F3FF"),
            std::u16string_view(u"thumbs_down_dark_skin_tone"),
            std::u16string_view(u"thumbs down: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A"),
            std::u16string_view(u"raised_fist"),
            std::u16string_view(u"raised fist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A\U0001F3FB"),
            std::u16string_view(u"raised_fist_light_skin_tone"),
            std::u16string_view(u"raised fist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A\U0001F3FC"),
            std::u16string_view(u"raised_fist_medium_light_skin_tone"),
            std::u16string_view(u"raised fist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A\U0001F3FD"),
            std::u16string_view(u"raised_fist_medium_skin_tone"),
            std::u16string_view(u"raised fist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A\U0001F3FE"),
            std::u16string_view(u"raised_fist_medium_dark_skin_tone"),
            std::u16string_view(u"raised fist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270A\U0001F3FF"),
            std::u16string_view(u"raised_fist_dark_skin_tone"),
            std::u16string_view(u"raised fist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A"),
            std::u16string_view(u"oncoming_fist"),
            std::u16string_view(u"oncoming fist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A\U0001F3FB"),
            std::u16string_view(u"oncoming_fist_light_skin_tone"),
            std::u16string_view(u"oncoming fist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A\U0001F3FC"),
            std::u16string_view(u"oncoming_fist_medium_light_skin_tone"),
            std::u16string_view(u"oncoming fist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A\U0001F3FD"),
            std::u16string_view(u"oncoming_fist_medium_skin_tone"),
            std::u16string_view(u"oncoming fist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A\U0001F3FE"),
            std::u16string_view(u"oncoming_fist_medium_dark_skin_tone"),
            std::u16string_view(u"oncoming fist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44A\U0001F3FF"),
            std::u16string_view(u"oncoming_fist_dark_skin_tone"),
            std::u16string_view(u"oncoming fist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B"),
            std::u16string_view(u"left_facing_fist"),
            std::u16string_view(u"left-facing fist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B\U0001F3FB"),
            std::u16string_view(u"left_facing_fist_light_skin_tone"),
            std::u16string_view(u"left-facing fist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B\U0001F3FC"),
            std::u16string_view(u"left_facing_fist_medium_light_skin_tone"),
            std::u16string_view(u"left-facing fist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B\U0001F3FD"),
            std::u16string_view(u"left_facing_fist_medium_skin_tone"),
            std::u16string_view(u"left-facing fist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B\U0001F3FE"),
            std::u16string_view(u"left_facing_fist_medium_dark_skin_tone"),
            std::u16string_view(u"left-facing fist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91B\U0001F3FF"),
            std::u16string_view(u"left_facing_fist_dark_skin_tone"),
            std::u16string_view(u"left-facing fist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C"),
            std::u16string_view(u"right_facing_fist"),
            std::u16string_view(u"right-facing fist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C\U0001F3FB"),
            std::u16string_view(u"right_facing_fist_light_skin_tone"),
            std::u16string_view(u"right-facing fist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C\U0001F3FC"),
            std::u16string_view(u"right_facing_fist_medium_light_skin_tone"),
            std::u16string_view(u"right-facing fist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C\U0001F3FD"),
            std::u16string_view(u"right_facing_fist_medium_skin_tone"),
            std::u16string_view(u"right-facing fist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C\U0001F3FE"),
            std::u16string_view(u"right_facing_fist_medium_dark_skin_tone"),
            std::u16string_view(u"right-facing fist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91C\U0001F3FF"),
            std::u16string_view(u"right_facing_fist_dark_skin_tone"),
            std::u16string_view(u"right-facing fist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F"),
            std::u16string_view(u"clapping_hands"),
            std::u16string_view(u"clapping hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F\U0001F3FB"),
            std::u16string_view(u"clapping_hands_light_skin_tone"),
            std::u16string_view(u"clapping hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F\U0001F3FC"),
            std::u16string_view(u"clapping_hands_medium_light_skin_tone"),
            std::u16string_view(u"clapping hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F\U0001F3FD"),
            std::u16string_view(u"clapping_hands_medium_skin_tone"),
            std::u16string_view(u"clapping hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F\U0001F3FE"),
            std::u16string_view(u"clapping_hands_medium_dark_skin_tone"),
            std::u16string_view(u"clapping hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F44F\U0001F3FF"),
            std::u16string_view(u"clapping_hands_dark_skin_tone"),
            std::u16string_view(u"clapping hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C"),
            std::u16string_view(u"raising_hands"),
            std::u16string_view(u"raising hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C\U0001F3FB"),
            std::u16string_view(u"raising_hands_light_skin_tone"),
            std::u16string_view(u"raising hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C\U0001F3FC"),
            std::u16string_view(u"raising_hands_medium_light_skin_tone"),
            std::u16string_view(u"raising hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C\U0001F3FD"),
            std::u16string_view(u"raising_hands_medium_skin_tone"),
            std::u16string_view(u"raising hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C\U0001F3FE"),
            std::u16string_view(u"raising_hands_medium_dark_skin_tone"),
            std::u16string_view(u"raising hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64C\U0001F3FF"),
            std::u16string_view(u"raising_hands_dark_skin_tone"),
            std::u16string_view(u"raising hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6"),
            std::u16string_view(u"heart_hands"),
            std::u16string_view(u"heart hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6\U0001F3FB"),
            std::u16string_view(u"heart_hands_light_skin_tone"),
            std::u16string_view(u"heart hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6\U0001F3FC"),
            std::u16string_view(u"heart_hands_medium_light_skin_tone"),
            std::u16string_view(u"heart hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6\U0001F3FD"),
            std::u16string_view(u"heart_hands_medium_skin_tone"),
            std::u16string_view(u"heart hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6\U0001F3FE"),
            std::u16string_view(u"heart_hands_medium_dark_skin_tone"),
            std::u16string_view(u"heart hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF6\U0001F3FF"),
            std::u16string_view(u"heart_hands_dark_skin_tone"),
            std::u16string_view(u"heart hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450"),
            std::u16string_view(u"open_hands"),
            std::u16string_view(u"open hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450\U0001F3FB"),
            std::u16string_view(u"open_hands_light_skin_tone"),
            std::u16string_view(u"open hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450\U0001F3FC"),
            std::u16string_view(u"open_hands_medium_light_skin_tone"),
            std::u16string_view(u"open hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450\U0001F3FD"),
            std::u16string_view(u"open_hands_medium_skin_tone"),
            std::u16string_view(u"open hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450\U0001F3FE"),
            std::u16string_view(u"open_hands_medium_dark_skin_tone"),
            std::u16string_view(u"open hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F450\U0001F3FF"),
            std::u16string_view(u"open_hands_dark_skin_tone"),
            std::u16string_view(u"open hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932"),
            std::u16string_view(u"palms_up_together"),
            std::u16string_view(u"palms up together"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932\U0001F3FB"),
            std::u16string_view(u"palms_up_together_light_skin_tone"),
            std::u16string_view(u"palms up together: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932\U0001F3FC"),
            std::u16string_view(u"palms_up_together_medium_light_skin_tone"),
            std::u16string_view(u"palms up together: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932\U0001F3FD"),
            std::u16string_view(u"palms_up_together_medium_skin_tone"),
            std::u16string_view(u"palms up together: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932\U0001F3FE"),
            std::u16string_view(u"palms_up_together_medium_dark_skin_tone"),
            std::u16string_view(u"palms up together: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F932\U0001F3FF"),
            std::u16string_view(u"palms_up_together_dark_skin_tone"),
            std::u16string_view(u"palms up together: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D"),
            std::u16string_view(u"handshake"),
            std::u16string_view(u"handshake"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D\U0001F3FB"),
            std::u16string_view(u"handshake_light_skin_tone"),
            std::u16string_view(u"handshake: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D\U0001F3FC"),
            std::u16string_view(u"handshake_medium_light_skin_tone"),
            std::u16string_view(u"handshake: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D\U0001F3FD"),
            std::u16string_view(u"handshake_medium_skin_tone"),
            std::u16string_view(u"handshake: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D\U0001F3FE"),
            std::u16string_view(u"handshake_medium_dark_skin_tone"),
            std::u16string_view(u"handshake: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F91D\U0001F3FF"),
            std::u16string_view(u"handshake_dark_skin_tone"),
            std::u16string_view(u"handshake: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FC"),
            std::u16string_view(u"handshake_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"handshake: light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FD"),
            std::u16string_view(u"handshake_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"handshake: light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FE"),
            std::u16string_view(u"handshake_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"handshake: light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FF"),
            std::u16string_view(u"handshake_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"handshake: light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FB"),
            std::u16string_view(u"handshake_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"handshake: medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FD"),
            std::u16string_view(u"handshake_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"handshake: medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FE"),
            std::u16string_view(u"handshake_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"handshake: medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FF"),
            std::u16string_view(u"handshake_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"handshake: medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FB"),
            std::u16string_view(u"handshake_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"handshake: medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FC"),
            std::u16string_view(u"handshake_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"handshake: medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FE"),
            std::u16string_view(u"handshake_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"handshake: medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FF"),
            std::u16string_view(u"handshake_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"handshake: medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FB"),
            std::u16string_view(u"handshake_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"handshake: medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FC"),
            std::u16string_view(u"handshake_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"handshake: medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FD"),
            std::u16string_view(u"handshake_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"handshake: medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FF"),
            std::u16string_view(u"handshake_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"handshake: medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FB"),
            std::u16string_view(u"handshake_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"handshake: dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FC"),
            std::u16string_view(u"handshake_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"handshake: dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FD"),
            std::u16string_view(u"handshake_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"handshake: dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FE"),
            std::u16string_view(u"handshake_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"handshake: dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F"),
            std::u16string_view(u"folded_hands"),
            std::u16string_view(u"folded hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F\U0001F3FB"),
            std::u16string_view(u"folded_hands_light_skin_tone"),
            std::u16string_view(u"folded hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F\U0001F3FC"),
            std::u16string_view(u"folded_hands_medium_light_skin_tone"),
            std::u16string_view(u"folded hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F\U0001F3FD"),
            std::u16string_view(u"folded_hands_medium_skin_tone"),
            std::u16string_view(u"folded hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F\U0001F3FE"),
            std::u16string_view(u"folded_hands_medium_dark_skin_tone"),
            std::u16string_view(u"folded hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64F\U0001F3FF"),
            std::u16string_view(u"folded_hands_dark_skin_tone"),
            std::u16string_view(u"folded hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0000FE0F"),
            std::u16string_view(u"writing_hand"),
            std::u16string_view(u"writing hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0001F3FB"),
            std::u16string_view(u"writing_hand_light_skin_tone"),
            std::u16string_view(u"writing hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0001F3FC"),
            std::u16string_view(u"writing_hand_medium_light_skin_tone"),
            std::u16string_view(u"writing hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0001F3FD"),
            std::u16string_view(u"writing_hand_medium_skin_tone"),
            std::u16string_view(u"writing hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0001F3FE"),
            std::u16string_view(u"writing_hand_medium_dark_skin_tone"),
            std::u16string_view(u"writing hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0000270D\U0001F3FF"),
            std::u16string_view(u"writing_hand_dark_skin_tone"),
            std::u16string_view(u"writing hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485"),
            std::u16string_view(u"nail_polish"),
            std::u16string_view(u"nail polish"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485\U0001F3FB"),
            std::u16string_view(u"nail_polish_light_skin_tone"),
            std::u16string_view(u"nail polish: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485\U0001F3FC"),
            std::u16string_view(u"nail_polish_medium_light_skin_tone"),
            std::u16string_view(u"nail polish: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485\U0001F3FD"),
            std::u16string_view(u"nail_polish_medium_skin_tone"),
            std::u16string_view(u"nail polish: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485\U0001F3FE"),
            std::u16string_view(u"nail_polish_medium_dark_skin_tone"),
            std::u16string_view(u"nail polish: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F485\U0001F3FF"),
            std::u16string_view(u"nail_polish_dark_skin_tone"),
            std::u16string_view(u"nail polish: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933"),
            std::u16string_view(u"selfie"),
            std::u16string_view(u"selfie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933\U0001F3FB"),
            std::u16string_view(u"selfie_light_skin_tone"),
            std::u16string_view(u"selfie: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933\U0001F3FC"),
            std::u16string_view(u"selfie_medium_light_skin_tone"),
            std::u16string_view(u"selfie: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933\U0001F3FD"),
            std::u16string_view(u"selfie_medium_skin_tone"),
            std::u16string_view(u"selfie: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933\U0001F3FE"),
            std::u16string_view(u"selfie_medium_dark_skin_tone"),
            std::u16string_view(u"selfie: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F933\U0001F3FF"),
            std::u16string_view(u"selfie_dark_skin_tone"),
            std::u16string_view(u"selfie: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA"),
            std::u16string_view(u"flexed_biceps"),
            std::u16string_view(u"flexed biceps"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA\U0001F3FB"),
            std::u16string_view(u"flexed_biceps_light_skin_tone"),
            std::u16string_view(u"flexed biceps: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA\U0001F3FC"),
            std::u16string_view(u"flexed_biceps_medium_light_skin_tone"),
            std::u16string_view(u"flexed biceps: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA\U0001F3FD"),
            std::u16string_view(u"flexed_biceps_medium_skin_tone"),
            std::u16string_view(u"flexed biceps: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA\U0001F3FE"),
            std::u16string_view(u"flexed_biceps_medium_dark_skin_tone"),
            std::u16string_view(u"flexed biceps: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F4AA\U0001F3FF"),
            std::u16string_view(u"flexed_biceps_dark_skin_tone"),
            std::u16string_view(u"flexed biceps: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BE"),
            std::u16string_view(u"mechanical_arm"),
            std::u16string_view(u"mechanical arm"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BF"),
            std::u16string_view(u"mechanical_leg"),
            std::u16string_view(u"mechanical leg"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5"),
            std::u16string_view(u"leg"),
            std::u16string_view(u"leg"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5\U0001F3FB"),
            std::u16string_view(u"leg_light_skin_tone"),
            std::u16string_view(u"leg: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5\U0001F3FC"),
            std::u16string_view(u"leg_medium_light_skin_tone"),
            std::u16string_view(u"leg: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5\U0001F3FD"),
            std::u16string_view(u"leg_medium_skin_tone"),
            std::u16string_view(u"leg: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5\U0001F3FE"),
            std::u16string_view(u"leg_medium_dark_skin_tone"),
            std::u16string_view(u"leg: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B5\U0001F3FF"),
            std::u16string_view(u"leg_dark_skin_tone"),
            std::u16string_view(u"leg: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6"),
            std::u16string_view(u"foot"),
            std::u16string_view(u"foot"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6\U0001F3FB"),
            std::u16string_view(u"foot_light_skin_tone"),
            std::u16string_view(u"foot: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6\U0001F3FC"),
            std::u16string_view(u"foot_medium_light_skin_tone"),
            std::u16string_view(u"foot: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6\U0001F3FD"),
            std::u16string_view(u"foot_medium_skin_tone"),
            std::u16string_view(u"foot: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6\U0001F3FE"),
            std::u16string_view(u"foot_medium_dark_skin_tone"),
            std::u16string_view(u"foot: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B6\U0001F3FF"),
            std::u16string_view(u"foot_dark_skin_tone"),
            std::u16string_view(u"foot: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442"),
            std::u16string_view(u"ear"),
            std::u16string_view(u"ear"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442\U0001F3FB"),
            std::u16string_view(u"ear_light_skin_tone"),
            std::u16string_view(u"ear: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442\U0001F3FC"),
            std::u16string_view(u"ear_medium_light_skin_tone"),
            std::u16string_view(u"ear: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442\U0001F3FD"),
            std::u16string_view(u"ear_medium_skin_tone"),
            std::u16string_view(u"ear: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442\U0001F3FE"),
            std::u16string_view(u"ear_medium_dark_skin_tone"),
            std::u16string_view(u"ear: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F442\U0001F3FF"),
            std::u16string_view(u"ear_dark_skin_tone"),
            std::u16string_view(u"ear: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB"),
            std::u16string_view(u"ear_with_hearing_aid"),
            std::u16string_view(u"ear with hearing aid"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB\U0001F3FB"),
            std::u16string_view(u"ear_with_hearing_aid_light_skin_tone"),
            std::u16string_view(u"ear with hearing aid: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB\U0001F3FC"),
            std::u16string_view(u"ear_with_hearing_aid_medium_light_skin_tone"),
            std::u16string_view(u"ear with hearing aid: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB\U0001F3FD"),
            std::u16string_view(u"ear_with_hearing_aid_medium_skin_tone"),
            std::u16string_view(u"ear with hearing aid: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB\U0001F3FE"),
            std::u16string_view(u"ear_with_hearing_aid_medium_dark_skin_tone"),
            std::u16string_view(u"ear with hearing aid: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9BB\U0001F3FF"),
            std::u16string_view(u"ear_with_hearing_aid_dark_skin_tone"),
            std::u16string_view(u"ear with hearing aid: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443"),
            std::u16string_view(u"nose"),
            std::u16string_view(u"nose"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443\U0001F3FB"),
            std::u16string_view(u"nose_light_skin_tone"),
            std::u16string_view(u"nose: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443\U0001F3FC"),
            std::u16string_view(u"nose_medium_light_skin_tone"),
            std::u16string_view(u"nose: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443\U0001F3FD"),
            std::u16string_view(u"nose_medium_skin_tone"),
            std::u16string_view(u"nose: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443\U0001F3FE"),
            std::u16string_view(u"nose_medium_dark_skin_tone"),
            std::u16string_view(u"nose: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F443\U0001F3FF"),
            std::u16string_view(u"nose_dark_skin_tone"),
            std::u16string_view(u"nose: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9E0"),
            std::u16string_view(u"brain"),
            std::u16string_view(u"brain"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC0"),
            std::u16string_view(u"anatomical_heart"),
            std::u16string_view(u"anatomical heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC1"),
            std::u16string_view(u"lungs"),
            std::u16string_view(u"lungs"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B7"),
            std::u16string_view(u"tooth"),
            std::u16string_view(u"tooth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B4"),
            std::u16string_view(u"bone"),
            std::u16string_view(u"bone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F440"),
            std::u16string_view(u"eyes"),
            std::u16string_view(u"eyes"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F441\U0000FE0F"),
            std::u16string_view(u"eye"),
            std::u16string_view(u"eye"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F445"),
            std::u16string_view(u"tongue"),
            std::u16string_view(u"tongue"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F444"),
            std::u16string_view(u"mouth"),
            std::u16string_view(u"mouth"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAE6"),
            std::u16string_view(u"biting_lip"),
            std::u16string_view(u"biting lip"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476"),
            std::u16string_view(u"baby"),
            std::u16string_view(u"baby"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476\U0001F3FB"),
            std::u16string_view(u"baby_light_skin_tone"),
            std::u16string_view(u"baby: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476\U0001F3FC"),
            std::u16string_view(u"baby_medium_light_skin_tone"),
            std::u16string_view(u"baby: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476\U0001F3FD"),
            std::u16string_view(u"baby_medium_skin_tone"),
            std::u16string_view(u"baby: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476\U0001F3FE"),
            std::u16string_view(u"baby_medium_dark_skin_tone"),
            std::u16string_view(u"baby: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F476\U0001F3FF"),
            std::u16string_view(u"baby_dark_skin_tone"),
            std::u16string_view(u"baby: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2"),
            std::u16string_view(u"child"),
            std::u16string_view(u"child"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2\U0001F3FB"),
            std::u16string_view(u"child_light_skin_tone"),
            std::u16string_view(u"child: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2\U0001F3FC"),
            std::u16string_view(u"child_medium_light_skin_tone"),
            std::u16string_view(u"child: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2\U0001F3FD"),
            std::u16string_view(u"child_medium_skin_tone"),
            std::u16string_view(u"child: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2\U0001F3FE"),
            std::u16string_view(u"child_medium_dark_skin_tone"),
            std::u16string_view(u"child: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D2\U0001F3FF"),
            std::u16string_view(u"child_dark_skin_tone"),
            std::u16string_view(u"child: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466"),
            std::u16string_view(u"boy"),
            std::u16string_view(u"boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466\U0001F3FB"),
            std::u16string_view(u"boy_light_skin_tone"),
            std::u16string_view(u"boy: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466\U0001F3FC"),
            std::u16string_view(u"boy_medium_light_skin_tone"),
            std::u16string_view(u"boy: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466\U0001F3FD"),
            std::u16string_view(u"boy_medium_skin_tone"),
            std::u16string_view(u"boy: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466\U0001F3FE"),
            std::u16string_view(u"boy_medium_dark_skin_tone"),
            std::u16string_view(u"boy: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F466\U0001F3FF"),
            std::u16string_view(u"boy_dark_skin_tone"),
            std::u16string_view(u"boy: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467"),
            std::u16string_view(u"girl"),
            std::u16string_view(u"girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467\U0001F3FB"),
            std::u16string_view(u"girl_light_skin_tone"),
            std::u16string_view(u"girl: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467\U0001F3FC"),
            std::u16string_view(u"girl_medium_light_skin_tone"),
            std::u16string_view(u"girl: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467\U0001F3FD"),
            std::u16string_view(u"girl_medium_skin_tone"),
            std::u16string_view(u"girl: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467\U0001F3FE"),
            std::u16string_view(u"girl_medium_dark_skin_tone"),
            std::u16string_view(u"girl: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F467\U0001F3FF"),
            std::u16string_view(u"girl_dark_skin_tone"),
            std::u16string_view(u"girl: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1"),
            std::u16string_view(u"person"),
            std::u16string_view(u"person"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"person_light_skin_tone"),
            std::u16string_view(u"person: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"person_medium_light_skin_tone"),
            std::u16string_view(u"person: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"person_medium_skin_tone"),
            std::u16string_view(u"person: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"person_medium_dark_skin_tone"),
            std::u16string_view(u"person: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"person_dark_skin_tone"),
            std::u16string_view(u"person: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471"),
            std::u16string_view(u"person_blond_hair"),
            std::u16string_view(u"person: blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FB"),
            std::u16string_view(u"person_light_skin_tone_blond_hair"),
            std::u16string_view(u"person: light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FC"),
            std::u16string_view(u"person_medium_light_skin_tone_blond_hair"),
            std::u16string_view(u"person: medium-light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FD"),
            std::u16string_view(u"person_medium_skin_tone_blond_hair"),
            std::u16string_view(u"person: medium skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FE"),
            std::u16string_view(u"person_medium_dark_skin_tone_blond_hair"),
            std::u16string_view(u"person: medium-dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FF"),
            std::u16string_view(u"person_dark_skin_tone_blond_hair"),
            std::u16string_view(u"person: dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468"),
            std::u16string_view(u"man"),
            std::u16string_view(u"man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB"),
            std::u16string_view(u"man_light_skin_tone"),
            std::u16string_view(u"man: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC"),
            std::u16string_view(u"man_medium_light_skin_tone"),
            std::u16string_view(u"man: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD"),
            std::u16string_view(u"man_medium_skin_tone"),
            std::u16string_view(u"man: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE"),
            std::u16string_view(u"man_medium_dark_skin_tone"),
            std::u16string_view(u"man: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF"),
            std::u16string_view(u"man_dark_skin_tone"),
            std::u16string_view(u"man: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4"),
            std::u16string_view(u"bearded_person"),
            std::u16string_view(u"person: beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FB"),
            std::u16string_view(u"person_light_skin_tone_beard"),
            std::u16string_view(u"person: light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FC"),
            std::u16string_view(u"person_medium_light_skin_tone_beard"),
            std::u16string_view(u"person: medium-light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FD"),
            std::u16string_view(u"person_medium_skin_tone_beard"),
            std::u16string_view(u"person: medium skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FE"),
            std::u16string_view(u"person_medium_dark_skin_tone_beard"),
            std::u16string_view(u"person: medium-dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FF"),
            std::u16string_view(u"person_dark_skin_tone_beard"),
            std::u16string_view(u"person: dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"bearded_man"),
            std::u16string_view(u"man: beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_light_skin_tone_beard"),
            std::u16string_view(u"man: light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_light_skin_tone_beard"),
            std::u16string_view(u"man: medium-light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_skin_tone_beard"),
            std::u16string_view(u"man: medium skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_dark_skin_tone_beard"),
            std::u16string_view(u"man: medium-dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_dark_skin_tone_beard"),
            std::u16string_view(u"man: dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"bearded_woman"),
            std::u16string_view(u"woman: beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_light_skin_tone_beard"),
            std::u16string_view(u"woman: light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_light_skin_tone_beard"),
            std::u16string_view(u"woman: medium-light skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_skin_tone_beard"),
            std::u16string_view(u"woman: medium skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_dark_skin_tone_beard"),
            std::u16string_view(u"woman: medium-dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D4\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_dark_skin_tone_beard"),
            std::u16string_view(u"woman: dark skin tone, beard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9B0"),
            std::u16string_view(u"red_haired_man"),
            std::u16string_view(u"man: red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9B0"),
            std::u16string_view(u"man_light_skin_tone_red_hair"),
            std::u16string_view(u"man: light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9B0"),
            std::u16string_view(u"man_medium_light_skin_tone_red_hair"),
            std::u16string_view(u"man: medium-light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9B0"),
            std::u16string_view(u"man_medium_skin_tone_red_hair"),
            std::u16string_view(u"man: medium skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9B0"),
            std::u16string_view(u"man_medium_dark_skin_tone_red_hair"),
            std::u16string_view(u"man: medium-dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9B0"),
            std::u16string_view(u"man_dark_skin_tone_red_hair"),
            std::u16string_view(u"man: dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9B1"),
            std::u16string_view(u"curly_haired_man"),
            std::u16string_view(u"man: curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9B1"),
            std::u16string_view(u"man_light_skin_tone_curly_hair"),
            std::u16string_view(u"man: light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9B1"),
            std::u16string_view(u"man_medium_light_skin_tone_curly_hair"),
            std::u16string_view(u"man: medium-light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9B1"),
            std::u16string_view(u"man_medium_skin_tone_curly_hair"),
            std::u16string_view(u"man: medium skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9B1"),
            std::u16string_view(u"man_medium_dark_skin_tone_curly_hair"),
            std::u16string_view(u"man: medium-dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9B1"),
            std::u16string_view(u"man_dark_skin_tone_curly_hair"),
            std::u16string_view(u"man: dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9B3"),
            std::u16string_view(u"white_haired_man"),
            std::u16string_view(u"man: white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9B3"),
            std::u16string_view(u"man_light_skin_tone_white_hair"),
            std::u16string_view(u"man: light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9B3"),
            std::u16string_view(u"man_medium_light_skin_tone_white_hair"),
            std::u16string_view(u"man: medium-light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9B3"),
            std::u16string_view(u"man_medium_skin_tone_white_hair"),
            std::u16string_view(u"man: medium skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9B3"),
            std::u16string_view(u"man_medium_dark_skin_tone_white_hair"),
            std::u16string_view(u"man: medium-dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9B3"),
            std::u16string_view(u"man_dark_skin_tone_white_hair"),
            std::u16string_view(u"man: dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9B2"),
            std::u16string_view(u"bald_man"),
            std::u16string_view(u"man: bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9B2"),
            std::u16string_view(u"man_light_skin_tone_bald"),
            std::u16string_view(u"man: light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9B2"),
            std::u16string_view(u"man_medium_light_skin_tone_bald"),
            std::u16string_view(u"man: medium-light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9B2"),
            std::u16string_view(u"man_medium_skin_tone_bald"),
            std::u16string_view(u"man: medium skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9B2"),
            std::u16string_view(u"man_medium_dark_skin_tone_bald"),
            std::u16string_view(u"man: medium-dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9B2"),
            std::u16string_view(u"man_dark_skin_tone_bald"),
            std::u16string_view(u"man: dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469"),
            std::u16string_view(u"woman"),
            std::u16string_view(u"woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB"),
            std::u16string_view(u"woman_light_skin_tone"),
            std::u16string_view(u"woman: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC"),
            std::u16string_view(u"woman_medium_light_skin_tone"),
            std::u16string_view(u"woman: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD"),
            std::u16string_view(u"woman_medium_skin_tone"),
            std::u16string_view(u"woman: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE"),
            std::u16string_view(u"woman_medium_dark_skin_tone"),
            std::u16string_view(u"woman: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF"),
            std::u16string_view(u"woman_dark_skin_tone"),
            std::u16string_view(u"woman: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9B0"),
            std::u16string_view(u"red_haired_woman"),
            std::u16string_view(u"woman: red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9B0"),
            std::u16string_view(u"woman_light_skin_tone_red_hair"),
            std::u16string_view(u"woman: light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9B0"),
            std::u16string_view(u"woman_medium_light_skin_tone_red_hair"),
            std::u16string_view(u"woman: medium-light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9B0"),
            std::u16string_view(u"woman_medium_skin_tone_red_hair"),
            std::u16string_view(u"woman: medium skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9B0"),
            std::u16string_view(u"woman_medium_dark_skin_tone_red_hair"),
            std::u16string_view(u"woman: medium-dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9B0"),
            std::u16string_view(u"woman_dark_skin_tone_red_hair"),
            std::u16string_view(u"woman: dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9B0"),
            std::u16string_view(u"red_haired_person"),
            std::u16string_view(u"person: red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9B0"),
            std::u16string_view(u"person_light_skin_tone_red_hair"),
            std::u16string_view(u"person: light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9B0"),
            std::u16string_view(u"person_medium_light_skin_tone_red_hair"),
            std::u16string_view(u"person: medium-light skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9B0"),
            std::u16string_view(u"person_medium_skin_tone_red_hair"),
            std::u16string_view(u"person: medium skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9B0"),
            std::u16string_view(u"person_medium_dark_skin_tone_red_hair"),
            std::u16string_view(u"person: medium-dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9B0"),
            std::u16string_view(u"person_dark_skin_tone_red_hair"),
            std::u16string_view(u"person: dark skin tone, red hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9B1"),
            std::u16string_view(u"curly_haired_woman"),
            std::u16string_view(u"woman: curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9B1"),
            std::u16string_view(u"woman_light_skin_tone_curly_hair"),
            std::u16string_view(u"woman: light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9B1"),
            std::u16string_view(u"woman_medium_light_skin_tone_curly_hair"),
            std::u16string_view(u"woman: medium-light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9B1"),
            std::u16string_view(u"woman_medium_skin_tone_curly_hair"),
            std::u16string_view(u"woman: medium skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9B1"),
            std::u16string_view(u"woman_medium_dark_skin_tone_curly_hair"),
            std::u16string_view(u"woman: medium-dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9B1"),
            std::u16string_view(u"woman_dark_skin_tone_curly_hair"),
            std::u16string_view(u"woman: dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9B1"),
            std::u16string_view(u"curly_haired_person"),
            std::u16string_view(u"person: curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9B1"),
            std::u16string_view(u"person_light_skin_tone_curly_hair"),
            std::u16string_view(u"person: light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9B1"),
            std::u16string_view(u"person_medium_light_skin_tone_curly_hair"),
            std::u16string_view(u"person: medium-light skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9B1"),
            std::u16string_view(u"person_medium_skin_tone_curly_hair"),
            std::u16string_view(u"person: medium skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9B1"),
            std::u16string_view(u"person_medium_dark_skin_tone_curly_hair"),
            std::u16string_view(u"person: medium-dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9B1"),
            std::u16string_view(u"person_dark_skin_tone_curly_hair"),
            std::u16string_view(u"person: dark skin tone, curly hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9B3"),
            std::u16string_view(u"white_haired_woman"),
            std::u16string_view(u"woman: white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9B3"),
            std::u16string_view(u"woman_light_skin_tone_white_hair"),
            std::u16string_view(u"woman: light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9B3"),
            std::u16string_view(u"woman_medium_light_skin_tone_white_hair"),
            std::u16string_view(u"woman: medium-light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9B3"),
            std::u16string_view(u"woman_medium_skin_tone_white_hair"),
            std::u16string_view(u"woman: medium skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9B3"),
            std::u16string_view(u"woman_medium_dark_skin_tone_white_hair"),
            std::u16string_view(u"woman: medium-dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9B3"),
            std::u16string_view(u"woman_dark_skin_tone_white_hair"),
            std::u16string_view(u"woman: dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9B3"),
            std::u16string_view(u"white_haired_person"),
            std::u16string_view(u"person: white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9B3"),
            std::u16string_view(u"person_light_skin_tone_white_hair"),
            std::u16string_view(u"person: light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9B3"),
            std::u16string_view(u"person_medium_light_skin_tone_white_hair"),
            std::u16string_view(u"person: medium-light skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9B3"),
            std::u16string_view(u"person_medium_skin_tone_white_hair"),
            std::u16string_view(u"person: medium skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9B3"),
            std::u16string_view(u"person_medium_dark_skin_tone_white_hair"),
            std::u16string_view(u"person: medium-dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9B3"),
            std::u16string_view(u"person_dark_skin_tone_white_hair"),
            std::u16string_view(u"person: dark skin tone, white hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9B2"),
            std::u16string_view(u"bald_woman"),
            std::u16string_view(u"woman: bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9B2"),
            std::u16string_view(u"woman_light_skin_tone_bald"),
            std::u16string_view(u"woman: light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9B2"),
            std::u16string_view(u"woman_medium_light_skin_tone_bald"),
            std::u16string_view(u"woman: medium-light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9B2"),
            std::u16string_view(u"woman_medium_skin_tone_bald"),
            std::u16string_view(u"woman: medium skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9B2"),
            std::u16string_view(u"woman_medium_dark_skin_tone_bald"),
            std::u16string_view(u"woman: medium-dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9B2"),
            std::u16string_view(u"woman_dark_skin_tone_bald"),
            std::u16string_view(u"woman: dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9B2"),
            std::u16string_view(u"bald_person"),
            std::u16string_view(u"person: bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9B2"),
            std::u16string_view(u"person_light_skin_tone_bald"),
            std::u16string_view(u"person: light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9B2"),
            std::u16string_view(u"person_medium_light_skin_tone_bald"),
            std::u16string_view(u"person: medium-light skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9B2"),
            std::u16string_view(u"person_medium_skin_tone_bald"),
            std::u16string_view(u"person: medium skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9B2"),
            std::u16string_view(u"person_medium_dark_skin_tone_bald"),
            std::u16string_view(u"person: medium-dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9B2"),
            std::u16string_view(u"person_dark_skin_tone_bald"),
            std::u16string_view(u"person: dark skin tone, bald"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_blond_hair"),
            std::u16string_view(u"woman: blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_light_skin_tone_blond_hair"),
            std::u16string_view(u"woman: light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_light_skin_tone_blond_hair"),
            std::u16string_view(u"woman: medium-light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_skin_tone_blond_hair"),
            std::u16string_view(u"woman: medium skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_medium_dark_skin_tone_blond_hair"),
            std::u16string_view(u"woman: medium-dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_dark_skin_tone_blond_hair"),
            std::u16string_view(u"woman: dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_blond_hair"),
            std::u16string_view(u"man: blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_light_skin_tone_blond_hair"),
            std::u16string_view(u"man: light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_light_skin_tone_blond_hair"),
            std::u16string_view(u"man: medium-light skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_skin_tone_blond_hair"),
            std::u16string_view(u"man: medium skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_medium_dark_skin_tone_blond_hair"),
            std::u16string_view(u"man: medium-dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F471\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_dark_skin_tone_blond_hair"),
            std::u16string_view(u"man: dark skin tone, blond hair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3"),
            std::u16string_view(u"older_person"),
            std::u16string_view(u"older person"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3\U0001F3FB"),
            std::u16string_view(u"older_person_light_skin_tone"),
            std::u16string_view(u"older person: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3\U0001F3FC"),
            std::u16string_view(u"older_person_medium_light_skin_tone"),
            std::u16string_view(u"older person: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3\U0001F3FD"),
            std::u16string_view(u"older_person_medium_skin_tone"),
            std::u16string_view(u"older person: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3\U0001F3FE"),
            std::u16string_view(u"older_person_medium_dark_skin_tone"),
            std::u16string_view(u"older person: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D3\U0001F3FF"),
            std::u16string_view(u"older_person_dark_skin_tone"),
            std::u16string_view(u"older person: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474"),
            std::u16string_view(u"old_man"),
            std::u16string_view(u"old man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474\U0001F3FB"),
            std::u16string_view(u"old_man_light_skin_tone"),
            std::u16string_view(u"old man: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474\U0001F3FC"),
            std::u16string_view(u"old_man_medium_light_skin_tone"),
            std::u16string_view(u"old man: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474\U0001F3FD"),
            std::u16string_view(u"old_man_medium_skin_tone"),
            std::u16string_view(u"old man: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474\U0001F3FE"),
            std::u16string_view(u"old_man_medium_dark_skin_tone"),
            std::u16string_view(u"old man: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F474\U0001F3FF"),
            std::u16string_view(u"old_man_dark_skin_tone"),
            std::u16string_view(u"old man: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475"),
            std::u16string_view(u"old_woman"),
            std::u16string_view(u"old woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475\U0001F3FB"),
            std::u16string_view(u"old_woman_light_skin_tone"),
            std::u16string_view(u"old woman: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475\U0001F3FC"),
            std::u16string_view(u"old_woman_medium_light_skin_tone"),
            std::u16string_view(u"old woman: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475\U0001F3FD"),
            std::u16string_view(u"old_woman_medium_skin_tone"),
            std::u16string_view(u"old woman: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475\U0001F3FE"),
            std::u16string_view(u"old_woman_medium_dark_skin_tone"),
            std::u16string_view(u"old woman: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F475\U0001F3FF"),
            std::u16string_view(u"old_woman_dark_skin_tone"),
            std::u16string_view(u"old woman: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D"),
            std::u16string_view(u"person_frowning"),
            std::u16string_view(u"person frowning"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FB"),
            std::u16string_view(u"person_frowning_light_skin_tone"),
            std::u16string_view(u"person frowning: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FC"),
            std::u16string_view(u"person_frowning_medium_light_skin_tone"),
            std::u16string_view(u"person frowning: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FD"),
            std::u16string_view(u"person_frowning_medium_skin_tone"),
            std::u16string_view(u"person frowning: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FE"),
            std::u16string_view(u"person_frowning_medium_dark_skin_tone"),
            std::u16string_view(u"person frowning: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FF"),
            std::u16string_view(u"person_frowning_dark_skin_tone"),
            std::u16string_view(u"person frowning: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning"),
            std::u16string_view(u"man frowning"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning_light_skin_tone"),
            std::u16string_view(u"man frowning: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning_medium_light_skin_tone"),
            std::u16string_view(u"man frowning: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning_medium_skin_tone"),
            std::u16string_view(u"man frowning: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning_medium_dark_skin_tone"),
            std::u16string_view(u"man frowning: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_frowning_dark_skin_tone"),
            std::u16string_view(u"man frowning: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning"),
            std::u16string_view(u"woman frowning"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning_light_skin_tone"),
            std::u16string_view(u"woman frowning: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning_medium_light_skin_tone"),
            std::u16string_view(u"woman frowning: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning_medium_skin_tone"),
            std::u16string_view(u"woman frowning: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning_medium_dark_skin_tone"),
            std::u16string_view(u"woman frowning: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64D\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_frowning_dark_skin_tone"),
            std::u16string_view(u"woman frowning: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E"),
            std::u16string_view(u"person_pouting"),
            std::u16string_view(u"person pouting"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FB"),
            std::u16string_view(u"person_pouting_light_skin_tone"),
            std::u16string_view(u"person pouting: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FC"),
            std::u16string_view(u"person_pouting_medium_light_skin_tone"),
            std::u16string_view(u"person pouting: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FD"),
            std::u16string_view(u"person_pouting_medium_skin_tone"),
            std::u16string_view(u"person pouting: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FE"),
            std::u16string_view(u"person_pouting_medium_dark_skin_tone"),
            std::u16string_view(u"person pouting: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FF"),
            std::u16string_view(u"person_pouting_dark_skin_tone"),
            std::u16string_view(u"person pouting: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting"),
            std::u16string_view(u"man pouting"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting_light_skin_tone"),
            std::u16string_view(u"man pouting: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting_medium_light_skin_tone"),
            std::u16string_view(u"man pouting: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting_medium_skin_tone"),
            std::u16string_view(u"man pouting: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting_medium_dark_skin_tone"),
            std::u16string_view(u"man pouting: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_pouting_dark_skin_tone"),
            std::u16string_view(u"man pouting: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting"),
            std::u16string_view(u"woman pouting"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting_light_skin_tone"),
            std::u16string_view(u"woman pouting: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting_medium_light_skin_tone"),
            std::u16string_view(u"woman pouting: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting_medium_skin_tone"),
            std::u16string_view(u"woman pouting: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting_medium_dark_skin_tone"),
            std::u16string_view(u"woman pouting: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64E\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_pouting_dark_skin_tone"),
            std::u16string_view(u"woman pouting: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645"),
            std::u16string_view(u"person_gesturing_no"),
            std::u16string_view(u"person gesturing NO"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FB"),
            std::u16string_view(u"person_gesturing_no_light_skin_tone"),
            std::u16string_view(u"person gesturing NO: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FC"),
            std::u16string_view(u"person_gesturing_no_medium_light_skin_tone"),
            std::u16string_view(u"person gesturing NO: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FD"),
            std::u16string_view(u"person_gesturing_no_medium_skin_tone"),
            std::u16string_view(u"person gesturing NO: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FE"),
            std::u16string_view(u"person_gesturing_no_medium_dark_skin_tone"),
            std::u16string_view(u"person gesturing NO: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FF"),
            std::u16string_view(u"person_gesturing_no_dark_skin_tone"),
            std::u16string_view(u"person gesturing NO: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no"),
            std::u16string_view(u"man gesturing NO"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no_light_skin_tone"),
            std::u16string_view(u"man gesturing NO: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no_medium_light_skin_tone"),
            std::u16string_view(u"man gesturing NO: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no_medium_skin_tone"),
            std::u16string_view(u"man gesturing NO: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no_medium_dark_skin_tone"),
            std::u16string_view(u"man gesturing NO: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_no_dark_skin_tone"),
            std::u16string_view(u"man gesturing NO: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no"),
            std::u16string_view(u"woman gesturing NO"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no_light_skin_tone"),
            std::u16string_view(u"woman gesturing NO: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no_medium_light_skin_tone"),
            std::u16string_view(u"woman gesturing NO: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no_medium_skin_tone"),
            std::u16string_view(u"woman gesturing NO: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no_medium_dark_skin_tone"),
            std::u16string_view(u"woman gesturing NO: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F645\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_no_dark_skin_tone"),
            std::u16string_view(u"woman gesturing NO: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646"),
            std::u16string_view(u"person_gesturing_ok"),
            std::u16string_view(u"person gesturing OK"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FB"),
            std::u16string_view(u"person_gesturing_ok_light_skin_tone"),
            std::u16string_view(u"person gesturing OK: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FC"),
            std::u16string_view(u"person_gesturing_ok_medium_light_skin_tone"),
            std::u16string_view(u"person gesturing OK: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FD"),
            std::u16string_view(u"person_gesturing_ok_medium_skin_tone"),
            std::u16string_view(u"person gesturing OK: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FE"),
            std::u16string_view(u"person_gesturing_ok_medium_dark_skin_tone"),
            std::u16string_view(u"person gesturing OK: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FF"),
            std::u16string_view(u"person_gesturing_ok_dark_skin_tone"),
            std::u16string_view(u"person gesturing OK: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok"),
            std::u16string_view(u"man gesturing OK"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok_light_skin_tone"),
            std::u16string_view(u"man gesturing OK: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok_medium_light_skin_tone"),
            std::u16string_view(u"man gesturing OK: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok_medium_skin_tone"),
            std::u16string_view(u"man gesturing OK: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok_medium_dark_skin_tone"),
            std::u16string_view(u"man gesturing OK: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_gesturing_ok_dark_skin_tone"),
            std::u16string_view(u"man gesturing OK: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok"),
            std::u16string_view(u"woman gesturing OK"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok_light_skin_tone"),
            std::u16string_view(u"woman gesturing OK: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok_medium_light_skin_tone"),
            std::u16string_view(u"woman gesturing OK: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok_medium_skin_tone"),
            std::u16string_view(u"woman gesturing OK: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok_medium_dark_skin_tone"),
            std::u16string_view(u"woman gesturing OK: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F646\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_gesturing_ok_dark_skin_tone"),
            std::u16string_view(u"woman gesturing OK: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481"),
            std::u16string_view(u"person_tipping_hand"),
            std::u16string_view(u"person tipping hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FB"),
            std::u16string_view(u"person_tipping_hand_light_skin_tone"),
            std::u16string_view(u"person tipping hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FC"),
            std::u16string_view(u"person_tipping_hand_medium_light_skin_tone"),
            std::u16string_view(u"person tipping hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FD"),
            std::u16string_view(u"person_tipping_hand_medium_skin_tone"),
            std::u16string_view(u"person tipping hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FE"),
            std::u16string_view(u"person_tipping_hand_medium_dark_skin_tone"),
            std::u16string_view(u"person tipping hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FF"),
            std::u16string_view(u"person_tipping_hand_dark_skin_tone"),
            std::u16string_view(u"person tipping hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand"),
            std::u16string_view(u"man tipping hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand_light_skin_tone"),
            std::u16string_view(u"man tipping hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand_medium_light_skin_tone"),
            std::u16string_view(u"man tipping hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand_medium_skin_tone"),
            std::u16string_view(u"man tipping hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand_medium_dark_skin_tone"),
            std::u16string_view(u"man tipping hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_tipping_hand_dark_skin_tone"),
            std::u16string_view(u"man tipping hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand"),
            std::u16string_view(u"woman tipping hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand_light_skin_tone"),
            std::u16string_view(u"woman tipping hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand_medium_light_skin_tone"),
            std::u16string_view(u"woman tipping hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand_medium_skin_tone"),
            std::u16string_view(u"woman tipping hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand_medium_dark_skin_tone"),
            std::u16string_view(u"woman tipping hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F481\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_tipping_hand_dark_skin_tone"),
            std::u16string_view(u"woman tipping hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B"),
            std::u16string_view(u"person_raising_hand"),
            std::u16string_view(u"person raising hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FB"),
            std::u16string_view(u"person_raising_hand_light_skin_tone"),
            std::u16string_view(u"person raising hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FC"),
            std::u16string_view(u"person_raising_hand_medium_light_skin_tone"),
            std::u16string_view(u"person raising hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FD"),
            std::u16string_view(u"person_raising_hand_medium_skin_tone"),
            std::u16string_view(u"person raising hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FE"),
            std::u16string_view(u"person_raising_hand_medium_dark_skin_tone"),
            std::u16string_view(u"person raising hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FF"),
            std::u16string_view(u"person_raising_hand_dark_skin_tone"),
            std::u16string_view(u"person raising hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand"),
            std::u16string_view(u"man raising hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand_light_skin_tone"),
            std::u16string_view(u"man raising hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand_medium_light_skin_tone"),
            std::u16string_view(u"man raising hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand_medium_skin_tone"),
            std::u16string_view(u"man raising hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand_medium_dark_skin_tone"),
            std::u16string_view(u"man raising hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_raising_hand_dark_skin_tone"),
            std::u16string_view(u"man raising hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand"),
            std::u16string_view(u"woman raising hand"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand_light_skin_tone"),
            std::u16string_view(u"woman raising hand: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand_medium_light_skin_tone"),
            std::u16string_view(u"woman raising hand: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand_medium_skin_tone"),
            std::u16string_view(u"woman raising hand: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand_medium_dark_skin_tone"),
            std::u16string_view(u"woman raising hand: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F64B\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_raising_hand_dark_skin_tone"),
            std::u16string_view(u"woman raising hand: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF"),
            std::u16string_view(u"deaf_person"),
            std::u16string_view(u"deaf person"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FB"),
            std::u16string_view(u"deaf_person_light_skin_tone"),
            std::u16string_view(u"deaf person: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FC"),
            std::u16string_view(u"deaf_person_medium_light_skin_tone"),
            std::u16string_view(u"deaf person: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FD"),
            std::u16string_view(u"deaf_person_medium_skin_tone"),
            std::u16string_view(u"deaf person: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FE"),
            std::u16string_view(u"deaf_person_medium_dark_skin_tone"),
            std::u16string_view(u"deaf person: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FF"),
            std::u16string_view(u"deaf_person_dark_skin_tone"),
            std::u16string_view(u"deaf person: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man"),
            std::u16string_view(u"deaf man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man_light_skin_tone"),
            std::u16string_view(u"deaf man: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man_medium_light_skin_tone"),
            std::u16string_view(u"deaf man: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man_medium_skin_tone"),
            std::u16string_view(u"deaf man: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man_medium_dark_skin_tone"),
            std::u16string_view(u"deaf man: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"deaf_man_dark_skin_tone"),
            std::u16string_view(u"deaf man: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman"),
            std::u16string_view(u"deaf woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman_light_skin_tone"),
            std::u16string_view(u"deaf woman: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman_medium_light_skin_tone"),
            std::u16string_view(u"deaf woman: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman_medium_skin_tone"),
            std::u16string_view(u"deaf woman: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman_medium_dark_skin_tone"),
            std::u16string_view(u"deaf woman: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CF\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"deaf_woman_dark_skin_tone"),
            std::u16string_view(u"deaf woman: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647"),
            std::u16string_view(u"person_bowing"),
            std::u16string_view(u"person bowing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FB"),
            std::u16string_view(u"person_bowing_light_skin_tone"),
            std::u16string_view(u"person bowing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FC"),
            std::u16string_view(u"person_bowing_medium_light_skin_tone"),
            std::u16string_view(u"person bowing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FD"),
            std::u16string_view(u"person_bowing_medium_skin_tone"),
            std::u16string_view(u"person bowing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FE"),
            std::u16string_view(u"person_bowing_medium_dark_skin_tone"),
            std::u16string_view(u"person bowing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FF"),
            std::u16string_view(u"person_bowing_dark_skin_tone"),
            std::u16string_view(u"person bowing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing"),
            std::u16string_view(u"man bowing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing_light_skin_tone"),
            std::u16string_view(u"man bowing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing_medium_light_skin_tone"),
            std::u16string_view(u"man bowing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing_medium_skin_tone"),
            std::u16string_view(u"man bowing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing_medium_dark_skin_tone"),
            std::u16string_view(u"man bowing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bowing_dark_skin_tone"),
            std::u16string_view(u"man bowing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing"),
            std::u16string_view(u"woman bowing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing_light_skin_tone"),
            std::u16string_view(u"woman bowing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing_medium_light_skin_tone"),
            std::u16string_view(u"woman bowing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing_medium_skin_tone"),
            std::u16string_view(u"woman bowing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing_medium_dark_skin_tone"),
            std::u16string_view(u"woman bowing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F647\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bowing_dark_skin_tone"),
            std::u16string_view(u"woman bowing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926"),
            std::u16string_view(u"person_facepalming"),
            std::u16string_view(u"person facepalming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FB"),
            std::u16string_view(u"person_facepalming_light_skin_tone"),
            std::u16string_view(u"person facepalming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FC"),
            std::u16string_view(u"person_facepalming_medium_light_skin_tone"),
            std::u16string_view(u"person facepalming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FD"),
            std::u16string_view(u"person_facepalming_medium_skin_tone"),
            std::u16string_view(u"person facepalming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FE"),
            std::u16string_view(u"person_facepalming_medium_dark_skin_tone"),
            std::u16string_view(u"person facepalming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FF"),
            std::u16string_view(u"person_facepalming_dark_skin_tone"),
            std::u16string_view(u"person facepalming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming"),
            std::u16string_view(u"man facepalming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming_light_skin_tone"),
            std::u16string_view(u"man facepalming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming_medium_light_skin_tone"),
            std::u16string_view(u"man facepalming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming_medium_skin_tone"),
            std::u16string_view(u"man facepalming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming_medium_dark_skin_tone"),
            std::u16string_view(u"man facepalming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_facepalming_dark_skin_tone"),
            std::u16string_view(u"man facepalming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming"),
            std::u16string_view(u"woman facepalming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming_light_skin_tone"),
            std::u16string_view(u"woman facepalming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming_medium_light_skin_tone"),
            std::u16string_view(u"woman facepalming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming_medium_skin_tone"),
            std::u16string_view(u"woman facepalming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming_medium_dark_skin_tone"),
            std::u16string_view(u"woman facepalming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F926\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_facepalming_dark_skin_tone"),
            std::u16string_view(u"woman facepalming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937"),
            std::u16string_view(u"shrug"),
            std::u16string_view(u"person shrugging"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FB"),
            std::u16string_view(u"person_shrugging_light_skin_tone"),
            std::u16string_view(u"person shrugging: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FC"),
            std::u16string_view(u"person_shrugging_medium_light_skin_tone"),
            std::u16string_view(u"person shrugging: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FD"),
            std::u16string_view(u"person_shrugging_medium_skin_tone"),
            std::u16string_view(u"person shrugging: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FE"),
            std::u16string_view(u"person_shrugging_medium_dark_skin_tone"),
            std::u16string_view(u"person shrugging: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FF"),
            std::u16string_view(u"person_shrugging_dark_skin_tone"),
            std::u16string_view(u"person shrugging: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging"),
            std::u16string_view(u"man shrugging"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging_light_skin_tone"),
            std::u16string_view(u"man shrugging: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging_medium_light_skin_tone"),
            std::u16string_view(u"man shrugging: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging_medium_skin_tone"),
            std::u16string_view(u"man shrugging: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging_medium_dark_skin_tone"),
            std::u16string_view(u"man shrugging: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_shrugging_dark_skin_tone"),
            std::u16string_view(u"man shrugging: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging"),
            std::u16string_view(u"woman shrugging"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging_light_skin_tone"),
            std::u16string_view(u"woman shrugging: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging_medium_light_skin_tone"),
            std::u16string_view(u"woman shrugging: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging_medium_skin_tone"),
            std::u16string_view(u"woman shrugging: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging_medium_dark_skin_tone"),
            std::u16string_view(u"woman shrugging: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F937\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_shrugging_dark_skin_tone"),
            std::u16string_view(u"woman shrugging: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker"),
            std::u16string_view(u"health worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker_light_skin_tone"),
            std::u16string_view(u"health worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker_medium_light_skin_tone"),
            std::u16string_view(u"health worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker_medium_skin_tone"),
            std::u16string_view(u"health worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker_medium_dark_skin_tone"),
            std::u16string_view(u"health worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"health_worker_dark_skin_tone"),
            std::u16string_view(u"health worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker"),
            std::u16string_view(u"man health worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker_light_skin_tone"),
            std::u16string_view(u"man health worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker_medium_light_skin_tone"),
            std::u16string_view(u"man health worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker_medium_skin_tone"),
            std::u16string_view(u"man health worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker_medium_dark_skin_tone"),
            std::u16string_view(u"man health worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"man_health_worker_dark_skin_tone"),
            std::u16string_view(u"man health worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker"),
            std::u16string_view(u"woman health worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker_light_skin_tone"),
            std::u16string_view(u"woman health worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker_medium_light_skin_tone"),
            std::u16string_view(u"woman health worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker_medium_skin_tone"),
            std::u16string_view(u"woman health worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker_medium_dark_skin_tone"),
            std::u16string_view(u"woman health worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002695\U0000FE0F"),
            std::u16string_view(u"woman_health_worker_dark_skin_tone"),
            std::u16string_view(u"woman health worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F393"),
            std::u16string_view(u"student"),
            std::u16string_view(u"student"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F393"),
            std::u16string_view(u"student_light_skin_tone"),
            std::u16string_view(u"student: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F393"),
            std::u16string_view(u"student_medium_light_skin_tone"),
            std::u16string_view(u"student: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F393"),
            std::u16string_view(u"student_medium_skin_tone"),
            std::u16string_view(u"student: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F393"),
            std::u16string_view(u"student_medium_dark_skin_tone"),
            std::u16string_view(u"student: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F393"),
            std::u16string_view(u"student_dark_skin_tone"),
            std::u16string_view(u"student: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F393"),
            std::u16string_view(u"man_student"),
            std::u16string_view(u"man student"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F393"),
            std::u16string_view(u"man_student_light_skin_tone"),
            std::u16string_view(u"man student: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F393"),
            std::u16string_view(u"man_student_medium_light_skin_tone"),
            std::u16string_view(u"man student: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F393"),
            std::u16string_view(u"man_student_medium_skin_tone"),
            std::u16string_view(u"man student: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F393"),
            std::u16string_view(u"man_student_medium_dark_skin_tone"),
            std::u16string_view(u"man student: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F393"),
            std::u16string_view(u"man_student_dark_skin_tone"),
            std::u16string_view(u"man student: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student"),
            std::u16string_view(u"woman student"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student_light_skin_tone"),
            std::u16string_view(u"woman student: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student_medium_light_skin_tone"),
            std::u16string_view(u"woman student: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student_medium_skin_tone"),
            std::u16string_view(u"woman student: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student_medium_dark_skin_tone"),
            std::u16string_view(u"woman student: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F393"),
            std::u16string_view(u"woman_student_dark_skin_tone"),
            std::u16string_view(u"woman student: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher"),
            std::u16string_view(u"teacher"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher_light_skin_tone"),
            std::u16string_view(u"teacher: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher_medium_light_skin_tone"),
            std::u16string_view(u"teacher: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher_medium_skin_tone"),
            std::u16string_view(u"teacher: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher_medium_dark_skin_tone"),
            std::u16string_view(u"teacher: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F3EB"),
            std::u16string_view(u"teacher_dark_skin_tone"),
            std::u16string_view(u"teacher: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher"),
            std::u16string_view(u"man teacher"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher_light_skin_tone"),
            std::u16string_view(u"man teacher: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher_medium_light_skin_tone"),
            std::u16string_view(u"man teacher: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher_medium_skin_tone"),
            std::u16string_view(u"man teacher: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher_medium_dark_skin_tone"),
            std::u16string_view(u"man teacher: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F3EB"),
            std::u16string_view(u"man_teacher_dark_skin_tone"),
            std::u16string_view(u"man teacher: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher"),
            std::u16string_view(u"woman teacher"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher_light_skin_tone"),
            std::u16string_view(u"woman teacher: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher_medium_light_skin_tone"),
            std::u16string_view(u"woman teacher: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher_medium_skin_tone"),
            std::u16string_view(u"woman teacher: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher_medium_dark_skin_tone"),
            std::u16string_view(u"woman teacher: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F3EB"),
            std::u16string_view(u"woman_teacher_dark_skin_tone"),
            std::u16string_view(u"woman teacher: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge"),
            std::u16string_view(u"judge"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge_light_skin_tone"),
            std::u16string_view(u"judge: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge_medium_light_skin_tone"),
            std::u16string_view(u"judge: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge_medium_skin_tone"),
            std::u16string_view(u"judge: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge_medium_dark_skin_tone"),
            std::u16string_view(u"judge: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"judge_dark_skin_tone"),
            std::u16string_view(u"judge: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge"),
            std::u16string_view(u"man judge"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge_light_skin_tone"),
            std::u16string_view(u"man judge: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge_medium_light_skin_tone"),
            std::u16string_view(u"man judge: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge_medium_skin_tone"),
            std::u16string_view(u"man judge: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge_medium_dark_skin_tone"),
            std::u16string_view(u"man judge: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"man_judge_dark_skin_tone"),
            std::u16string_view(u"man judge: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge"),
            std::u16string_view(u"woman judge"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge_light_skin_tone"),
            std::u16string_view(u"woman judge: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge_medium_light_skin_tone"),
            std::u16string_view(u"woman judge: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge_medium_skin_tone"),
            std::u16string_view(u"woman judge: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge_medium_dark_skin_tone"),
            std::u16string_view(u"woman judge: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002696\U0000FE0F"),
            std::u16string_view(u"woman_judge_dark_skin_tone"),
            std::u16string_view(u"woman judge: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer"),
            std::u16string_view(u"farmer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer_light_skin_tone"),
            std::u16string_view(u"farmer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer_medium_light_skin_tone"),
            std::u16string_view(u"farmer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer_medium_skin_tone"),
            std::u16string_view(u"farmer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer_medium_dark_skin_tone"),
            std::u16string_view(u"farmer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F33E"),
            std::u16string_view(u"farmer_dark_skin_tone"),
            std::u16string_view(u"farmer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer"),
            std::u16string_view(u"man farmer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer_light_skin_tone"),
            std::u16string_view(u"man farmer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer_medium_light_skin_tone"),
            std::u16string_view(u"man farmer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer_medium_skin_tone"),
            std::u16string_view(u"man farmer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer_medium_dark_skin_tone"),
            std::u16string_view(u"man farmer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F33E"),
            std::u16string_view(u"man_farmer_dark_skin_tone"),
            std::u16string_view(u"man farmer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer"),
            std::u16string_view(u"woman farmer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer_light_skin_tone"),
            std::u16string_view(u"woman farmer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer_medium_light_skin_tone"),
            std::u16string_view(u"woman farmer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer_medium_skin_tone"),
            std::u16string_view(u"woman farmer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer_medium_dark_skin_tone"),
            std::u16string_view(u"woman farmer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F33E"),
            std::u16string_view(u"woman_farmer_dark_skin_tone"),
            std::u16string_view(u"woman farmer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F373"),
            std::u16string_view(u"cook"),
            std::u16string_view(u"cook"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F373"),
            std::u16string_view(u"cook_light_skin_tone"),
            std::u16string_view(u"cook: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F373"),
            std::u16string_view(u"cook_medium_light_skin_tone"),
            std::u16string_view(u"cook: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F373"),
            std::u16string_view(u"cook_medium_skin_tone"),
            std::u16string_view(u"cook: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F373"),
            std::u16string_view(u"cook_medium_dark_skin_tone"),
            std::u16string_view(u"cook: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F373"),
            std::u16string_view(u"cook_dark_skin_tone"),
            std::u16string_view(u"cook: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook"),
            std::u16string_view(u"man cook"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook_light_skin_tone"),
            std::u16string_view(u"man cook: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook_medium_light_skin_tone"),
            std::u16string_view(u"man cook: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook_medium_skin_tone"),
            std::u16string_view(u"man cook: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook_medium_dark_skin_tone"),
            std::u16string_view(u"man cook: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F373"),
            std::u16string_view(u"man_cook_dark_skin_tone"),
            std::u16string_view(u"man cook: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook"),
            std::u16string_view(u"woman cook"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook_light_skin_tone"),
            std::u16string_view(u"woman cook: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook_medium_light_skin_tone"),
            std::u16string_view(u"woman cook: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook_medium_skin_tone"),
            std::u16string_view(u"woman cook: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook_medium_dark_skin_tone"),
            std::u16string_view(u"woman cook: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F373"),
            std::u16string_view(u"woman_cook_dark_skin_tone"),
            std::u16string_view(u"woman cook: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic"),
            std::u16string_view(u"mechanic"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic_light_skin_tone"),
            std::u16string_view(u"mechanic: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic_medium_light_skin_tone"),
            std::u16string_view(u"mechanic: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic_medium_skin_tone"),
            std::u16string_view(u"mechanic: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic_medium_dark_skin_tone"),
            std::u16string_view(u"mechanic: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F527"),
            std::u16string_view(u"mechanic_dark_skin_tone"),
            std::u16string_view(u"mechanic: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic"),
            std::u16string_view(u"man mechanic"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic_light_skin_tone"),
            std::u16string_view(u"man mechanic: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic_medium_light_skin_tone"),
            std::u16string_view(u"man mechanic: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic_medium_skin_tone"),
            std::u16string_view(u"man mechanic: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic_medium_dark_skin_tone"),
            std::u16string_view(u"man mechanic: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F527"),
            std::u16string_view(u"man_mechanic_dark_skin_tone"),
            std::u16string_view(u"man mechanic: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic"),
            std::u16string_view(u"woman mechanic"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic_light_skin_tone"),
            std::u16string_view(u"woman mechanic: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic_medium_light_skin_tone"),
            std::u16string_view(u"woman mechanic: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic_medium_skin_tone"),
            std::u16string_view(u"woman mechanic: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic_medium_dark_skin_tone"),
            std::u16string_view(u"woman mechanic: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F527"),
            std::u16string_view(u"woman_mechanic_dark_skin_tone"),
            std::u16string_view(u"woman mechanic: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker"),
            std::u16string_view(u"factory worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker_light_skin_tone"),
            std::u16string_view(u"factory worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker_medium_light_skin_tone"),
            std::u16string_view(u"factory worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker_medium_skin_tone"),
            std::u16string_view(u"factory worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker_medium_dark_skin_tone"),
            std::u16string_view(u"factory worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F3ED"),
            std::u16string_view(u"factory_worker_dark_skin_tone"),
            std::u16string_view(u"factory worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker"),
            std::u16string_view(u"man factory worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker_light_skin_tone"),
            std::u16string_view(u"man factory worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker_medium_light_skin_tone"),
            std::u16string_view(u"man factory worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker_medium_skin_tone"),
            std::u16string_view(u"man factory worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker_medium_dark_skin_tone"),
            std::u16string_view(u"man factory worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F3ED"),
            std::u16string_view(u"man_factory_worker_dark_skin_tone"),
            std::u16string_view(u"man factory worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker"),
            std::u16string_view(u"woman factory worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker_light_skin_tone"),
            std::u16string_view(u"woman factory worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker_medium_light_skin_tone"),
            std::u16string_view(u"woman factory worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker_medium_skin_tone"),
            std::u16string_view(u"woman factory worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker_medium_dark_skin_tone"),
            std::u16string_view(u"woman factory worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F3ED"),
            std::u16string_view(u"woman_factory_worker_dark_skin_tone"),
            std::u16string_view(u"woman factory worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker"),
            std::u16string_view(u"office worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker_light_skin_tone"),
            std::u16string_view(u"office worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker_medium_light_skin_tone"),
            std::u16string_view(u"office worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker_medium_skin_tone"),
            std::u16string_view(u"office worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker_medium_dark_skin_tone"),
            std::u16string_view(u"office worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F4BC"),
            std::u16string_view(u"office_worker_dark_skin_tone"),
            std::u16string_view(u"office worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker"),
            std::u16string_view(u"man office worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker_light_skin_tone"),
            std::u16string_view(u"man office worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker_medium_light_skin_tone"),
            std::u16string_view(u"man office worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker_medium_skin_tone"),
            std::u16string_view(u"man office worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker_medium_dark_skin_tone"),
            std::u16string_view(u"man office worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F4BC"),
            std::u16string_view(u"man_office_worker_dark_skin_tone"),
            std::u16string_view(u"man office worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker"),
            std::u16string_view(u"woman office worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker_light_skin_tone"),
            std::u16string_view(u"woman office worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker_medium_light_skin_tone"),
            std::u16string_view(u"woman office worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker_medium_skin_tone"),
            std::u16string_view(u"woman office worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker_medium_dark_skin_tone"),
            std::u16string_view(u"woman office worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F4BC"),
            std::u16string_view(u"woman_office_worker_dark_skin_tone"),
            std::u16string_view(u"woman office worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist"),
            std::u16string_view(u"scientist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist_light_skin_tone"),
            std::u16string_view(u"scientist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist_medium_light_skin_tone"),
            std::u16string_view(u"scientist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist_medium_skin_tone"),
            std::u16string_view(u"scientist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist_medium_dark_skin_tone"),
            std::u16string_view(u"scientist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F52C"),
            std::u16string_view(u"scientist_dark_skin_tone"),
            std::u16string_view(u"scientist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist"),
            std::u16string_view(u"man scientist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist_light_skin_tone"),
            std::u16string_view(u"man scientist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist_medium_light_skin_tone"),
            std::u16string_view(u"man scientist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist_medium_skin_tone"),
            std::u16string_view(u"man scientist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist_medium_dark_skin_tone"),
            std::u16string_view(u"man scientist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F52C"),
            std::u16string_view(u"man_scientist_dark_skin_tone"),
            std::u16string_view(u"man scientist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist"),
            std::u16string_view(u"woman scientist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist_light_skin_tone"),
            std::u16string_view(u"woman scientist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist_medium_light_skin_tone"),
            std::u16string_view(u"woman scientist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist_medium_skin_tone"),
            std::u16string_view(u"woman scientist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist_medium_dark_skin_tone"),
            std::u16string_view(u"woman scientist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F52C"),
            std::u16string_view(u"woman_scientist_dark_skin_tone"),
            std::u16string_view(u"woman scientist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist"),
            std::u16string_view(u"technologist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist_light_skin_tone"),
            std::u16string_view(u"technologist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist_medium_light_skin_tone"),
            std::u16string_view(u"technologist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist_medium_skin_tone"),
            std::u16string_view(u"technologist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist_medium_dark_skin_tone"),
            std::u16string_view(u"technologist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F4BB"),
            std::u16string_view(u"technologist_dark_skin_tone"),
            std::u16string_view(u"technologist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist"),
            std::u16string_view(u"man technologist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist_light_skin_tone"),
            std::u16string_view(u"man technologist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist_medium_light_skin_tone"),
            std::u16string_view(u"man technologist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist_medium_skin_tone"),
            std::u16string_view(u"man technologist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist_medium_dark_skin_tone"),
            std::u16string_view(u"man technologist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F4BB"),
            std::u16string_view(u"man_technologist_dark_skin_tone"),
            std::u16string_view(u"man technologist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist"),
            std::u16string_view(u"woman technologist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist_light_skin_tone"),
            std::u16string_view(u"woman technologist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist_medium_light_skin_tone"),
            std::u16string_view(u"woman technologist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist_medium_skin_tone"),
            std::u16string_view(u"woman technologist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist_medium_dark_skin_tone"),
            std::u16string_view(u"woman technologist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F4BB"),
            std::u16string_view(u"woman_technologist_dark_skin_tone"),
            std::u16string_view(u"woman technologist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer"),
            std::u16string_view(u"singer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer_light_skin_tone"),
            std::u16string_view(u"singer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer_medium_light_skin_tone"),
            std::u16string_view(u"singer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer_medium_skin_tone"),
            std::u16string_view(u"singer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer_medium_dark_skin_tone"),
            std::u16string_view(u"singer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F3A4"),
            std::u16string_view(u"singer_dark_skin_tone"),
            std::u16string_view(u"singer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer"),
            std::u16string_view(u"man singer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer_light_skin_tone"),
            std::u16string_view(u"man singer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer_medium_light_skin_tone"),
            std::u16string_view(u"man singer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer_medium_skin_tone"),
            std::u16string_view(u"man singer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer_medium_dark_skin_tone"),
            std::u16string_view(u"man singer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F3A4"),
            std::u16string_view(u"man_singer_dark_skin_tone"),
            std::u16string_view(u"man singer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer"),
            std::u16string_view(u"woman singer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer_light_skin_tone"),
            std::u16string_view(u"woman singer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer_medium_light_skin_tone"),
            std::u16string_view(u"woman singer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer_medium_skin_tone"),
            std::u16string_view(u"woman singer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer_medium_dark_skin_tone"),
            std::u16string_view(u"woman singer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F3A4"),
            std::u16string_view(u"woman_singer_dark_skin_tone"),
            std::u16string_view(u"woman singer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist"),
            std::u16string_view(u"artist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist_light_skin_tone"),
            std::u16string_view(u"artist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist_medium_light_skin_tone"),
            std::u16string_view(u"artist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist_medium_skin_tone"),
            std::u16string_view(u"artist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist_medium_dark_skin_tone"),
            std::u16string_view(u"artist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F3A8"),
            std::u16string_view(u"artist_dark_skin_tone"),
            std::u16string_view(u"artist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist"),
            std::u16string_view(u"man artist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist_light_skin_tone"),
            std::u16string_view(u"man artist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist_medium_light_skin_tone"),
            std::u16string_view(u"man artist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist_medium_skin_tone"),
            std::u16string_view(u"man artist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist_medium_dark_skin_tone"),
            std::u16string_view(u"man artist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F3A8"),
            std::u16string_view(u"man_artist_dark_skin_tone"),
            std::u16string_view(u"man artist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist"),
            std::u16string_view(u"woman artist"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist_light_skin_tone"),
            std::u16string_view(u"woman artist: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist_medium_light_skin_tone"),
            std::u16string_view(u"woman artist: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist_medium_skin_tone"),
            std::u16string_view(u"woman artist: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist_medium_dark_skin_tone"),
            std::u16string_view(u"woman artist: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F3A8"),
            std::u16string_view(u"woman_artist_dark_skin_tone"),
            std::u16string_view(u"woman artist: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot"),
            std::u16string_view(u"pilot"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot_light_skin_tone"),
            std::u16string_view(u"pilot: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot_medium_light_skin_tone"),
            std::u16string_view(u"pilot: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot_medium_skin_tone"),
            std::u16string_view(u"pilot: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot_medium_dark_skin_tone"),
            std::u16string_view(u"pilot: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"pilot_dark_skin_tone"),
            std::u16string_view(u"pilot: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot"),
            std::u16string_view(u"man pilot"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot_light_skin_tone"),
            std::u16string_view(u"man pilot: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot_medium_light_skin_tone"),
            std::u16string_view(u"man pilot: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot_medium_skin_tone"),
            std::u16string_view(u"man pilot: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot_medium_dark_skin_tone"),
            std::u16string_view(u"man pilot: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"man_pilot_dark_skin_tone"),
            std::u16string_view(u"man pilot: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot"),
            std::u16string_view(u"woman pilot"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot_light_skin_tone"),
            std::u16string_view(u"woman pilot: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot_medium_light_skin_tone"),
            std::u16string_view(u"woman pilot: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot_medium_skin_tone"),
            std::u16string_view(u"woman pilot: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot_medium_dark_skin_tone"),
            std::u16string_view(u"woman pilot: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002708\U0000FE0F"),
            std::u16string_view(u"woman_pilot_dark_skin_tone"),
            std::u16string_view(u"woman pilot: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut"),
            std::u16string_view(u"astronaut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut_light_skin_tone"),
            std::u16string_view(u"astronaut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut_medium_light_skin_tone"),
            std::u16string_view(u"astronaut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut_medium_skin_tone"),
            std::u16string_view(u"astronaut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut_medium_dark_skin_tone"),
            std::u16string_view(u"astronaut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F680"),
            std::u16string_view(u"astronaut_dark_skin_tone"),
            std::u16string_view(u"astronaut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut"),
            std::u16string_view(u"man astronaut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut_light_skin_tone"),
            std::u16string_view(u"man astronaut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut_medium_light_skin_tone"),
            std::u16string_view(u"man astronaut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut_medium_skin_tone"),
            std::u16string_view(u"man astronaut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut_medium_dark_skin_tone"),
            std::u16string_view(u"man astronaut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F680"),
            std::u16string_view(u"man_astronaut_dark_skin_tone"),
            std::u16string_view(u"man astronaut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut"),
            std::u16string_view(u"woman astronaut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut_light_skin_tone"),
            std::u16string_view(u"woman astronaut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut_medium_light_skin_tone"),
            std::u16string_view(u"woman astronaut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut_medium_skin_tone"),
            std::u16string_view(u"woman astronaut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut_medium_dark_skin_tone"),
            std::u16string_view(u"woman astronaut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F680"),
            std::u16string_view(u"woman_astronaut_dark_skin_tone"),
            std::u16string_view(u"woman astronaut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter"),
            std::u16string_view(u"firefighter"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter_light_skin_tone"),
            std::u16string_view(u"firefighter: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter_medium_light_skin_tone"),
            std::u16string_view(u"firefighter: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter_medium_skin_tone"),
            std::u16string_view(u"firefighter: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter_medium_dark_skin_tone"),
            std::u16string_view(u"firefighter: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F692"),
            std::u16string_view(u"firefighter_dark_skin_tone"),
            std::u16string_view(u"firefighter: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter"),
            std::u16string_view(u"man firefighter"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter_light_skin_tone"),
            std::u16string_view(u"man firefighter: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter_medium_light_skin_tone"),
            std::u16string_view(u"man firefighter: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter_medium_skin_tone"),
            std::u16string_view(u"man firefighter: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter_medium_dark_skin_tone"),
            std::u16string_view(u"man firefighter: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F692"),
            std::u16string_view(u"man_firefighter_dark_skin_tone"),
            std::u16string_view(u"man firefighter: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter"),
            std::u16string_view(u"woman firefighter"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter_light_skin_tone"),
            std::u16string_view(u"woman firefighter: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter_medium_light_skin_tone"),
            std::u16string_view(u"woman firefighter: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter_medium_skin_tone"),
            std::u16string_view(u"woman firefighter: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter_medium_dark_skin_tone"),
            std::u16string_view(u"woman firefighter: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F692"),
            std::u16string_view(u"woman_firefighter_dark_skin_tone"),
            std::u16string_view(u"woman firefighter: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E"),
            std::u16string_view(u"police_officer"),
            std::u16string_view(u"police officer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FB"),
            std::u16string_view(u"police_officer_light_skin_tone"),
            std::u16string_view(u"police officer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FC"),
            std::u16string_view(u"police_officer_medium_light_skin_tone"),
            std::u16string_view(u"police officer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FD"),
            std::u16string_view(u"police_officer_medium_skin_tone"),
            std::u16string_view(u"police officer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FE"),
            std::u16string_view(u"police_officer_medium_dark_skin_tone"),
            std::u16string_view(u"police officer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FF"),
            std::u16string_view(u"police_officer_dark_skin_tone"),
            std::u16string_view(u"police officer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer"),
            std::u16string_view(u"man police officer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer_light_skin_tone"),
            std::u16string_view(u"man police officer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer_medium_light_skin_tone"),
            std::u16string_view(u"man police officer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer_medium_skin_tone"),
            std::u16string_view(u"man police officer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer_medium_dark_skin_tone"),
            std::u16string_view(u"man police officer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_police_officer_dark_skin_tone"),
            std::u16string_view(u"man police officer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer"),
            std::u16string_view(u"woman police officer"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer_light_skin_tone"),
            std::u16string_view(u"woman police officer: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer_medium_light_skin_tone"),
            std::u16string_view(u"woman police officer: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer_medium_skin_tone"),
            std::u16string_view(u"woman police officer: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer_medium_dark_skin_tone"),
            std::u16string_view(u"woman police officer: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46E\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_police_officer_dark_skin_tone"),
            std::u16string_view(u"woman police officer: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0000FE0F"),
            std::u16string_view(u"detective"),
            std::u16string_view(u"detective"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FB"),
            std::u16string_view(u"detective_light_skin_tone"),
            std::u16string_view(u"detective: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FC"),
            std::u16string_view(u"detective_medium_light_skin_tone"),
            std::u16string_view(u"detective: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FD"),
            std::u16string_view(u"detective_medium_skin_tone"),
            std::u16string_view(u"detective: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FE"),
            std::u16string_view(u"detective_medium_dark_skin_tone"),
            std::u16string_view(u"detective: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FF"),
            std::u16string_view(u"detective_dark_skin_tone"),
            std::u16string_view(u"detective: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0000FE0F\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective"),
            std::u16string_view(u"man detective"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective_light_skin_tone"),
            std::u16string_view(u"man detective: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective_medium_light_skin_tone"),
            std::u16string_view(u"man detective: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective_medium_skin_tone"),
            std::u16string_view(u"man detective: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective_medium_dark_skin_tone"),
            std::u16string_view(u"man detective: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_detective_dark_skin_tone"),
            std::u16string_view(u"man detective: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0000FE0F\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective"),
            std::u16string_view(u"woman detective"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective_light_skin_tone"),
            std::u16string_view(u"woman detective: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective_medium_light_skin_tone"),
            std::u16string_view(u"woman detective: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective_medium_skin_tone"),
            std::u16string_view(u"woman detective: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective_medium_dark_skin_tone"),
            std::u16string_view(u"woman detective: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F575\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_detective_dark_skin_tone"),
            std::u16string_view(u"woman detective: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482"),
            std::u16string_view(u"guard"),
            std::u16string_view(u"guard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FB"),
            std::u16string_view(u"guard_light_skin_tone"),
            std::u16string_view(u"guard: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FC"),
            std::u16string_view(u"guard_medium_light_skin_tone"),
            std::u16string_view(u"guard: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FD"),
            std::u16string_view(u"guard_medium_skin_tone"),
            std::u16string_view(u"guard: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FE"),
            std::u16string_view(u"guard_medium_dark_skin_tone"),
            std::u16string_view(u"guard: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FF"),
            std::u16string_view(u"guard_dark_skin_tone"),
            std::u16string_view(u"guard: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard"),
            std::u16string_view(u"man guard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard_light_skin_tone"),
            std::u16string_view(u"man guard: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard_medium_light_skin_tone"),
            std::u16string_view(u"man guard: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard_medium_skin_tone"),
            std::u16string_view(u"man guard: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard_medium_dark_skin_tone"),
            std::u16string_view(u"man guard: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_guard_dark_skin_tone"),
            std::u16string_view(u"man guard: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard"),
            std::u16string_view(u"woman guard"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard_light_skin_tone"),
            std::u16string_view(u"woman guard: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard_medium_light_skin_tone"),
            std::u16string_view(u"woman guard: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard_medium_skin_tone"),
            std::u16string_view(u"woman guard: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard_medium_dark_skin_tone"),
            std::u16string_view(u"woman guard: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F482\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_guard_dark_skin_tone"),
            std::u16string_view(u"woman guard: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977"),
            std::u16string_view(u"ninja"),
            std::u16string_view(u"ninja"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977\U0001F3FB"),
            std::u16string_view(u"ninja_light_skin_tone"),
            std::u16string_view(u"ninja: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977\U0001F3FC"),
            std::u16string_view(u"ninja_medium_light_skin_tone"),
            std::u16string_view(u"ninja: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977\U0001F3FD"),
            std::u16string_view(u"ninja_medium_skin_tone"),
            std::u16string_view(u"ninja: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977\U0001F3FE"),
            std::u16string_view(u"ninja_medium_dark_skin_tone"),
            std::u16string_view(u"ninja: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F977\U0001F3FF"),
            std::u16string_view(u"ninja_dark_skin_tone"),
            std::u16string_view(u"ninja: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477"),
            std::u16string_view(u"construction_worker"),
            std::u16string_view(u"construction worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FB"),
            std::u16string_view(u"construction_worker_light_skin_tone"),
            std::u16string_view(u"construction worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FC"),
            std::u16string_view(u"construction_worker_medium_light_skin_tone"),
            std::u16string_view(u"construction worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FD"),
            std::u16string_view(u"construction_worker_medium_skin_tone"),
            std::u16string_view(u"construction worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FE"),
            std::u16string_view(u"construction_worker_medium_dark_skin_tone"),
            std::u16string_view(u"construction worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FF"),
            std::u16string_view(u"construction_worker_dark_skin_tone"),
            std::u16string_view(u"construction worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker"),
            std::u16string_view(u"man construction worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker_light_skin_tone"),
            std::u16string_view(u"man construction worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker_medium_light_skin_tone"),
            std::u16string_view(u"man construction worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker_medium_skin_tone"),
            std::u16string_view(u"man construction worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker_medium_dark_skin_tone"),
            std::u16string_view(u"man construction worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_construction_worker_dark_skin_tone"),
            std::u16string_view(u"man construction worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker"),
            std::u16string_view(u"woman construction worker"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker_light_skin_tone"),
            std::u16string_view(u"woman construction worker: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker_medium_light_skin_tone"),
            std::u16string_view(u"woman construction worker: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker_medium_skin_tone"),
            std::u16string_view(u"woman construction worker: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker_medium_dark_skin_tone"),
            std::u16string_view(u"woman construction worker: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F477\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_construction_worker_dark_skin_tone"),
            std::u16string_view(u"woman construction worker: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5"),
            std::u16string_view(u"person_with_crown"),
            std::u16string_view(u"person with crown"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5\U0001F3FB"),
            std::u16string_view(u"person_with_crown_light_skin_tone"),
            std::u16string_view(u"person with crown: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5\U0001F3FC"),
            std::u16string_view(u"person_with_crown_medium_light_skin_tone"),
            std::u16string_view(u"person with crown: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5\U0001F3FD"),
            std::u16string_view(u"person_with_crown_medium_skin_tone"),
            std::u16string_view(u"person with crown: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5\U0001F3FE"),
            std::u16string_view(u"person_with_crown_medium_dark_skin_tone"),
            std::u16string_view(u"person with crown: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC5\U0001F3FF"),
            std::u16string_view(u"person_with_crown_dark_skin_tone"),
            std::u16string_view(u"person with crown: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934"),
            std::u16string_view(u"prince"),
            std::u16string_view(u"prince"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934\U0001F3FB"),
            std::u16string_view(u"prince_light_skin_tone"),
            std::u16string_view(u"prince: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934\U0001F3FC"),
            std::u16string_view(u"prince_medium_light_skin_tone"),
            std::u16string_view(u"prince: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934\U0001F3FD"),
            std::u16string_view(u"prince_medium_skin_tone"),
            std::u16string_view(u"prince: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934\U0001F3FE"),
            std::u16string_view(u"prince_medium_dark_skin_tone"),
            std::u16string_view(u"prince: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F934\U0001F3FF"),
            std::u16string_view(u"prince_dark_skin_tone"),
            std::u16string_view(u"prince: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478"),
            std::u16string_view(u"princess"),
            std::u16string_view(u"princess"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478\U0001F3FB"),
            std::u16string_view(u"princess_light_skin_tone"),
            std::u16string_view(u"princess: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478\U0001F3FC"),
            std::u16string_view(u"princess_medium_light_skin_tone"),
            std::u16string_view(u"princess: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478\U0001F3FD"),
            std::u16string_view(u"princess_medium_skin_tone"),
            std::u16string_view(u"princess: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478\U0001F3FE"),
            std::u16string_view(u"princess_medium_dark_skin_tone"),
            std::u16string_view(u"princess: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F478\U0001F3FF"),
            std::u16string_view(u"princess_dark_skin_tone"),
            std::u16string_view(u"princess: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473"),
            std::u16string_view(u"person_wearing_turban"),
            std::u16string_view(u"person wearing turban"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FB"),
            std::u16string_view(u"person_wearing_turban_light_skin_tone"),
            std::u16string_view(u"person wearing turban: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FC"),
            std::u16string_view(u"person_wearing_turban_medium_light_skin_tone"),
            std::u16string_view(u"person wearing turban: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FD"),
            std::u16string_view(u"person_wearing_turban_medium_skin_tone"),
            std::u16string_view(u"person wearing turban: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FE"),
            std::u16string_view(u"person_wearing_turban_medium_dark_skin_tone"),
            std::u16string_view(u"person wearing turban: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FF"),
            std::u16string_view(u"person_wearing_turban_dark_skin_tone"),
            std::u16string_view(u"person wearing turban: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban"),
            std::u16string_view(u"man wearing turban"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban_light_skin_tone"),
            std::u16string_view(u"man wearing turban: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban_medium_light_skin_tone"),
            std::u16string_view(u"man wearing turban: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban_medium_skin_tone"),
            std::u16string_view(u"man wearing turban: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban_medium_dark_skin_tone"),
            std::u16string_view(u"man wearing turban: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_wearing_turban_dark_skin_tone"),
            std::u16string_view(u"man wearing turban: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban"),
            std::u16string_view(u"woman wearing turban"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban_light_skin_tone"),
            std::u16string_view(u"woman wearing turban: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban_medium_light_skin_tone"),
            std::u16string_view(u"woman wearing turban: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban_medium_skin_tone"),
            std::u16string_view(u"woman wearing turban: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban_medium_dark_skin_tone"),
            std::u16string_view(u"woman wearing turban: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F473\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_wearing_turban_dark_skin_tone"),
            std::u16string_view(u"woman wearing turban: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472"),
            std::u16string_view(u"person_with_skullcap"),
            std::u16string_view(u"person with skullcap"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472\U0001F3FB"),
            std::u16string_view(u"person_with_skullcap_light_skin_tone"),
            std::u16string_view(u"person with skullcap: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472\U0001F3FC"),
            std::u16string_view(u"person_with_skullcap_medium_light_skin_tone"),
            std::u16string_view(u"person with skullcap: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472\U0001F3FD"),
            std::u16string_view(u"person_with_skullcap_medium_skin_tone"),
            std::u16string_view(u"person with skullcap: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472\U0001F3FE"),
            std::u16string_view(u"person_with_skullcap_medium_dark_skin_tone"),
            std::u16string_view(u"person with skullcap: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F472\U0001F3FF"),
            std::u16string_view(u"person_with_skullcap_dark_skin_tone"),
            std::u16string_view(u"person with skullcap: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5"),
            std::u16string_view(u"woman_with_headscarf"),
            std::u16string_view(u"woman with headscarf"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5\U0001F3FB"),
            std::u16string_view(u"woman_with_headscarf_light_skin_tone"),
            std::u16string_view(u"woman with headscarf: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5\U0001F3FC"),
            std::u16string_view(u"woman_with_headscarf_medium_light_skin_tone"),
            std::u16string_view(u"woman with headscarf: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5\U0001F3FD"),
            std::u16string_view(u"woman_with_headscarf_medium_skin_tone"),
            std::u16string_view(u"woman with headscarf: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5\U0001F3FE"),
            std::u16string_view(u"woman_with_headscarf_medium_dark_skin_tone"),
            std::u16string_view(u"woman with headscarf: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D5\U0001F3FF"),
            std::u16string_view(u"woman_with_headscarf_dark_skin_tone"),
            std::u16string_view(u"woman with headscarf: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935"),
            std::u16string_view(u"person_in_tuxedo"),
            std::u16string_view(u"person in tuxedo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FB"),
            std::u16string_view(u"person_in_tuxedo_light_skin_tone"),
            std::u16string_view(u"person in tuxedo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FC"),
            std::u16string_view(u"person_in_tuxedo_medium_light_skin_tone"),
            std::u16string_view(u"person in tuxedo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FD"),
            std::u16string_view(u"person_in_tuxedo_medium_skin_tone"),
            std::u16string_view(u"person in tuxedo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FE"),
            std::u16string_view(u"person_in_tuxedo_medium_dark_skin_tone"),
            std::u16string_view(u"person in tuxedo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FF"),
            std::u16string_view(u"person_in_tuxedo_dark_skin_tone"),
            std::u16string_view(u"person in tuxedo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo"),
            std::u16string_view(u"man in tuxedo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo_light_skin_tone"),
            std::u16string_view(u"man in tuxedo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo_medium_light_skin_tone"),
            std::u16string_view(u"man in tuxedo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo_medium_skin_tone"),
            std::u16string_view(u"man in tuxedo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo_medium_dark_skin_tone"),
            std::u16string_view(u"man in tuxedo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_tuxedo_dark_skin_tone"),
            std::u16string_view(u"man in tuxedo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo"),
            std::u16string_view(u"woman in tuxedo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo_light_skin_tone"),
            std::u16string_view(u"woman in tuxedo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo_medium_light_skin_tone"),
            std::u16string_view(u"woman in tuxedo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo_medium_skin_tone"),
            std::u16string_view(u"woman in tuxedo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo_medium_dark_skin_tone"),
            std::u16string_view(u"woman in tuxedo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F935\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_tuxedo_dark_skin_tone"),
            std::u16string_view(u"woman in tuxedo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470"),
            std::u16string_view(u"person_with_veil"),
            std::u16string_view(u"person with veil"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FB"),
            std::u16string_view(u"person_with_veil_light_skin_tone"),
            std::u16string_view(u"person with veil: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FC"),
            std::u16string_view(u"person_with_veil_medium_light_skin_tone"),
            std::u16string_view(u"person with veil: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FD"),
            std::u16string_view(u"person_with_veil_medium_skin_tone"),
            std::u16string_view(u"person with veil: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FE"),
            std::u16string_view(u"person_with_veil_medium_dark_skin_tone"),
            std::u16string_view(u"person with veil: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FF"),
            std::u16string_view(u"person_with_veil_dark_skin_tone"),
            std::u16string_view(u"person with veil: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil"),
            std::u16string_view(u"man with veil"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil_light_skin_tone"),
            std::u16string_view(u"man with veil: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil_medium_light_skin_tone"),
            std::u16string_view(u"man with veil: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil_medium_skin_tone"),
            std::u16string_view(u"man with veil: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil_medium_dark_skin_tone"),
            std::u16string_view(u"man with veil: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_with_veil_dark_skin_tone"),
            std::u16string_view(u"man with veil: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil"),
            std::u16string_view(u"woman with veil"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil_light_skin_tone"),
            std::u16string_view(u"woman with veil: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil_medium_light_skin_tone"),
            std::u16string_view(u"woman with veil: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil_medium_skin_tone"),
            std::u16string_view(u"woman with veil: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil_medium_dark_skin_tone"),
            std::u16string_view(u"woman with veil: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F470\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_with_veil_dark_skin_tone"),
            std::u16string_view(u"woman with veil: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930"),
            std::u16string_view(u"pregnant_woman"),
            std::u16string_view(u"pregnant woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930\U0001F3FB"),
            std::u16string_view(u"pregnant_woman_light_skin_tone"),
            std::u16string_view(u"pregnant woman: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930\U0001F3FC"),
            std::u16string_view(u"pregnant_woman_medium_light_skin_tone"),
            std::u16string_view(u"pregnant woman: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930\U0001F3FD"),
            std::u16string_view(u"pregnant_woman_medium_skin_tone"),
            std::u16string_view(u"pregnant woman: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930\U0001F3FE"),
            std::u16string_view(u"pregnant_woman_medium_dark_skin_tone"),
            std::u16string_view(u"pregnant woman: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F930\U0001F3FF"),
            std::u16string_view(u"pregnant_woman_dark_skin_tone"),
            std::u16string_view(u"pregnant woman: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3"),
            std::u16string_view(u"pregnant_man"),
            std::u16string_view(u"pregnant man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3\U0001F3FB"),
            std::u16string_view(u"pregnant_man_light_skin_tone"),
            std::u16string_view(u"pregnant man: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3\U0001F3FC"),
            std::u16string_view(u"pregnant_man_medium_light_skin_tone"),
            std::u16string_view(u"pregnant man: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3\U0001F3FD"),
            std::u16string_view(u"pregnant_man_medium_skin_tone"),
            std::u16string_view(u"pregnant man: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3\U0001F3FE"),
            std::u16string_view(u"pregnant_man_medium_dark_skin_tone"),
            std::u16string_view(u"pregnant man: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC3\U0001F3FF"),
            std::u16string_view(u"pregnant_man_dark_skin_tone"),
            std::u16string_view(u"pregnant man: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4"),
            std::u16string_view(u"pregnant_person"),
            std::u16string_view(u"pregnant person"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4\U0001F3FB"),
            std::u16string_view(u"pregnant_person_light_skin_tone"),
            std::u16string_view(u"pregnant person: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4\U0001F3FC"),
            std::u16string_view(u"pregnant_person_medium_light_skin_tone"),
            std::u16string_view(u"pregnant person: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4\U0001F3FD"),
            std::u16string_view(u"pregnant_person_medium_skin_tone"),
            std::u16string_view(u"pregnant person: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4\U0001F3FE"),
            std::u16string_view(u"pregnant_person_medium_dark_skin_tone"),
            std::u16string_view(u"pregnant person: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC4\U0001F3FF"),
            std::u16string_view(u"pregnant_person_dark_skin_tone"),
            std::u16string_view(u"pregnant person: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931"),
            std::u16string_view(u"breast_feeding"),
            std::u16string_view(u"breast-feeding"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931\U0001F3FB"),
            std::u16string_view(u"breast_feeding_light_skin_tone"),
            std::u16string_view(u"breast-feeding: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931\U0001F3FC"),
            std::u16string_view(u"breast_feeding_medium_light_skin_tone"),
            std::u16string_view(u"breast-feeding: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931\U0001F3FD"),
            std::u16string_view(u"breast_feeding_medium_skin_tone"),
            std::u16string_view(u"breast-feeding: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931\U0001F3FE"),
            std::u16string_view(u"breast_feeding_medium_dark_skin_tone"),
            std::u16string_view(u"breast-feeding: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F931\U0001F3FF"),
            std::u16string_view(u"breast_feeding_dark_skin_tone"),
            std::u16string_view(u"breast-feeding: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby"),
            std::u16string_view(u"woman feeding baby"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby_light_skin_tone"),
            std::u16string_view(u"woman feeding baby: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby_medium_light_skin_tone"),
            std::u16string_view(u"woman feeding baby: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby_medium_skin_tone"),
            std::u16string_view(u"woman feeding baby: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby_medium_dark_skin_tone"),
            std::u16string_view(u"woman feeding baby: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F37C"),
            std::u16string_view(u"woman_feeding_baby_dark_skin_tone"),
            std::u16string_view(u"woman feeding baby: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby"),
            std::u16string_view(u"man feeding baby"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby_light_skin_tone"),
            std::u16string_view(u"man feeding baby: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby_medium_light_skin_tone"),
            std::u16string_view(u"man feeding baby: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby_medium_skin_tone"),
            std::u16string_view(u"man feeding baby: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby_medium_dark_skin_tone"),
            std::u16string_view(u"man feeding baby: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F37C"),
            std::u16string_view(u"man_feeding_baby_dark_skin_tone"),
            std::u16string_view(u"man feeding baby: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby"),
            std::u16string_view(u"person feeding baby"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby_light_skin_tone"),
            std::u16string_view(u"person feeding baby: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby_medium_light_skin_tone"),
            std::u16string_view(u"person feeding baby: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby_medium_skin_tone"),
            std::u16string_view(u"person feeding baby: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby_medium_dark_skin_tone"),
            std::u16string_view(u"person feeding baby: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F37C"),
            std::u16string_view(u"person_feeding_baby_dark_skin_tone"),
            std::u16string_view(u"person feeding baby: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C"),
            std::u16string_view(u"baby_angel"),
            std::u16string_view(u"baby angel"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C\U0001F3FB"),
            std::u16string_view(u"baby_angel_light_skin_tone"),
            std::u16string_view(u"baby angel: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C\U0001F3FC"),
            std::u16string_view(u"baby_angel_medium_light_skin_tone"),
            std::u16string_view(u"baby angel: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C\U0001F3FD"),
            std::u16string_view(u"baby_angel_medium_skin_tone"),
            std::u16string_view(u"baby angel: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C\U0001F3FE"),
            std::u16string_view(u"baby_angel_medium_dark_skin_tone"),
            std::u16string_view(u"baby angel: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F47C\U0001F3FF"),
            std::u16string_view(u"baby_angel_dark_skin_tone"),
            std::u16string_view(u"baby angel: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385"),
            std::u16string_view(u"santa_claus"),
            std::u16string_view(u"Santa Claus"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385\U0001F3FB"),
            std::u16string_view(u"santa_claus_light_skin_tone"),
            std::u16string_view(u"Santa Claus: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385\U0001F3FC"),
            std::u16string_view(u"santa_claus_medium_light_skin_tone"),
            std::u16string_view(u"Santa Claus: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385\U0001F3FD"),
            std::u16string_view(u"santa_claus_medium_skin_tone"),
            std::u16string_view(u"Santa Claus: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385\U0001F3FE"),
            std::u16string_view(u"santa_claus_medium_dark_skin_tone"),
            std::u16string_view(u"Santa Claus: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F385\U0001F3FF"),
            std::u16string_view(u"santa_claus_dark_skin_tone"),
            std::u16string_view(u"Santa Claus: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936"),
            std::u16string_view(u"mrs_claus"),
            std::u16string_view(u"Mrs. Claus"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936\U0001F3FB"),
            std::u16string_view(u"mrs_claus_light_skin_tone"),
            std::u16string_view(u"Mrs. Claus: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936\U0001F3FC"),
            std::u16string_view(u"mrs_claus_medium_light_skin_tone"),
            std::u16string_view(u"Mrs. Claus: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936\U0001F3FD"),
            std::u16string_view(u"mrs_claus_medium_skin_tone"),
            std::u16string_view(u"Mrs. Claus: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936\U0001F3FE"),
            std::u16string_view(u"mrs_claus_medium_dark_skin_tone"),
            std::u16string_view(u"Mrs. Claus: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F936\U0001F3FF"),
            std::u16string_view(u"mrs_claus_dark_skin_tone"),
            std::u16string_view(u"Mrs. Claus: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus"),
            std::u16string_view(u"mx claus"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus_light_skin_tone"),
            std::u16string_view(u"mx claus: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus_medium_light_skin_tone"),
            std::u16string_view(u"mx claus: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus_medium_skin_tone"),
            std::u16string_view(u"mx claus: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus_medium_dark_skin_tone"),
            std::u16string_view(u"mx claus: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F384"),
            std::u16string_view(u"mx_claus_dark_skin_tone"),
            std::u16string_view(u"mx claus: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8"),
            std::u16string_view(u"superhero"),
            std::u16string_view(u"superhero"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FB"),
            std::u16string_view(u"superhero_light_skin_tone"),
            std::u16string_view(u"superhero: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FC"),
            std::u16string_view(u"superhero_medium_light_skin_tone"),
            std::u16string_view(u"superhero: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FD"),
            std::u16string_view(u"superhero_medium_skin_tone"),
            std::u16string_view(u"superhero: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FE"),
            std::u16string_view(u"superhero_medium_dark_skin_tone"),
            std::u16string_view(u"superhero: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FF"),
            std::u16string_view(u"superhero_dark_skin_tone"),
            std::u16string_view(u"superhero: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero"),
            std::u16string_view(u"man superhero"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero_light_skin_tone"),
            std::u16string_view(u"man superhero: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero_medium_light_skin_tone"),
            std::u16string_view(u"man superhero: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero_medium_skin_tone"),
            std::u16string_view(u"man superhero: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero_medium_dark_skin_tone"),
            std::u16string_view(u"man superhero: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_superhero_dark_skin_tone"),
            std::u16string_view(u"man superhero: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero"),
            std::u16string_view(u"woman superhero"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero_light_skin_tone"),
            std::u16string_view(u"woman superhero: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero_medium_light_skin_tone"),
            std::u16string_view(u"woman superhero: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero_medium_skin_tone"),
            std::u16string_view(u"woman superhero: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero_medium_dark_skin_tone"),
            std::u16string_view(u"woman superhero: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B8\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_superhero_dark_skin_tone"),
            std::u16string_view(u"woman superhero: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9"),
            std::u16string_view(u"supervillain"),
            std::u16string_view(u"supervillain"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FB"),
            std::u16string_view(u"supervillain_light_skin_tone"),
            std::u16string_view(u"supervillain: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FC"),
            std::u16string_view(u"supervillain_medium_light_skin_tone"),
            std::u16string_view(u"supervillain: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FD"),
            std::u16string_view(u"supervillain_medium_skin_tone"),
            std::u16string_view(u"supervillain: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FE"),
            std::u16string_view(u"supervillain_medium_dark_skin_tone"),
            std::u16string_view(u"supervillain: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FF"),
            std::u16string_view(u"supervillain_dark_skin_tone"),
            std::u16string_view(u"supervillain: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain"),
            std::u16string_view(u"man supervillain"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain_light_skin_tone"),
            std::u16string_view(u"man supervillain: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain_medium_light_skin_tone"),
            std::u16string_view(u"man supervillain: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain_medium_skin_tone"),
            std::u16string_view(u"man supervillain: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain_medium_dark_skin_tone"),
            std::u16string_view(u"man supervillain: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_supervillain_dark_skin_tone"),
            std::u16string_view(u"man supervillain: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain"),
            std::u16string_view(u"woman supervillain"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain_light_skin_tone"),
            std::u16string_view(u"woman supervillain: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain_medium_light_skin_tone"),
            std::u16string_view(u"woman supervillain: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain_medium_skin_tone"),
            std::u16string_view(u"woman supervillain: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain_medium_dark_skin_tone"),
            std::u16string_view(u"woman supervillain: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9B9\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_supervillain_dark_skin_tone"),
            std::u16string_view(u"woman supervillain: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9"),
            std::u16string_view(u"mage"),
            std::u16string_view(u"mage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FB"),
            std::u16string_view(u"mage_light_skin_tone"),
            std::u16string_view(u"mage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FC"),
            std::u16string_view(u"mage_medium_light_skin_tone"),
            std::u16string_view(u"mage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FD"),
            std::u16string_view(u"mage_medium_skin_tone"),
            std::u16string_view(u"mage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FE"),
            std::u16string_view(u"mage_medium_dark_skin_tone"),
            std::u16string_view(u"mage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FF"),
            std::u16string_view(u"mage_dark_skin_tone"),
            std::u16string_view(u"mage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage"),
            std::u16string_view(u"man mage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage_light_skin_tone"),
            std::u16string_view(u"man mage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage_medium_light_skin_tone"),
            std::u16string_view(u"man mage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage_medium_skin_tone"),
            std::u16string_view(u"man mage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage_medium_dark_skin_tone"),
            std::u16string_view(u"man mage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mage_dark_skin_tone"),
            std::u16string_view(u"man mage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage"),
            std::u16string_view(u"woman mage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage_light_skin_tone"),
            std::u16string_view(u"woman mage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage_medium_light_skin_tone"),
            std::u16string_view(u"woman mage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage_medium_skin_tone"),
            std::u16string_view(u"woman mage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage_medium_dark_skin_tone"),
            std::u16string_view(u"woman mage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D9\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mage_dark_skin_tone"),
            std::u16string_view(u"woman mage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA"),
            std::u16string_view(u"fairy"),
            std::u16string_view(u"fairy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FB"),
            std::u16string_view(u"fairy_light_skin_tone"),
            std::u16string_view(u"fairy: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FC"),
            std::u16string_view(u"fairy_medium_light_skin_tone"),
            std::u16string_view(u"fairy: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FD"),
            std::u16string_view(u"fairy_medium_skin_tone"),
            std::u16string_view(u"fairy: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FE"),
            std::u16string_view(u"fairy_medium_dark_skin_tone"),
            std::u16string_view(u"fairy: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FF"),
            std::u16string_view(u"fairy_dark_skin_tone"),
            std::u16string_view(u"fairy: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy"),
            std::u16string_view(u"man fairy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy_light_skin_tone"),
            std::u16string_view(u"man fairy: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy_medium_light_skin_tone"),
            std::u16string_view(u"man fairy: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy_medium_skin_tone"),
            std::u16string_view(u"man fairy: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy_medium_dark_skin_tone"),
            std::u16string_view(u"man fairy: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_fairy_dark_skin_tone"),
            std::u16string_view(u"man fairy: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy"),
            std::u16string_view(u"woman fairy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy_light_skin_tone"),
            std::u16string_view(u"woman fairy: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy_medium_light_skin_tone"),
            std::u16string_view(u"woman fairy: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy_medium_skin_tone"),
            std::u16string_view(u"woman fairy: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy_medium_dark_skin_tone"),
            std::u16string_view(u"woman fairy: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DA\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_fairy_dark_skin_tone"),
            std::u16string_view(u"woman fairy: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB"),
            std::u16string_view(u"vampire"),
            std::u16string_view(u"vampire"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FB"),
            std::u16string_view(u"vampire_light_skin_tone"),
            std::u16string_view(u"vampire: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FC"),
            std::u16string_view(u"vampire_medium_light_skin_tone"),
            std::u16string_view(u"vampire: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FD"),
            std::u16string_view(u"vampire_medium_skin_tone"),
            std::u16string_view(u"vampire: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FE"),
            std::u16string_view(u"vampire_medium_dark_skin_tone"),
            std::u16string_view(u"vampire: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FF"),
            std::u16string_view(u"vampire_dark_skin_tone"),
            std::u16string_view(u"vampire: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire"),
            std::u16string_view(u"man vampire"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire_light_skin_tone"),
            std::u16string_view(u"man vampire: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire_medium_light_skin_tone"),
            std::u16string_view(u"man vampire: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire_medium_skin_tone"),
            std::u16string_view(u"man vampire: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire_medium_dark_skin_tone"),
            std::u16string_view(u"man vampire: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_vampire_dark_skin_tone"),
            std::u16string_view(u"man vampire: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire"),
            std::u16string_view(u"woman vampire"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire_light_skin_tone"),
            std::u16string_view(u"woman vampire: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire_medium_light_skin_tone"),
            std::u16string_view(u"woman vampire: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire_medium_skin_tone"),
            std::u16string_view(u"woman vampire: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire_medium_dark_skin_tone"),
            std::u16string_view(u"woman vampire: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DB\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_vampire_dark_skin_tone"),
            std::u16string_view(u"woman vampire: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC"),
            std::u16string_view(u"merperson"),
            std::u16string_view(u"merperson"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FB"),
            std::u16string_view(u"merperson_light_skin_tone"),
            std::u16string_view(u"merperson: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FC"),
            std::u16string_view(u"merperson_medium_light_skin_tone"),
            std::u16string_view(u"merperson: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FD"),
            std::u16string_view(u"merperson_medium_skin_tone"),
            std::u16string_view(u"merperson: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FE"),
            std::u16string_view(u"merperson_medium_dark_skin_tone"),
            std::u16string_view(u"merperson: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FF"),
            std::u16string_view(u"merperson_dark_skin_tone"),
            std::u16string_view(u"merperson: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman"),
            std::u16string_view(u"merman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman_light_skin_tone"),
            std::u16string_view(u"merman: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman_medium_light_skin_tone"),
            std::u16string_view(u"merman: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman_medium_skin_tone"),
            std::u16string_view(u"merman: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman_medium_dark_skin_tone"),
            std::u16string_view(u"merman: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"merman_dark_skin_tone"),
            std::u16string_view(u"merman: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid"),
            std::u16string_view(u"mermaid"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid_light_skin_tone"),
            std::u16string_view(u"mermaid: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid_medium_light_skin_tone"),
            std::u16string_view(u"mermaid: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid_medium_skin_tone"),
            std::u16string_view(u"mermaid: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid_medium_dark_skin_tone"),
            std::u16string_view(u"mermaid: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DC\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"mermaid_dark_skin_tone"),
            std::u16string_view(u"mermaid: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD"),
            std::u16string_view(u"elf"),
            std::u16string_view(u"elf"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FB"),
            std::u16string_view(u"elf_light_skin_tone"),
            std::u16string_view(u"elf: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FC"),
            std::u16string_view(u"elf_medium_light_skin_tone"),
            std::u16string_view(u"elf: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FD"),
            std::u16string_view(u"elf_medium_skin_tone"),
            std::u16string_view(u"elf: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FE"),
            std::u16string_view(u"elf_medium_dark_skin_tone"),
            std::u16string_view(u"elf: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FF"),
            std::u16string_view(u"elf_dark_skin_tone"),
            std::u16string_view(u"elf: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf"),
            std::u16string_view(u"man elf"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf_light_skin_tone"),
            std::u16string_view(u"man elf: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf_medium_light_skin_tone"),
            std::u16string_view(u"man elf: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf_medium_skin_tone"),
            std::u16string_view(u"man elf: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf_medium_dark_skin_tone"),
            std::u16string_view(u"man elf: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_elf_dark_skin_tone"),
            std::u16string_view(u"man elf: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf"),
            std::u16string_view(u"woman elf"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf_light_skin_tone"),
            std::u16string_view(u"woman elf: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf_medium_light_skin_tone"),
            std::u16string_view(u"woman elf: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf_medium_skin_tone"),
            std::u16string_view(u"woman elf: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf_medium_dark_skin_tone"),
            std::u16string_view(u"woman elf: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DD\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_elf_dark_skin_tone"),
            std::u16string_view(u"woman elf: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DE"),
            std::u16string_view(u"genie"),
            std::u16string_view(u"genie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_genie"),
            std::u16string_view(u"man genie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_genie"),
            std::u16string_view(u"woman genie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DF"),
            std::u16string_view(u"zombie"),
            std::u16string_view(u"zombie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_zombie"),
            std::u16string_view(u"man zombie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9DF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_zombie"),
            std::u16string_view(u"woman zombie"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CC"),
            std::u16string_view(u"troll"),
            std::u16string_view(u"troll"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486"),
            std::u16string_view(u"person_getting_massage"),
            std::u16string_view(u"person getting massage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FB"),
            std::u16string_view(u"person_getting_massage_light_skin_tone"),
            std::u16string_view(u"person getting massage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FC"),
            std::u16string_view(u"person_getting_massage_medium_light_skin_tone"),
            std::u16string_view(u"person getting massage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FD"),
            std::u16string_view(u"person_getting_massage_medium_skin_tone"),
            std::u16string_view(u"person getting massage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FE"),
            std::u16string_view(u"person_getting_massage_medium_dark_skin_tone"),
            std::u16string_view(u"person getting massage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FF"),
            std::u16string_view(u"person_getting_massage_dark_skin_tone"),
            std::u16string_view(u"person getting massage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage"),
            std::u16string_view(u"man getting massage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage_light_skin_tone"),
            std::u16string_view(u"man getting massage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage_medium_light_skin_tone"),
            std::u16string_view(u"man getting massage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage_medium_skin_tone"),
            std::u16string_view(u"man getting massage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage_medium_dark_skin_tone"),
            std::u16string_view(u"man getting massage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_massage_dark_skin_tone"),
            std::u16string_view(u"man getting massage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage"),
            std::u16string_view(u"woman getting massage"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage_light_skin_tone"),
            std::u16string_view(u"woman getting massage: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage_medium_light_skin_tone"),
            std::u16string_view(u"woman getting massage: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage_medium_skin_tone"),
            std::u16string_view(u"woman getting massage: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage_medium_dark_skin_tone"),
            std::u16string_view(u"woman getting massage: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F486\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_massage_dark_skin_tone"),
            std::u16string_view(u"woman getting massage: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487"),
            std::u16string_view(u"person_getting_haircut"),
            std::u16string_view(u"person getting haircut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FB"),
            std::u16string_view(u"person_getting_haircut_light_skin_tone"),
            std::u16string_view(u"person getting haircut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FC"),
            std::u16string_view(u"person_getting_haircut_medium_light_skin_tone"),
            std::u16string_view(u"person getting haircut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FD"),
            std::u16string_view(u"person_getting_haircut_medium_skin_tone"),
            std::u16string_view(u"person getting haircut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FE"),
            std::u16string_view(u"person_getting_haircut_medium_dark_skin_tone"),
            std::u16string_view(u"person getting haircut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FF"),
            std::u16string_view(u"person_getting_haircut_dark_skin_tone"),
            std::u16string_view(u"person getting haircut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut"),
            std::u16string_view(u"man getting haircut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut_light_skin_tone"),
            std::u16string_view(u"man getting haircut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut_medium_light_skin_tone"),
            std::u16string_view(u"man getting haircut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut_medium_skin_tone"),
            std::u16string_view(u"man getting haircut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut_medium_dark_skin_tone"),
            std::u16string_view(u"man getting haircut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_getting_haircut_dark_skin_tone"),
            std::u16string_view(u"man getting haircut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut"),
            std::u16string_view(u"woman getting haircut"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut_light_skin_tone"),
            std::u16string_view(u"woman getting haircut: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut_medium_light_skin_tone"),
            std::u16string_view(u"woman getting haircut: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut_medium_skin_tone"),
            std::u16string_view(u"woman getting haircut: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut_medium_dark_skin_tone"),
            std::u16string_view(u"woman getting haircut: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F487\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_getting_haircut_dark_skin_tone"),
            std::u16string_view(u"woman getting haircut: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6"),
            std::u16string_view(u"person_walking"),
            std::u16string_view(u"person walking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FB"),
            std::u16string_view(u"person_walking_light_skin_tone"),
            std::u16string_view(u"person walking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FC"),
            std::u16string_view(u"person_walking_medium_light_skin_tone"),
            std::u16string_view(u"person walking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FD"),
            std::u16string_view(u"person_walking_medium_skin_tone"),
            std::u16string_view(u"person walking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FE"),
            std::u16string_view(u"person_walking_medium_dark_skin_tone"),
            std::u16string_view(u"person walking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FF"),
            std::u16string_view(u"person_walking_dark_skin_tone"),
            std::u16string_view(u"person walking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking"),
            std::u16string_view(u"man walking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking_light_skin_tone"),
            std::u16string_view(u"man walking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking_medium_light_skin_tone"),
            std::u16string_view(u"man walking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking_medium_skin_tone"),
            std::u16string_view(u"man walking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking_medium_dark_skin_tone"),
            std::u16string_view(u"man walking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_walking_dark_skin_tone"),
            std::u16string_view(u"man walking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking"),
            std::u16string_view(u"woman walking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking_light_skin_tone"),
            std::u16string_view(u"woman walking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking_medium_light_skin_tone"),
            std::u16string_view(u"woman walking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking_medium_skin_tone"),
            std::u16string_view(u"woman walking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking_medium_dark_skin_tone"),
            std::u16string_view(u"woman walking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_walking_dark_skin_tone"),
            std::u16string_view(u"woman walking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right"),
            std::u16string_view(u"person walking facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FB\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right_light_skin_tone"),
            std::u16string_view(u"person walking facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"person walking facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right_medium_skin_tone"),
            std::u16string_view(u"person walking facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FE\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"person walking facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B6\U0001F3FF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_walking_facing_right_dark_skin_tone"),
            std::u16string_view(u"person walking facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right"),
            std::u16string_view(u"woman walking facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FB\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right_light_skin_tone"),
            std::u16string_view(u"woman walking facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FC\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"woman walking facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FD\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman walking facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FE\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"woman walking facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FF\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_walking_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman walking facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right"),
            std::u16string_view(u"man walking facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FB\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right_light_skin_tone"),
            std::u16string_view(u"man walking facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FC\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"man walking facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FD\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right_medium_skin_tone"),
            std::u16string_view(u"man walking facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FE\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man walking facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F6B6\U0001F3FF\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_walking_facing_right_dark_skin_tone"),
            std::u16string_view(u"man walking facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD"),
            std::u16string_view(u"person_standing"),
            std::u16string_view(u"person standing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FB"),
            std::u16string_view(u"person_standing_light_skin_tone"),
            std::u16string_view(u"person standing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FC"),
            std::u16string_view(u"person_standing_medium_light_skin_tone"),
            std::u16string_view(u"person standing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FD"),
            std::u16string_view(u"person_standing_medium_skin_tone"),
            std::u16string_view(u"person standing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FE"),
            std::u16string_view(u"person_standing_medium_dark_skin_tone"),
            std::u16string_view(u"person standing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FF"),
            std::u16string_view(u"person_standing_dark_skin_tone"),
            std::u16string_view(u"person standing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing"),
            std::u16string_view(u"man standing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing_light_skin_tone"),
            std::u16string_view(u"man standing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing_medium_light_skin_tone"),
            std::u16string_view(u"man standing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing_medium_skin_tone"),
            std::u16string_view(u"man standing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing_medium_dark_skin_tone"),
            std::u16string_view(u"man standing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_standing_dark_skin_tone"),
            std::u16string_view(u"man standing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing"),
            std::u16string_view(u"woman standing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing_light_skin_tone"),
            std::u16string_view(u"woman standing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing_medium_light_skin_tone"),
            std::u16string_view(u"woman standing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing_medium_skin_tone"),
            std::u16string_view(u"woman standing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing_medium_dark_skin_tone"),
            std::u16string_view(u"woman standing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CD\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_standing_dark_skin_tone"),
            std::u16string_view(u"woman standing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE"),
            std::u16string_view(u"person_kneeling"),
            std::u16string_view(u"person kneeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FB"),
            std::u16string_view(u"person_kneeling_light_skin_tone"),
            std::u16string_view(u"person kneeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FC"),
            std::u16string_view(u"person_kneeling_medium_light_skin_tone"),
            std::u16string_view(u"person kneeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FD"),
            std::u16string_view(u"person_kneeling_medium_skin_tone"),
            std::u16string_view(u"person kneeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FE"),
            std::u16string_view(u"person_kneeling_medium_dark_skin_tone"),
            std::u16string_view(u"person kneeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FF"),
            std::u16string_view(u"person_kneeling_dark_skin_tone"),
            std::u16string_view(u"person kneeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling"),
            std::u16string_view(u"man kneeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling_light_skin_tone"),
            std::u16string_view(u"man kneeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling_medium_light_skin_tone"),
            std::u16string_view(u"man kneeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling_medium_skin_tone"),
            std::u16string_view(u"man kneeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling_medium_dark_skin_tone"),
            std::u16string_view(u"man kneeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_kneeling_dark_skin_tone"),
            std::u16string_view(u"man kneeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling"),
            std::u16string_view(u"woman kneeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_light_skin_tone"),
            std::u16string_view(u"woman kneeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_medium_light_skin_tone"),
            std::u16string_view(u"woman kneeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_medium_skin_tone"),
            std::u16string_view(u"woman kneeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_medium_dark_skin_tone"),
            std::u16string_view(u"woman kneeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_dark_skin_tone"),
            std::u16string_view(u"woman kneeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right"),
            std::u16string_view(u"person kneeling facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FB\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right_light_skin_tone"),
            std::u16string_view(u"person kneeling facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"person kneeling facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right_medium_skin_tone"),
            std::u16string_view(u"person kneeling facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FE\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"person kneeling facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9CE\U0001F3FF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_kneeling_facing_right_dark_skin_tone"),
            std::u16string_view(u"person kneeling facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right"),
            std::u16string_view(u"woman kneeling facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FB\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right_light_skin_tone"),
            std::u16string_view(u"woman kneeling facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FC\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"woman kneeling facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FD\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman kneeling facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FE\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"woman kneeling facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FF\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_kneeling_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman kneeling facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right"),
            std::u16string_view(u"man kneeling facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FB\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right_light_skin_tone"),
            std::u16string_view(u"man kneeling facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FC\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"man kneeling facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FD\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right_medium_skin_tone"),
            std::u16string_view(u"man kneeling facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FE\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man kneeling facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9CE\U0001F3FF\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_kneeling_facing_right_dark_skin_tone"),
            std::u16string_view(u"man kneeling facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane"),
            std::u16string_view(u"person with white cane"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane_light_skin_tone"),
            std::u16string_view(u"person with white cane: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane_medium_light_skin_tone"),
            std::u16string_view(u"person with white cane: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane_medium_skin_tone"),
            std::u16string_view(u"person with white cane: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane_medium_dark_skin_tone"),
            std::u16string_view(u"person with white cane: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9AF"),
            std::u16string_view(u"person_with_white_cane_dark_skin_tone"),
            std::u16string_view(u"person with white cane: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right"),
            std::u16string_view(u"person with white cane facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right_light_skin_tone"),
            std::u16string_view(u"person with white cane facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"person with white cane facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right_medium_skin_tone"),
            std::u16string_view(u"person with white cane facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"person with white cane facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_with_white_cane_facing_right_dark_skin_tone"),
            std::u16string_view(u"person with white cane facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane"),
            std::u16string_view(u"man with white cane"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane_light_skin_tone"),
            std::u16string_view(u"man with white cane: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane_medium_light_skin_tone"),
            std::u16string_view(u"man with white cane: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane_medium_skin_tone"),
            std::u16string_view(u"man with white cane: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane_medium_dark_skin_tone"),
            std::u16string_view(u"man with white cane: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9AF"),
            std::u16string_view(u"man_with_white_cane_dark_skin_tone"),
            std::u16string_view(u"man with white cane: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right"),
            std::u16string_view(u"man with white cane facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right_light_skin_tone"),
            std::u16string_view(u"man with white cane facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"man with white cane facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right_medium_skin_tone"),
            std::u16string_view(u"man with white cane facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man with white cane facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_with_white_cane_facing_right_dark_skin_tone"),
            std::u16string_view(u"man with white cane facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane"),
            std::u16string_view(u"woman with white cane"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane_light_skin_tone"),
            std::u16string_view(u"woman with white cane: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane_medium_light_skin_tone"),
            std::u16string_view(u"woman with white cane: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane_medium_skin_tone"),
            std::u16string_view(u"woman with white cane: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane_medium_dark_skin_tone"),
            std::u16string_view(u"woman with white cane: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9AF"),
            std::u16string_view(u"woman_with_white_cane_dark_skin_tone"),
            std::u16string_view(u"woman with white cane: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right"),
            std::u16string_view(u"woman with white cane facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right_light_skin_tone"),
            std::u16string_view(u"woman with white cane facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"woman with white cane facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman with white cane facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"woman with white cane facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F9AF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_with_white_cane_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman with white cane facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair"),
            std::u16string_view(u"person in motorized wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair_light_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair_medium_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9BC"),
            std::u16string_view(u"person_in_motorized_wheelchair_dark_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_motorized_wheelchair_facing_right"),
            std::u16string_view(u"person in motorized wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_motorized_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"person_in_motorized_wheelchair_facing_right_medium_light_skin_tone"),
        std::u16string_view(u"person in motorized wheelchair facing right: medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_motorized_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"person_in_motorized_wheelchair_facing_right_medium_dark_skin_tone"),
        std::u16string_view(u"person in motorized wheelchair facing right: medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_motorized_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"person in motorized wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair"),
            std::u16string_view(u"man in motorized wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair_light_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair_medium_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9BC"),
            std::u16string_view(u"man_in_motorized_wheelchair_dark_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_motorized_wheelchair_facing_right"),
            std::u16string_view(u"man in motorized wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_motorized_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FC\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"man_in_motorized_wheelchair_facing_right_medium_light_skin_tone"),
        std::u16string_view(u"man in motorized wheelchair facing right: medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_motorized_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_motorized_wheelchair_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_motorized_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"man in motorized wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair"),
            std::u16string_view(u"woman in motorized wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair_light_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair_medium_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9BC"),
            std::u16string_view(u"woman_in_motorized_wheelchair_dark_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_motorized_wheelchair_facing_right"),
            std::u16string_view(u"woman in motorized wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_motorized_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"woman_in_motorized_wheelchair_facing_right_medium_light_skin_tone"),
        std::u16string_view(u"woman in motorized wheelchair facing right: medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_motorized_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"woman_in_motorized_wheelchair_facing_right_medium_dark_skin_tone"),
        std::u16string_view(u"woman in motorized wheelchair facing right: medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F9BC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_motorized_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman in motorized wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair"),
            std::u16string_view(u"person in manual wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair_light_skin_tone"),
            std::u16string_view(u"person in manual wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"person in manual wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair_medium_skin_tone"),
            std::u16string_view(u"person in manual wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"person in manual wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9BD"),
            std::u16string_view(u"person_in_manual_wheelchair_dark_skin_tone"),
            std::u16string_view(u"person in manual wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_manual_wheelchair_facing_right"),
            std::u16string_view(u"person in manual wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_manual_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"person in manual wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FC\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
        std::u16string_view(u"person_in_manual_wheelchair_facing_right_medium_light_skin_tone"),
        std::u16string_view(u"person in manual wheelchair facing right: medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_manual_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"person in manual wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_manual_wheelchair_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"person in manual wheelchair facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_in_manual_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"person in manual wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair"),
            std::u16string_view(u"man in manual wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair_light_skin_tone"),
            std::u16string_view(u"man in manual wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"man in manual wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair_medium_skin_tone"),
            std::u16string_view(u"man in manual wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"man in manual wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U0001F9BD"),
            std::u16string_view(u"man_in_manual_wheelchair_dark_skin_tone"),
            std::u16string_view(u"man in manual wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right"),
            std::u16string_view(u"man in manual wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"man in manual wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"man in manual wheelchair facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"man in manual wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man in manual wheelchair facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_in_manual_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"man in manual wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair"),
            std::u16string_view(u"woman in manual wheelchair"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair_light_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair_medium_light_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair_medium_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair_medium_dark_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U0001F9BD"),
            std::u16string_view(u"woman_in_manual_wheelchair_dark_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right"),
            std::u16string_view(u"woman in manual wheelchair facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right_light_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F9BD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_in_manual_wheelchair_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman in manual wheelchair facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3"),
            std::u16string_view(u"person_running"),
            std::u16string_view(u"person running"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FB"),
            std::u16string_view(u"person_running_light_skin_tone"),
            std::u16string_view(u"person running: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FC"),
            std::u16string_view(u"person_running_medium_light_skin_tone"),
            std::u16string_view(u"person running: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FD"),
            std::u16string_view(u"person_running_medium_skin_tone"),
            std::u16string_view(u"person running: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FE"),
            std::u16string_view(u"person_running_medium_dark_skin_tone"),
            std::u16string_view(u"person running: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FF"),
            std::u16string_view(u"person_running_dark_skin_tone"),
            std::u16string_view(u"person running: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running"),
            std::u16string_view(u"man running"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running_light_skin_tone"),
            std::u16string_view(u"man running: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running_medium_light_skin_tone"),
            std::u16string_view(u"man running: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running_medium_skin_tone"),
            std::u16string_view(u"man running: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running_medium_dark_skin_tone"),
            std::u16string_view(u"man running: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_running_dark_skin_tone"),
            std::u16string_view(u"man running: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running"),
            std::u16string_view(u"woman running"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running_light_skin_tone"),
            std::u16string_view(u"woman running: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running_medium_light_skin_tone"),
            std::u16string_view(u"woman running: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running_medium_skin_tone"),
            std::u16string_view(u"woman running: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running_medium_dark_skin_tone"),
            std::u16string_view(u"woman running: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_running_dark_skin_tone"),
            std::u16string_view(u"woman running: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right"),
            std::u16string_view(u"person running facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FB\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right_light_skin_tone"),
            std::u16string_view(u"person running facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FC\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"person running facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FD\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right_medium_skin_tone"),
            std::u16string_view(u"person running facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FE\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"person running facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C3\U0001F3FF\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"person_running_facing_right_dark_skin_tone"),
            std::u16string_view(u"person running facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right"),
            std::u16string_view(u"woman running facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FB\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right_light_skin_tone"),
            std::u16string_view(u"woman running facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FC\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"woman running facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FD\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right_medium_skin_tone"),
            std::u16string_view(u"woman running facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FE\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"woman running facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FF\U0000200D\U00002640\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"woman_running_facing_right_dark_skin_tone"),
            std::u16string_view(u"woman running facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right"),
            std::u16string_view(u"man running facing right"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FB\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right_light_skin_tone"),
            std::u16string_view(u"man running facing right: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FC\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right_medium_light_skin_tone"),
            std::u16string_view(u"man running facing right: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FD\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right_medium_skin_tone"),
            std::u16string_view(u"man running facing right: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FE\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right_medium_dark_skin_tone"),
            std::u16string_view(u"man running facing right: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F3C3\U0001F3FF\U0000200D\U00002642\U0000FE0F\U0000200D\U000027A1\U0000FE0F"),
            std::u16string_view(u"man_running_facing_right_dark_skin_tone"),
            std::u16string_view(u"man running facing right: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483"),
            std::u16string_view(u"woman_dancing"),
            std::u16string_view(u"woman dancing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483\U0001F3FB"),
            std::u16string_view(u"woman_dancing_light_skin_tone"),
            std::u16string_view(u"woman dancing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483\U0001F3FC"),
            std::u16string_view(u"woman_dancing_medium_light_skin_tone"),
            std::u16string_view(u"woman dancing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483\U0001F3FD"),
            std::u16string_view(u"woman_dancing_medium_skin_tone"),
            std::u16string_view(u"woman dancing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483\U0001F3FE"),
            std::u16string_view(u"woman_dancing_medium_dark_skin_tone"),
            std::u16string_view(u"woman dancing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F483\U0001F3FF"),
            std::u16string_view(u"woman_dancing_dark_skin_tone"),
            std::u16string_view(u"woman dancing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A"),
            std::u16string_view(u"man_dancing"),
            std::u16string_view(u"man dancing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A\U0001F3FB"),
            std::u16string_view(u"man_dancing_light_skin_tone"),
            std::u16string_view(u"man dancing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A\U0001F3FC"),
            std::u16string_view(u"man_dancing_medium_light_skin_tone"),
            std::u16string_view(u"man dancing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A\U0001F3FD"),
            std::u16string_view(u"man_dancing_medium_skin_tone"),
            std::u16string_view(u"man dancing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A\U0001F3FE"),
            std::u16string_view(u"man_dancing_medium_dark_skin_tone"),
            std::u16string_view(u"man dancing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F57A\U0001F3FF"),
            std::u16string_view(u"man_dancing_dark_skin_tone"),
            std::u16string_view(u"man dancing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0000FE0F"),
            std::u16string_view(u"person_in_suit_levitating"),
            std::u16string_view(u"person in suit levitating"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0001F3FB"),
            std::u16string_view(u"person_in_suit_levitating_light_skin_tone"),
            std::u16string_view(u"person in suit levitating: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0001F3FC"),
            std::u16string_view(u"person_in_suit_levitating_medium_light_skin_tone"),
            std::u16string_view(u"person in suit levitating: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0001F3FD"),
            std::u16string_view(u"person_in_suit_levitating_medium_skin_tone"),
            std::u16string_view(u"person in suit levitating: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0001F3FE"),
            std::u16string_view(u"person_in_suit_levitating_medium_dark_skin_tone"),
            std::u16string_view(u"person in suit levitating: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F574\U0001F3FF"),
            std::u16string_view(u"person_in_suit_levitating_dark_skin_tone"),
            std::u16string_view(u"person in suit levitating: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46F"),
            std::u16string_view(u"people_with_bunny_ears"),
            std::u16string_view(u"people with bunny ears"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46F\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"men_with_bunny_ears"),
            std::u16string_view(u"men with bunny ears"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46F\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"women_with_bunny_ears"),
            std::u16string_view(u"women with bunny ears"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6"),
            std::u16string_view(u"person_in_steamy_room"),
            std::u16string_view(u"person in steamy room"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FB"),
            std::u16string_view(u"person_in_steamy_room_light_skin_tone"),
            std::u16string_view(u"person in steamy room: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FC"),
            std::u16string_view(u"person_in_steamy_room_medium_light_skin_tone"),
            std::u16string_view(u"person in steamy room: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FD"),
            std::u16string_view(u"person_in_steamy_room_medium_skin_tone"),
            std::u16string_view(u"person in steamy room: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FE"),
            std::u16string_view(u"person_in_steamy_room_medium_dark_skin_tone"),
            std::u16string_view(u"person in steamy room: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FF"),
            std::u16string_view(u"person_in_steamy_room_dark_skin_tone"),
            std::u16string_view(u"person in steamy room: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room"),
            std::u16string_view(u"man in steamy room"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room_light_skin_tone"),
            std::u16string_view(u"man in steamy room: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room_medium_light_skin_tone"),
            std::u16string_view(u"man in steamy room: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room_medium_skin_tone"),
            std::u16string_view(u"man in steamy room: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room_medium_dark_skin_tone"),
            std::u16string_view(u"man in steamy room: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_steamy_room_dark_skin_tone"),
            std::u16string_view(u"man in steamy room: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room"),
            std::u16string_view(u"woman in steamy room"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room_light_skin_tone"),
            std::u16string_view(u"woman in steamy room: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room_medium_light_skin_tone"),
            std::u16string_view(u"woman in steamy room: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room_medium_skin_tone"),
            std::u16string_view(u"woman in steamy room: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room_medium_dark_skin_tone"),
            std::u16string_view(u"woman in steamy room: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D6\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_steamy_room_dark_skin_tone"),
            std::u16string_view(u"woman in steamy room: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7"),
            std::u16string_view(u"person_climbing"),
            std::u16string_view(u"person climbing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FB"),
            std::u16string_view(u"person_climbing_light_skin_tone"),
            std::u16string_view(u"person climbing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FC"),
            std::u16string_view(u"person_climbing_medium_light_skin_tone"),
            std::u16string_view(u"person climbing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FD"),
            std::u16string_view(u"person_climbing_medium_skin_tone"),
            std::u16string_view(u"person climbing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FE"),
            std::u16string_view(u"person_climbing_medium_dark_skin_tone"),
            std::u16string_view(u"person climbing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FF"),
            std::u16string_view(u"person_climbing_dark_skin_tone"),
            std::u16string_view(u"person climbing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing"),
            std::u16string_view(u"man climbing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing_light_skin_tone"),
            std::u16string_view(u"man climbing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing_medium_light_skin_tone"),
            std::u16string_view(u"man climbing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing_medium_skin_tone"),
            std::u16string_view(u"man climbing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing_medium_dark_skin_tone"),
            std::u16string_view(u"man climbing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_climbing_dark_skin_tone"),
            std::u16string_view(u"man climbing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing"),
            std::u16string_view(u"woman climbing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing_light_skin_tone"),
            std::u16string_view(u"woman climbing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing_medium_light_skin_tone"),
            std::u16string_view(u"woman climbing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing_medium_skin_tone"),
            std::u16string_view(u"woman climbing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing_medium_dark_skin_tone"),
            std::u16string_view(u"woman climbing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D7\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_climbing_dark_skin_tone"),
            std::u16string_view(u"woman climbing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93A"),
            std::u16string_view(u"person_fencing"),
            std::u16string_view(u"person fencing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7"),
            std::u16string_view(u"horse_racing"),
            std::u16string_view(u"horse racing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7\U0001F3FB"),
            std::u16string_view(u"horse_racing_light_skin_tone"),
            std::u16string_view(u"horse racing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7\U0001F3FC"),
            std::u16string_view(u"horse_racing_medium_light_skin_tone"),
            std::u16string_view(u"horse racing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7\U0001F3FD"),
            std::u16string_view(u"horse_racing_medium_skin_tone"),
            std::u16string_view(u"horse racing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7\U0001F3FE"),
            std::u16string_view(u"horse_racing_medium_dark_skin_tone"),
            std::u16string_view(u"horse racing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C7\U0001F3FF"),
            std::u16string_view(u"horse_racing_dark_skin_tone"),
            std::u16string_view(u"horse racing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F7\U0000FE0F"),
            std::u16string_view(u"skier"),
            std::u16string_view(u"skier"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2"),
            std::u16string_view(u"snowboarder"),
            std::u16string_view(u"snowboarder"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2\U0001F3FB"),
            std::u16string_view(u"snowboarder_light_skin_tone"),
            std::u16string_view(u"snowboarder: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2\U0001F3FC"),
            std::u16string_view(u"snowboarder_medium_light_skin_tone"),
            std::u16string_view(u"snowboarder: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2\U0001F3FD"),
            std::u16string_view(u"snowboarder_medium_skin_tone"),
            std::u16string_view(u"snowboarder: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2\U0001F3FE"),
            std::u16string_view(u"snowboarder_medium_dark_skin_tone"),
            std::u16string_view(u"snowboarder: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C2\U0001F3FF"),
            std::u16string_view(u"snowboarder_dark_skin_tone"),
            std::u16string_view(u"snowboarder: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0000FE0F"),
            std::u16string_view(u"person_golfing"),
            std::u16string_view(u"person golfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FB"),
            std::u16string_view(u"person_golfing_light_skin_tone"),
            std::u16string_view(u"person golfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FC"),
            std::u16string_view(u"person_golfing_medium_light_skin_tone"),
            std::u16string_view(u"person golfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FD"),
            std::u16string_view(u"person_golfing_medium_skin_tone"),
            std::u16string_view(u"person golfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FE"),
            std::u16string_view(u"person_golfing_medium_dark_skin_tone"),
            std::u16string_view(u"person golfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FF"),
            std::u16string_view(u"person_golfing_dark_skin_tone"),
            std::u16string_view(u"person golfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0000FE0F\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing"),
            std::u16string_view(u"man golfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing_light_skin_tone"),
            std::u16string_view(u"man golfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing_medium_light_skin_tone"),
            std::u16string_view(u"man golfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing_medium_skin_tone"),
            std::u16string_view(u"man golfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing_medium_dark_skin_tone"),
            std::u16string_view(u"man golfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_golfing_dark_skin_tone"),
            std::u16string_view(u"man golfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0000FE0F\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing"),
            std::u16string_view(u"woman golfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing_light_skin_tone"),
            std::u16string_view(u"woman golfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing_medium_light_skin_tone"),
            std::u16string_view(u"woman golfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing_medium_skin_tone"),
            std::u16string_view(u"woman golfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing_medium_dark_skin_tone"),
            std::u16string_view(u"woman golfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CC\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_golfing_dark_skin_tone"),
            std::u16string_view(u"woman golfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4"),
            std::u16string_view(u"person_surfing"),
            std::u16string_view(u"person surfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FB"),
            std::u16string_view(u"person_surfing_light_skin_tone"),
            std::u16string_view(u"person surfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FC"),
            std::u16string_view(u"person_surfing_medium_light_skin_tone"),
            std::u16string_view(u"person surfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FD"),
            std::u16string_view(u"person_surfing_medium_skin_tone"),
            std::u16string_view(u"person surfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FE"),
            std::u16string_view(u"person_surfing_medium_dark_skin_tone"),
            std::u16string_view(u"person surfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FF"),
            std::u16string_view(u"person_surfing_dark_skin_tone"),
            std::u16string_view(u"person surfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing"),
            std::u16string_view(u"man surfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing_light_skin_tone"),
            std::u16string_view(u"man surfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing_medium_light_skin_tone"),
            std::u16string_view(u"man surfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing_medium_skin_tone"),
            std::u16string_view(u"man surfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing_medium_dark_skin_tone"),
            std::u16string_view(u"man surfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_surfing_dark_skin_tone"),
            std::u16string_view(u"man surfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing"),
            std::u16string_view(u"woman surfing"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing_light_skin_tone"),
            std::u16string_view(u"woman surfing: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing_medium_light_skin_tone"),
            std::u16string_view(u"woman surfing: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing_medium_skin_tone"),
            std::u16string_view(u"woman surfing: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing_medium_dark_skin_tone"),
            std::u16string_view(u"woman surfing: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3C4\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_surfing_dark_skin_tone"),
            std::u16string_view(u"woman surfing: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3"),
            std::u16string_view(u"person_rowing_boat"),
            std::u16string_view(u"person rowing boat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FB"),
            std::u16string_view(u"person_rowing_boat_light_skin_tone"),
            std::u16string_view(u"person rowing boat: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FC"),
            std::u16string_view(u"person_rowing_boat_medium_light_skin_tone"),
            std::u16string_view(u"person rowing boat: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FD"),
            std::u16string_view(u"person_rowing_boat_medium_skin_tone"),
            std::u16string_view(u"person rowing boat: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FE"),
            std::u16string_view(u"person_rowing_boat_medium_dark_skin_tone"),
            std::u16string_view(u"person rowing boat: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FF"),
            std::u16string_view(u"person_rowing_boat_dark_skin_tone"),
            std::u16string_view(u"person rowing boat: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat"),
            std::u16string_view(u"man rowing boat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat_light_skin_tone"),
            std::u16string_view(u"man rowing boat: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat_medium_light_skin_tone"),
            std::u16string_view(u"man rowing boat: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat_medium_skin_tone"),
            std::u16string_view(u"man rowing boat: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat_medium_dark_skin_tone"),
            std::u16string_view(u"man rowing boat: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_rowing_boat_dark_skin_tone"),
            std::u16string_view(u"man rowing boat: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat"),
            std::u16string_view(u"woman rowing boat"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat_light_skin_tone"),
            std::u16string_view(u"woman rowing boat: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat_medium_light_skin_tone"),
            std::u16string_view(u"woman rowing boat: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat_medium_skin_tone"),
            std::u16string_view(u"woman rowing boat: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat_medium_dark_skin_tone"),
            std::u16string_view(u"woman rowing boat: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6A3\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_rowing_boat_dark_skin_tone"),
            std::u16string_view(u"woman rowing boat: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA"),
            std::u16string_view(u"person_swimming"),
            std::u16string_view(u"person swimming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FB"),
            std::u16string_view(u"person_swimming_light_skin_tone"),
            std::u16string_view(u"person swimming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FC"),
            std::u16string_view(u"person_swimming_medium_light_skin_tone"),
            std::u16string_view(u"person swimming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FD"),
            std::u16string_view(u"person_swimming_medium_skin_tone"),
            std::u16string_view(u"person swimming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FE"),
            std::u16string_view(u"person_swimming_medium_dark_skin_tone"),
            std::u16string_view(u"person swimming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FF"),
            std::u16string_view(u"person_swimming_dark_skin_tone"),
            std::u16string_view(u"person swimming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming"),
            std::u16string_view(u"man swimming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming_light_skin_tone"),
            std::u16string_view(u"man swimming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming_medium_light_skin_tone"),
            std::u16string_view(u"man swimming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming_medium_skin_tone"),
            std::u16string_view(u"man swimming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming_medium_dark_skin_tone"),
            std::u16string_view(u"man swimming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_swimming_dark_skin_tone"),
            std::u16string_view(u"man swimming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming"),
            std::u16string_view(u"woman swimming"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming_light_skin_tone"),
            std::u16string_view(u"woman swimming: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming_medium_light_skin_tone"),
            std::u16string_view(u"woman swimming: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming_medium_skin_tone"),
            std::u16string_view(u"woman swimming: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming_medium_dark_skin_tone"),
            std::u16string_view(u"woman swimming: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CA\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_swimming_dark_skin_tone"),
            std::u16string_view(u"woman swimming: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0000FE0F"),
            std::u16string_view(u"person_bouncing_ball"),
            std::u16string_view(u"person bouncing ball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FB"),
            std::u16string_view(u"person_bouncing_ball_light_skin_tone"),
            std::u16string_view(u"person bouncing ball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FC"),
            std::u16string_view(u"person_bouncing_ball_medium_light_skin_tone"),
            std::u16string_view(u"person bouncing ball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FD"),
            std::u16string_view(u"person_bouncing_ball_medium_skin_tone"),
            std::u16string_view(u"person bouncing ball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FE"),
            std::u16string_view(u"person_bouncing_ball_medium_dark_skin_tone"),
            std::u16string_view(u"person bouncing ball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FF"),
            std::u16string_view(u"person_bouncing_ball_dark_skin_tone"),
            std::u16string_view(u"person bouncing ball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0000FE0F\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball"),
            std::u16string_view(u"man bouncing ball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball_light_skin_tone"),
            std::u16string_view(u"man bouncing ball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball_medium_light_skin_tone"),
            std::u16string_view(u"man bouncing ball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball_medium_skin_tone"),
            std::u16string_view(u"man bouncing ball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball_medium_dark_skin_tone"),
            std::u16string_view(u"man bouncing ball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_bouncing_ball_dark_skin_tone"),
            std::u16string_view(u"man bouncing ball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0000FE0F\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball"),
            std::u16string_view(u"woman bouncing ball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball_light_skin_tone"),
            std::u16string_view(u"woman bouncing ball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball_medium_light_skin_tone"),
            std::u16string_view(u"woman bouncing ball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball_medium_skin_tone"),
            std::u16string_view(u"woman bouncing ball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball_medium_dark_skin_tone"),
            std::u16string_view(u"woman bouncing ball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U000026F9\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_bouncing_ball_dark_skin_tone"),
            std::u16string_view(u"woman bouncing ball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0000FE0F"),
            std::u16string_view(u"person_lifting_weights"),
            std::u16string_view(u"person lifting weights"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FB"),
            std::u16string_view(u"person_lifting_weights_light_skin_tone"),
            std::u16string_view(u"person lifting weights: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FC"),
            std::u16string_view(u"person_lifting_weights_medium_light_skin_tone"),
            std::u16string_view(u"person lifting weights: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FD"),
            std::u16string_view(u"person_lifting_weights_medium_skin_tone"),
            std::u16string_view(u"person lifting weights: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FE"),
            std::u16string_view(u"person_lifting_weights_medium_dark_skin_tone"),
            std::u16string_view(u"person lifting weights: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FF"),
            std::u16string_view(u"person_lifting_weights_dark_skin_tone"),
            std::u16string_view(u"person lifting weights: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0000FE0F\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights"),
            std::u16string_view(u"man lifting weights"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights_light_skin_tone"),
            std::u16string_view(u"man lifting weights: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights_medium_light_skin_tone"),
            std::u16string_view(u"man lifting weights: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights_medium_skin_tone"),
            std::u16string_view(u"man lifting weights: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights_medium_dark_skin_tone"),
            std::u16string_view(u"man lifting weights: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_lifting_weights_dark_skin_tone"),
            std::u16string_view(u"man lifting weights: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0000FE0F\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights"),
            std::u16string_view(u"woman lifting weights"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights_light_skin_tone"),
            std::u16string_view(u"woman lifting weights: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights_medium_light_skin_tone"),
            std::u16string_view(u"woman lifting weights: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights_medium_skin_tone"),
            std::u16string_view(u"woman lifting weights: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights_medium_dark_skin_tone"),
            std::u16string_view(u"woman lifting weights: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F3CB\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_lifting_weights_dark_skin_tone"),
            std::u16string_view(u"woman lifting weights: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4"),
            std::u16string_view(u"person_biking"),
            std::u16string_view(u"person biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FB"),
            std::u16string_view(u"person_biking_light_skin_tone"),
            std::u16string_view(u"person biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FC"),
            std::u16string_view(u"person_biking_medium_light_skin_tone"),
            std::u16string_view(u"person biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FD"),
            std::u16string_view(u"person_biking_medium_skin_tone"),
            std::u16string_view(u"person biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FE"),
            std::u16string_view(u"person_biking_medium_dark_skin_tone"),
            std::u16string_view(u"person biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FF"),
            std::u16string_view(u"person_biking_dark_skin_tone"),
            std::u16string_view(u"person biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking"),
            std::u16string_view(u"man biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking_light_skin_tone"),
            std::u16string_view(u"man biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking_medium_light_skin_tone"),
            std::u16string_view(u"man biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking_medium_skin_tone"),
            std::u16string_view(u"man biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking_medium_dark_skin_tone"),
            std::u16string_view(u"man biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_biking_dark_skin_tone"),
            std::u16string_view(u"man biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking"),
            std::u16string_view(u"woman biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking_light_skin_tone"),
            std::u16string_view(u"woman biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking_medium_light_skin_tone"),
            std::u16string_view(u"woman biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking_medium_skin_tone"),
            std::u16string_view(u"woman biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking_medium_dark_skin_tone"),
            std::u16string_view(u"woman biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B4\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_biking_dark_skin_tone"),
            std::u16string_view(u"woman biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5"),
            std::u16string_view(u"person_mountain_biking"),
            std::u16string_view(u"person mountain biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FB"),
            std::u16string_view(u"person_mountain_biking_light_skin_tone"),
            std::u16string_view(u"person mountain biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FC"),
            std::u16string_view(u"person_mountain_biking_medium_light_skin_tone"),
            std::u16string_view(u"person mountain biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FD"),
            std::u16string_view(u"person_mountain_biking_medium_skin_tone"),
            std::u16string_view(u"person mountain biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FE"),
            std::u16string_view(u"person_mountain_biking_medium_dark_skin_tone"),
            std::u16string_view(u"person mountain biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FF"),
            std::u16string_view(u"person_mountain_biking_dark_skin_tone"),
            std::u16string_view(u"person mountain biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking"),
            std::u16string_view(u"man mountain biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking_light_skin_tone"),
            std::u16string_view(u"man mountain biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking_medium_light_skin_tone"),
            std::u16string_view(u"man mountain biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking_medium_skin_tone"),
            std::u16string_view(u"man mountain biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking_medium_dark_skin_tone"),
            std::u16string_view(u"man mountain biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_mountain_biking_dark_skin_tone"),
            std::u16string_view(u"man mountain biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking"),
            std::u16string_view(u"woman mountain biking"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking_light_skin_tone"),
            std::u16string_view(u"woman mountain biking: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking_medium_light_skin_tone"),
            std::u16string_view(u"woman mountain biking: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking_medium_skin_tone"),
            std::u16string_view(u"woman mountain biking: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking_medium_dark_skin_tone"),
            std::u16string_view(u"woman mountain biking: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6B5\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_mountain_biking_dark_skin_tone"),
            std::u16string_view(u"woman mountain biking: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938"),
            std::u16string_view(u"person_cartwheeling"),
            std::u16string_view(u"person cartwheeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FB"),
            std::u16string_view(u"person_cartwheeling_light_skin_tone"),
            std::u16string_view(u"person cartwheeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FC"),
            std::u16string_view(u"person_cartwheeling_medium_light_skin_tone"),
            std::u16string_view(u"person cartwheeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FD"),
            std::u16string_view(u"person_cartwheeling_medium_skin_tone"),
            std::u16string_view(u"person cartwheeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FE"),
            std::u16string_view(u"person_cartwheeling_medium_dark_skin_tone"),
            std::u16string_view(u"person cartwheeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FF"),
            std::u16string_view(u"person_cartwheeling_dark_skin_tone"),
            std::u16string_view(u"person cartwheeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling"),
            std::u16string_view(u"man cartwheeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling_light_skin_tone"),
            std::u16string_view(u"man cartwheeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling_medium_light_skin_tone"),
            std::u16string_view(u"man cartwheeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling_medium_skin_tone"),
            std::u16string_view(u"man cartwheeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling_medium_dark_skin_tone"),
            std::u16string_view(u"man cartwheeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_cartwheeling_dark_skin_tone"),
            std::u16string_view(u"man cartwheeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling"),
            std::u16string_view(u"woman cartwheeling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling_light_skin_tone"),
            std::u16string_view(u"woman cartwheeling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling_medium_light_skin_tone"),
            std::u16string_view(u"woman cartwheeling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling_medium_skin_tone"),
            std::u16string_view(u"woman cartwheeling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling_medium_dark_skin_tone"),
            std::u16string_view(u"woman cartwheeling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F938\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_cartwheeling_dark_skin_tone"),
            std::u16string_view(u"woman cartwheeling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93C"),
            std::u16string_view(u"people_wrestling"),
            std::u16string_view(u"people wrestling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93C\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"men_wrestling"),
            std::u16string_view(u"men wrestling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93C\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"women_wrestling"),
            std::u16string_view(u"women wrestling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D"),
            std::u16string_view(u"person_playing_water_polo"),
            std::u16string_view(u"person playing water polo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FB"),
            std::u16string_view(u"person_playing_water_polo_light_skin_tone"),
            std::u16string_view(u"person playing water polo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FC"),
            std::u16string_view(u"person_playing_water_polo_medium_light_skin_tone"),
            std::u16string_view(u"person playing water polo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FD"),
            std::u16string_view(u"person_playing_water_polo_medium_skin_tone"),
            std::u16string_view(u"person playing water polo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FE"),
            std::u16string_view(u"person_playing_water_polo_medium_dark_skin_tone"),
            std::u16string_view(u"person playing water polo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FF"),
            std::u16string_view(u"person_playing_water_polo_dark_skin_tone"),
            std::u16string_view(u"person playing water polo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo"),
            std::u16string_view(u"man playing water polo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo_light_skin_tone"),
            std::u16string_view(u"man playing water polo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo_medium_light_skin_tone"),
            std::u16string_view(u"man playing water polo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo_medium_skin_tone"),
            std::u16string_view(u"man playing water polo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo_medium_dark_skin_tone"),
            std::u16string_view(u"man playing water polo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_water_polo_dark_skin_tone"),
            std::u16string_view(u"man playing water polo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo"),
            std::u16string_view(u"woman playing water polo"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo_light_skin_tone"),
            std::u16string_view(u"woman playing water polo: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo_medium_light_skin_tone"),
            std::u16string_view(u"woman playing water polo: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo_medium_skin_tone"),
            std::u16string_view(u"woman playing water polo: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo_medium_dark_skin_tone"),
            std::u16string_view(u"woman playing water polo: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93D\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_water_polo_dark_skin_tone"),
            std::u16string_view(u"woman playing water polo: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E"),
            std::u16string_view(u"person_playing_handball"),
            std::u16string_view(u"person playing handball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FB"),
            std::u16string_view(u"person_playing_handball_light_skin_tone"),
            std::u16string_view(u"person playing handball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FC"),
            std::u16string_view(u"person_playing_handball_medium_light_skin_tone"),
            std::u16string_view(u"person playing handball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FD"),
            std::u16string_view(u"person_playing_handball_medium_skin_tone"),
            std::u16string_view(u"person playing handball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FE"),
            std::u16string_view(u"person_playing_handball_medium_dark_skin_tone"),
            std::u16string_view(u"person playing handball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FF"),
            std::u16string_view(u"person_playing_handball_dark_skin_tone"),
            std::u16string_view(u"person playing handball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball"),
            std::u16string_view(u"man playing handball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball_light_skin_tone"),
            std::u16string_view(u"man playing handball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball_medium_light_skin_tone"),
            std::u16string_view(u"man playing handball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball_medium_skin_tone"),
            std::u16string_view(u"man playing handball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball_medium_dark_skin_tone"),
            std::u16string_view(u"man playing handball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_playing_handball_dark_skin_tone"),
            std::u16string_view(u"man playing handball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball"),
            std::u16string_view(u"woman playing handball"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball_light_skin_tone"),
            std::u16string_view(u"woman playing handball: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball_medium_light_skin_tone"),
            std::u16string_view(u"woman playing handball: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball_medium_skin_tone"),
            std::u16string_view(u"woman playing handball: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball_medium_dark_skin_tone"),
            std::u16string_view(u"woman playing handball: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F93E\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_playing_handball_dark_skin_tone"),
            std::u16string_view(u"woman playing handball: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939"),
            std::u16string_view(u"person_juggling"),
            std::u16string_view(u"person juggling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FB"),
            std::u16string_view(u"person_juggling_light_skin_tone"),
            std::u16string_view(u"person juggling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FC"),
            std::u16string_view(u"person_juggling_medium_light_skin_tone"),
            std::u16string_view(u"person juggling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FD"),
            std::u16string_view(u"person_juggling_medium_skin_tone"),
            std::u16string_view(u"person juggling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FE"),
            std::u16string_view(u"person_juggling_medium_dark_skin_tone"),
            std::u16string_view(u"person juggling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FF"),
            std::u16string_view(u"person_juggling_dark_skin_tone"),
            std::u16string_view(u"person juggling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling"),
            std::u16string_view(u"man juggling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling_light_skin_tone"),
            std::u16string_view(u"man juggling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling_medium_light_skin_tone"),
            std::u16string_view(u"man juggling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling_medium_skin_tone"),
            std::u16string_view(u"man juggling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling_medium_dark_skin_tone"),
            std::u16string_view(u"man juggling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_juggling_dark_skin_tone"),
            std::u16string_view(u"man juggling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling"),
            std::u16string_view(u"woman juggling"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling_light_skin_tone"),
            std::u16string_view(u"woman juggling: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling_medium_light_skin_tone"),
            std::u16string_view(u"woman juggling: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling_medium_skin_tone"),
            std::u16string_view(u"woman juggling: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling_medium_dark_skin_tone"),
            std::u16string_view(u"woman juggling: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F939\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_juggling_dark_skin_tone"),
            std::u16string_view(u"woman juggling: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8"),
            std::u16string_view(u"person_in_lotus_position"),
            std::u16string_view(u"person in lotus position"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FB"),
            std::u16string_view(u"person_in_lotus_position_light_skin_tone"),
            std::u16string_view(u"person in lotus position: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FC"),
            std::u16string_view(u"person_in_lotus_position_medium_light_skin_tone"),
            std::u16string_view(u"person in lotus position: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FD"),
            std::u16string_view(u"person_in_lotus_position_medium_skin_tone"),
            std::u16string_view(u"person in lotus position: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FE"),
            std::u16string_view(u"person_in_lotus_position_medium_dark_skin_tone"),
            std::u16string_view(u"person in lotus position: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FF"),
            std::u16string_view(u"person_in_lotus_position_dark_skin_tone"),
            std::u16string_view(u"person in lotus position: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position"),
            std::u16string_view(u"man in lotus position"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FB\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position_light_skin_tone"),
            std::u16string_view(u"man in lotus position: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FC\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position_medium_light_skin_tone"),
            std::u16string_view(u"man in lotus position: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FD\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position_medium_skin_tone"),
            std::u16string_view(u"man in lotus position: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FE\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position_medium_dark_skin_tone"),
            std::u16string_view(u"man in lotus position: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FF\U0000200D\U00002642\U0000FE0F"),
            std::u16string_view(u"man_in_lotus_position_dark_skin_tone"),
            std::u16string_view(u"man in lotus position: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position"),
            std::u16string_view(u"woman in lotus position"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FB\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position_light_skin_tone"),
            std::u16string_view(u"woman in lotus position: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FC\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position_medium_light_skin_tone"),
            std::u16string_view(u"woman in lotus position: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FD\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position_medium_skin_tone"),
            std::u16string_view(u"woman in lotus position: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FE\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position_medium_dark_skin_tone"),
            std::u16string_view(u"woman in lotus position: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D8\U0001F3FF\U0000200D\U00002640\U0000FE0F"),
            std::u16string_view(u"woman_in_lotus_position_dark_skin_tone"),
            std::u16string_view(u"woman in lotus position: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0"),
            std::u16string_view(u"person_taking_bath"),
            std::u16string_view(u"person taking bath"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0\U0001F3FB"),
            std::u16string_view(u"person_taking_bath_light_skin_tone"),
            std::u16string_view(u"person taking bath: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0\U0001F3FC"),
            std::u16string_view(u"person_taking_bath_medium_light_skin_tone"),
            std::u16string_view(u"person taking bath: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0\U0001F3FD"),
            std::u16string_view(u"person_taking_bath_medium_skin_tone"),
            std::u16string_view(u"person taking bath: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0\U0001F3FE"),
            std::u16string_view(u"person_taking_bath_medium_dark_skin_tone"),
            std::u16string_view(u"person taking bath: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6C0\U0001F3FF"),
            std::u16string_view(u"person_taking_bath_dark_skin_tone"),
            std::u16string_view(u"person taking bath: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC"),
            std::u16string_view(u"person_in_bed"),
            std::u16string_view(u"person in bed"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC\U0001F3FB"),
            std::u16string_view(u"person_in_bed_light_skin_tone"),
            std::u16string_view(u"person in bed: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC\U0001F3FC"),
            std::u16string_view(u"person_in_bed_medium_light_skin_tone"),
            std::u16string_view(u"person in bed: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC\U0001F3FD"),
            std::u16string_view(u"person_in_bed_medium_skin_tone"),
            std::u16string_view(u"person in bed: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC\U0001F3FE"),
            std::u16string_view(u"person_in_bed_medium_dark_skin_tone"),
            std::u16string_view(u"person in bed: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F6CC\U0001F3FF"),
            std::u16string_view(u"person_in_bed_dark_skin_tone"),
            std::u16string_view(u"person in bed: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F91D\U0000200D\U0001F9D1"),
            std::u16string_view(u"people_holding_hands"),
            std::u16string_view(u"people holding hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"people_holding_hands_light_skin_tone"),
            std::u16string_view(u"people holding hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"people_holding_hands_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"people holding hands: light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"people_holding_hands_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"people holding hands: light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"people_holding_hands_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"people holding hands: light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"people_holding_hands_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"people holding hands: light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"people_holding_hands_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"people holding hands: medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"people_holding_hands_medium_light_skin_tone"),
            std::u16string_view(u"people holding hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"people_holding_hands_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"people holding hands: medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"),
        std::u16string_view(u"people_holding_hands_medium_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"people holding hands: medium-light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"people_holding_hands_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"people holding hands: medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"people_holding_hands_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"people holding hands: medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"people_holding_hands_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"people holding hands: medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"people_holding_hands_medium_skin_tone"),
            std::u16string_view(u"people holding hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"people_holding_hands_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"people holding hands: medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"people_holding_hands_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"people holding hands: medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"people_holding_hands_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"people holding hands: medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"),
        std::u16string_view(u"people_holding_hands_medium_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"people holding hands: medium-dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"people_holding_hands_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"people holding hands: medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"people_holding_hands_medium_dark_skin_tone"),
            std::u16string_view(u"people holding hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"people_holding_hands_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"people holding hands: medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"people_holding_hands_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"people holding hands: dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"people_holding_hands_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"people holding hands: dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"people_holding_hands_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"people holding hands: dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"people_holding_hands_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"people holding hands: dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"people_holding_hands_dark_skin_tone"),
            std::u16string_view(u"people holding hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D"),
            std::u16string_view(u"women_holding_hands"),
            std::u16string_view(u"women holding hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D\U0001F3FB"),
            std::u16string_view(u"women_holding_hands_light_skin_tone"),
            std::u16string_view(u"women holding hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"women_holding_hands_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"women holding hands: light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"women_holding_hands_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"women holding hands: light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"women_holding_hands_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"women holding hands: light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"women_holding_hands_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"women holding hands: light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"women_holding_hands_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"women holding hands: medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D\U0001F3FC"),
            std::u16string_view(u"women_holding_hands_medium_light_skin_tone"),
            std::u16string_view(u"women holding hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"women_holding_hands_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"women holding hands: medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"),
        std::u16string_view(u"women_holding_hands_medium_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"women holding hands: medium-light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"women_holding_hands_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"women holding hands: medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"women_holding_hands_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"women holding hands: medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"women_holding_hands_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"women holding hands: medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D\U0001F3FD"),
            std::u16string_view(u"women_holding_hands_medium_skin_tone"),
            std::u16string_view(u"women holding hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"women_holding_hands_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"women holding hands: medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"women_holding_hands_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"women holding hands: medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"women_holding_hands_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"women holding hands: medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"),
        std::u16string_view(u"women_holding_hands_medium_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"women holding hands: medium-dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"women_holding_hands_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"women holding hands: medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D\U0001F3FE"),
            std::u16string_view(u"women_holding_hands_medium_dark_skin_tone"),
            std::u16string_view(u"women holding hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"women_holding_hands_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"women holding hands: medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"women_holding_hands_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"women holding hands: dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"women_holding_hands_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"women holding hands: dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"women_holding_hands_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"women holding hands: dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"women_holding_hands_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"women holding hands: dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46D\U0001F3FF"),
            std::u16string_view(u"women_holding_hands_dark_skin_tone"),
            std::u16string_view(u"women holding hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B"),
            std::u16string_view(u"woman_and_man_holding_hands"),
            std::u16string_view(u"woman and man holding hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B\U0001F3FB"),
            std::u16string_view(u"woman_and_man_holding_hands_light_skin_tone"),
            std::u16string_view(u"woman and man holding hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"woman_and_man_holding_hands_light_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: light skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"woman_and_man_holding_hands_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"woman and man holding hands: light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"woman_and_man_holding_hands_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"woman and man holding hands: light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"woman_and_man_holding_hands_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"woman and man holding hands: light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_light_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: medium-light skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B\U0001F3FC"),
            std::u16string_view(u"woman_and_man_holding_hands_medium_light_skin_tone"),
            std::u16string_view(u"woman and man holding hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_light_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: medium-light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(
              u"woman_and_man_holding_hands_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"woman and man holding hands: medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_light_skin_tone_dark_skin_tone"),
        std::u16string_view(u"woman and man holding hands: medium-light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"woman_and_man_holding_hands_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"woman and man holding hands: medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: medium skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B\U0001F3FD"),
            std::u16string_view(u"woman_and_man_holding_hands_medium_skin_tone"),
            std::u16string_view(u"woman and man holding hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: medium skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"woman_and_man_holding_hands_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"woman and man holding hands: medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_dark_skin_tone_light_skin_tone"),
        std::u16string_view(u"woman and man holding hands: medium-dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(
              u"woman_and_man_holding_hands_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"woman and man holding hands: medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_dark_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"woman and man holding hands: medium-dark skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B\U0001F3FE"),
            std::u16string_view(u"woman_and_man_holding_hands_medium_dark_skin_tone"),
            std::u16string_view(u"woman and man holding hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"woman_and_man_holding_hands_medium_dark_skin_tone_dark_skin_tone"),
        std::u16string_view(u"woman and man holding hands: medium-dark skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"woman_and_man_holding_hands_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"woman and man holding hands: dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"woman_and_man_holding_hands_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"woman and man holding hands: dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"woman_and_man_holding_hands_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"woman and man holding hands: dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"woman_and_man_holding_hands_dark_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"woman and man holding hands: dark skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46B\U0001F3FF"),
            std::u16string_view(u"woman_and_man_holding_hands_dark_skin_tone"),
            std::u16string_view(u"woman and man holding hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C"),
            std::u16string_view(u"men_holding_hands"),
            std::u16string_view(u"men holding hands"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C\U0001F3FB"),
            std::u16string_view(u"men_holding_hands_light_skin_tone"),
            std::u16string_view(u"men holding hands: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"men_holding_hands_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"men holding hands: light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"men_holding_hands_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"men holding hands: light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"men_holding_hands_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"men holding hands: light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"men_holding_hands_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"men holding hands: light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"men_holding_hands_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"men holding hands: medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C\U0001F3FC"),
            std::u16string_view(u"men_holding_hands_medium_light_skin_tone"),
            std::u16string_view(u"men holding hands: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"men_holding_hands_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"men holding hands: medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"men_holding_hands_medium_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"men holding hands: medium-light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"men_holding_hands_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"men holding hands: medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"men_holding_hands_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"men holding hands: medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"men_holding_hands_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"men holding hands: medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C\U0001F3FD"),
            std::u16string_view(u"men_holding_hands_medium_skin_tone"),
            std::u16string_view(u"men holding hands: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"men_holding_hands_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"men holding hands: medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"men_holding_hands_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"men holding hands: medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"men_holding_hands_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"men holding hands: medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"men_holding_hands_medium_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"men holding hands: medium-dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"men_holding_hands_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"men holding hands: medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C\U0001F3FE"),
            std::u16string_view(u"men_holding_hands_medium_dark_skin_tone"),
            std::u16string_view(u"men holding hands: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"men_holding_hands_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"men holding hands: medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"men_holding_hands_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"men holding hands: dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"men_holding_hands_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"men holding hands: dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"men_holding_hands_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"men holding hands: dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"men_holding_hands_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"men holding hands: dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46C\U0001F3FF"),
            std::u16string_view(u"men_holding_hands_dark_skin_tone"),
            std::u16string_view(u"men holding hands: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F"),
            std::u16string_view(u"kiss"),
            std::u16string_view(u"kiss"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F\U0001F3FB"),
            std::u16string_view(u"kiss_light_skin_tone"),
            std::u16string_view(u"kiss: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F\U0001F3FC"),
            std::u16string_view(u"kiss_medium_light_skin_tone"),
            std::u16string_view(u"kiss: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F\U0001F3FD"),
            std::u16string_view(u"kiss_medium_skin_tone"),
            std::u16string_view(u"kiss: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F\U0001F3FE"),
            std::u16string_view(u"kiss_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F48F\U0001F3FF"),
            std::u16string_view(u"kiss_dark_skin_tone"),
            std::u16string_view(u"kiss: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"kiss_person_person_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"kiss_person_person_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: person, person, light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"kiss_person_person_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"kiss_person_person_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"kiss_person_person_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"kiss_person_person_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                            u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"),
        std::u16string_view(u"kiss_person_person_medium_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"kiss: person, person, medium-light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"kiss_person_person_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"kiss_person_person_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"kiss_person_person_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"kiss_person_person_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"kiss_person_person_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"kiss_person_person_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                            u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"),
        std::u16string_view(u"kiss_person_person_medium_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"kiss: person, person, medium-dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"kiss_person_person_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(u"kiss_person_person_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(u"kiss_person_person_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(u"kiss_person_person_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: person, person, dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(u"kiss_person_person_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: person, person, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(u"kiss_person_person_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: person, person, dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468"),
            std::u16string_view(u"kiss_woman_man"),
            std::u16string_view(u"kiss: woman, man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_woman_man_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_woman_man_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_woman_man_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, man, light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_woman_man_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_woman_man_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_woman_man_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_woman_man_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_woman_man_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_woman_man_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_woman_man_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_woman_man_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_woman_man_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_woman_man_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_woman_man_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_woman_man_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_woman_man_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_woman_man_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_woman_man_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_woman_man_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_woman_man_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_woman_man_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_woman_man_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, man, dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_woman_man_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, man, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_woman_man_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_woman_man_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, man, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468"),
            std::u16string_view(u"kiss_man_man"),
            std::u16string_view(u"kiss: man, man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_man_man_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_man_man_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_man_man_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: man, man, light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_man_man_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_man_man_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_man_man_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_man_man_medium_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_man_man_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_man_man_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_man_man_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_man_man_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_man_man_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_man_man_medium_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_man_man_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_man_man_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_man_man_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_man_man_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_man_man_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_man_man_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_man_man_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"kiss_man_man_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"kiss_man_man_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: man, man, dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"kiss_man_man_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: man, man, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"kiss_man_man_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"kiss_man_man_dark_skin_tone"),
            std::u16string_view(u"kiss: man, man, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469"),
            std::u16string_view(u"kiss_woman_woman"),
            std::u16string_view(u"kiss: woman, woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"kiss_woman_woman_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"kiss_woman_woman_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"kiss_woman_woman_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"kiss_woman_woman_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"kiss_woman_woman_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"kiss_woman_woman_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"kiss_woman_woman_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"kiss_woman_woman_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                            u"\U0001F48B\U0000200D\U0001F469\U0001F3FE"),
        std::u16string_view(u"kiss_woman_woman_medium_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"kiss: woman, woman, medium-light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"kiss_woman_woman_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"kiss_woman_woman_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"kiss_woman_woman_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"kiss_woman_woman_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"kiss_woman_woman_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"kiss_woman_woman_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"kiss_woman_woman_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                            u"\U0001F48B\U0000200D\U0001F469\U0001F3FC"),
        std::u16string_view(u"kiss_woman_woman_medium_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"kiss: woman, woman, medium-dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"kiss_woman_woman_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"kiss_woman_woman_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"kiss_woman_woman_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"kiss_woman_woman_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"kiss_woman_woman_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"kiss_woman_woman_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"kiss_woman_woman_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D"
                                u"\U0001F48B\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"kiss_woman_woman_dark_skin_tone"),
            std::u16string_view(u"kiss: woman, woman, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491"),
            std::u16string_view(u"couple_with_heart"),
            std::u16string_view(u"couple with heart"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_light_skin_tone"),
            std::u16string_view(u"couple with heart: light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491\U0001F3FC"),
            std::u16string_view(u"couple_with_heart_medium_light_skin_tone"),
            std::u16string_view(u"couple with heart: medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_medium_skin_tone"),
            std::u16string_view(u"couple with heart: medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491\U0001F3FE"),
            std::u16string_view(u"couple_with_heart_medium_dark_skin_tone"),
            std::u16string_view(u"couple with heart: medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F491\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_dark_skin_tone"),
            std::u16string_view(u"couple with heart: dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_person_person_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_person_person_light_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"couple with heart: person, person, light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_person_person_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_person_person_light_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: person, person, light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_light_skin_tone_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_person_person_medium_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: person, person, medium skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_person_person_medium_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: person, person, medium skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_dark_skin_tone_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"),
            std::u16string_view(
              u"couple_with_heart_person_person_medium_dark_skin_tone_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, medium-dark skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_person_person_dark_skin_tone_light_skin_tone"),
        std::u16string_view(u"couple with heart: person, person, dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_person_person_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_person_person_dark_skin_tone_medium_skin_tone"),
        std::u16string_view(u"couple with heart: person, person, dark skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_person_person_dark_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: person, person, dark skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468"),
            std::u16string_view(u"couple_with_heart_woman_man"),
            std::u16string_view(u"couple with heart: woman, man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_woman_man_light_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_woman_man_light_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, light skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_woman_man_light_skin_tone_medium_skin_tone"),
        std::u16string_view(u"couple with heart: woman, man, light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_woman_man_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_woman_man_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_light_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-light skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"couple_with_heart_woman_man_medium_light_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_light_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_woman_man_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, man, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_light_skin_tone_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_skin_tone_light_skin_tone"),
        std::u16string_view(u"couple with heart: woman, man, medium skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_woman_man_medium_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_woman_man_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_dark_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_woman_man_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, man, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_dark_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-dark skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"couple_with_heart_woman_man_medium_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_man_medium_dark_skin_tone_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, medium-dark skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_woman_man_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_woman_man_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_woman_man_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_woman_man_dark_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, man, dark skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_woman_man_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, man, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468"),
            std::u16string_view(u"couple_with_heart_man_man"),
            std::u16string_view(u"couple with heart: man, man"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_man_man_light_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_man_man_light_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, light skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_man_man_light_skin_tone_medium_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_man_man_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_man_man_light_skin_tone_dark_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, light skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_man_man_medium_light_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, medium-light skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(u"couple_with_heart_man_man_medium_light_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_man_man_medium_light_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, medium-light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_man_man_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: man, man, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_man_man_medium_light_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, medium-light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_man_man_medium_skin_tone_light_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, medium skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_man_man_medium_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, medium skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_man_man_medium_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_man_man_medium_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, medium skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_man_man_medium_skin_tone_dark_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, medium skin tone, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_man_man_medium_dark_skin_tone_light_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, medium-dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_man_man_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: man, man, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_man_man_medium_dark_skin_tone_medium_skin_tone"),
        std::u16string_view(
          u"couple with heart: man, man, medium-dark skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
            std::u16string_view(u"couple_with_heart_man_man_medium_dark_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_man_man_medium_dark_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, medium-dark skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_man_man_dark_skin_tone_light_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, dark skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_man_man_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_man_man_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_man_man_dark_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(u"couple with heart: man, man, dark skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_man_man_dark_skin_tone"),
            std::u16string_view(u"couple with heart: man, man, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469"),
            std::u16string_view(u"couple_with_heart_woman_woman"),
            std::u16string_view(u"couple with heart: woman, woman"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(u"couple_with_heart_woman_woman_light_skin_tone"),
            std::u16string_view(u"couple with heart: woman, woman, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_light_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, light skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_woman_woman_light_skin_tone_medium_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, light skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_woman_woman_light_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, light skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_woman_light_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_light_skin_tone_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium-light skin tone, light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(u"couple_with_heart_woman_woman_medium_light_skin_tone"),
            std::u16string_view(u"couple with heart: woman, woman, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_light_skin_tone_medium_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium-light skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_light_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium-light skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_woman_medium_light_skin_tone_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, medium-light skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_woman_medium_skin_tone_light_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, medium skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(u"couple_with_heart_woman_woman_medium_skin_tone"),
            std::u16string_view(u"couple with heart: woman, woman, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_skin_tone_medium_dark_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium skin tone, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_woman_medium_skin_tone_dark_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, medium skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_woman_medium_dark_skin_tone_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, medium-dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_dark_skin_tone_medium_light_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium-dark skin tone, medium-light skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"),
            std::u16string_view(
              u"couple_with_heart_woman_woman_medium_dark_skin_tone_medium_skin_tone"),
            std::u16string_view(
              u"couple with heart: woman, woman, medium-dark skin tone, medium skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"),
            std::u16string_view(u"couple_with_heart_woman_woman_medium_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, woman, medium-dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"),
        std::u16string_view(u"couple_with_heart_woman_woman_medium_dark_skin_tone_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, medium-dark skin tone, dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"),
        std::u16string_view(u"couple_with_heart_woman_woman_dark_skin_tone_light_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, dark skin tone, light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"),
        std::u16string_view(u"couple_with_heart_woman_woman_dark_skin_tone_medium_light_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, dark skin tone, medium-light skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"),
        std::u16string_view(u"couple_with_heart_woman_woman_dark_skin_tone_medium_skin_tone"),
        std::u16string_view(u"couple with heart: woman, woman, dark skin tone, medium skin tone"),
        emoji::Emoji::Category::People},
      Emoji{
        std::u16string_view(
          u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"),
        std::u16string_view(u"couple_with_heart_woman_woman_dark_skin_tone_medium_dark_skin_tone"),
        std::u16string_view(
          u"couple with heart: woman, woman, dark skin tone, medium-dark skin tone"),
        emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"),
            std::u16string_view(u"couple_with_heart_woman_woman_dark_skin_tone"),
            std::u16string_view(u"couple with heart: woman, woman, dark skin tone"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F469\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_woman_boy"),
            std::u16string_view(u"family: man, woman, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F469\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_woman_girl"),
            std::u16string_view(u"family: man, woman, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F469\U0000200D\U0001F467\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_woman_girl_boy"),
            std::u16string_view(u"family: man, woman, girl, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F469\U0000200D\U0001F466\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_woman_boy_boy"),
            std::u16string_view(u"family: man, woman, boy, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F469\U0000200D\U0001F467\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_woman_girl_girl"),
            std::u16string_view(u"family: man, woman, girl, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F468\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_man_boy"),
            std::u16string_view(u"family: man, man, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F468\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_man_girl"),
            std::u16string_view(u"family: man, man, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F468\U0000200D\U0001F467\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_man_girl_boy"),
            std::u16string_view(u"family: man, man, girl, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F468\U0000200D\U0001F466\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_man_boy_boy"),
            std::u16string_view(u"family: man, man, boy, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F468\U0000200D\U0001F468\U0000200D\U0001F467\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_man_girl_girl"),
            std::u16string_view(u"family: man, man, girl, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F469\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_woman_boy"),
            std::u16string_view(u"family: woman, woman, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F469\U0000200D\U0001F467"),
            std::u16string_view(u"family_woman_woman_girl"),
            std::u16string_view(u"family: woman, woman, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0000200D\U0001F469\U0000200D\U0001F467\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_woman_girl_boy"),
            std::u16string_view(u"family: woman, woman, girl, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0000200D\U0001F469\U0000200D\U0001F466\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_woman_boy_boy"),
            std::u16string_view(u"family: woman, woman, boy, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F469\U0000200D\U0001F469\U0000200D\U0001F467\U0000200D\U0001F467"),
            std::u16string_view(u"family_woman_woman_girl_girl"),
            std::u16string_view(u"family: woman, woman, girl, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_boy"),
            std::u16string_view(u"family: man, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F466\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_boy_boy"),
            std::u16string_view(u"family: man, boy, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_girl"),
            std::u16string_view(u"family: man, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F467\U0000200D\U0001F466"),
            std::u16string_view(u"family_man_girl_boy"),
            std::u16string_view(u"family: man, girl, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F468\U0000200D\U0001F467\U0000200D\U0001F467"),
            std::u16string_view(u"family_man_girl_girl"),
            std::u16string_view(u"family: man, girl, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_boy"),
            std::u16string_view(u"family: woman, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F466\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_boy_boy"),
            std::u16string_view(u"family: woman, boy, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F467"),
            std::u16string_view(u"family_woman_girl"),
            std::u16string_view(u"family: woman, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F467\U0000200D\U0001F466"),
            std::u16string_view(u"family_woman_girl_boy"),
            std::u16string_view(u"family: woman, girl, boy"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F469\U0000200D\U0001F467\U0000200D\U0001F467"),
            std::u16string_view(u"family_woman_girl_girl"),
            std::u16string_view(u"family: woman, girl, girl"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F5E3\U0000FE0F"),
            std::u16string_view(u"speaking_head"),
            std::u16string_view(u"speaking head"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F464"),
            std::u16string_view(u"bust_in_silhouette"),
            std::u16string_view(u"bust in silhouette"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F465"),
            std::u16string_view(u"busts_in_silhouette"),
            std::u16string_view(u"busts in silhouette"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001FAC2"),
            std::u16string_view(u"people_hugging"),
            std::u16string_view(u"people hugging"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F46A"),
            std::u16string_view(u"family"),
            std::u16string_view(u"family"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9D1\U0000200D\U0001F9D2"),
            std::u16string_view(u"family_adult_adult_child"),
            std::u16string_view(u"family: adult, adult, child"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(
              u"\U0001F9D1\U0000200D\U0001F9D1\U0000200D\U0001F9D2\U0000200D\U0001F9D2"),
            std::u16string_view(u"family_adult_adult_child_child"),
            std::u16string_view(u"family: adult, adult, child, child"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9D2"),
            std::u16string_view(u"family_adult_child"),
            std::u16string_view(u"family: adult, child"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F9D1\U0000200D\U0001F9D2\U0000200D\U0001F9D2"),
            std::u16string_view(u"family_adult_child_child"),
            std::u16string_view(u"family: adult, child, child"),
            emoji::Emoji::Category::People},
      Emoji{std::u16string_view(u"\U0001F463"),
            std::u16string_view(u"footprints"),
            std::u16string_view(u"footprints"),
            emoji::Emoji::Category::People},
      // Nature
      Emoji{std::u16string_view(u"\U0001F435"),
            std::u16string_view(u"monkey"),
            std::u16string_view(u"monkey face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F412"),
            std::u16string_view(u"monkey"),
            std::u16string_view(u"monkey"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98D"),
            std::u16string_view(u"gorilla"),
            std::u16string_view(u"gorilla"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A7"),
            std::u16string_view(u"orangutan"),
            std::u16string_view(u"orangutan"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F436"),
            std::u16string_view(u"dog"),
            std::u16string_view(u"dog face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F415"),
            std::u16string_view(u"dog"),
            std::u16string_view(u"dog"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9AE"),
            std::u16string_view(u"guide_dog"),
            std::u16string_view(u"guide dog"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F415\U0000200D\U0001F9BA"),
            std::u16string_view(u"service_dog"),
            std::u16string_view(u"service dog"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F429"),
            std::u16string_view(u"poodle"),
            std::u16string_view(u"poodle"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43A"),
            std::u16string_view(u"wolf"),
            std::u16string_view(u"wolf"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98A"),
            std::u16string_view(u"fox"),
            std::u16string_view(u"fox"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F99D"),
            std::u16string_view(u"raccoon"),
            std::u16string_view(u"raccoon"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F431"),
            std::u16string_view(u"cat"),
            std::u16string_view(u"cat face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F408"),
            std::u16string_view(u"cat"),
            std::u16string_view(u"cat"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F408\U0000200D\U00002B1B"),
            std::u16string_view(u"black_cat"),
            std::u16string_view(u"black cat"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F981"),
            std::u16string_view(u"lion"),
            std::u16string_view(u"lion"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42F"),
            std::u16string_view(u"tiger"),
            std::u16string_view(u"tiger face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F405"),
            std::u16string_view(u"tiger"),
            std::u16string_view(u"tiger"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F406"),
            std::u16string_view(u"leopard"),
            std::u16string_view(u"leopard"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F434"),
            std::u16string_view(u"horse"),
            std::u16string_view(u"horse face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FACE"),
            std::u16string_view(u"moose"),
            std::u16string_view(u"moose"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FACF"),
            std::u16string_view(u"donkey"),
            std::u16string_view(u"donkey"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40E"),
            std::u16string_view(u"horse"),
            std::u16string_view(u"horse"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F984"),
            std::u16string_view(u"unicorn"),
            std::u16string_view(u"unicorn"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F993"),
            std::u16string_view(u"zebra"),
            std::u16string_view(u"zebra"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98C"),
            std::u16string_view(u"deer"),
            std::u16string_view(u"deer"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9AC"),
            std::u16string_view(u"bison"),
            std::u16string_view(u"bison"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42E"),
            std::u16string_view(u"cow"),
            std::u16string_view(u"cow face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F402"),
            std::u16string_view(u"ox"),
            std::u16string_view(u"ox"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F403"),
            std::u16string_view(u"water_buffalo"),
            std::u16string_view(u"water buffalo"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F404"),
            std::u16string_view(u"cow"),
            std::u16string_view(u"cow"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F437"),
            std::u16string_view(u"pig"),
            std::u16string_view(u"pig face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F416"),
            std::u16string_view(u"pig"),
            std::u16string_view(u"pig"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F417"),
            std::u16string_view(u"boar"),
            std::u16string_view(u"boar"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43D"),
            std::u16string_view(u"pig_nose"),
            std::u16string_view(u"pig nose"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40F"),
            std::u16string_view(u"ram"),
            std::u16string_view(u"ram"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F411"),
            std::u16string_view(u"ewe"),
            std::u16string_view(u"ewe"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F410"),
            std::u16string_view(u"goat"),
            std::u16string_view(u"goat"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42A"),
            std::u16string_view(u"camel"),
            std::u16string_view(u"camel"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42B"),
            std::u16string_view(u"two_hump_camel"),
            std::u16string_view(u"two-hump camel"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F999"),
            std::u16string_view(u"llama"),
            std::u16string_view(u"llama"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F992"),
            std::u16string_view(u"giraffe"),
            std::u16string_view(u"giraffe"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F418"),
            std::u16string_view(u"elephant"),
            std::u16string_view(u"elephant"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A3"),
            std::u16string_view(u"mammoth"),
            std::u16string_view(u"mammoth"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98F"),
            std::u16string_view(u"rhinoceros"),
            std::u16string_view(u"rhinoceros"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F99B"),
            std::u16string_view(u"hippopotamus"),
            std::u16string_view(u"hippopotamus"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42D"),
            std::u16string_view(u"mouse"),
            std::u16string_view(u"mouse face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F401"),
            std::u16string_view(u"mouse"),
            std::u16string_view(u"mouse"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F400"),
            std::u16string_view(u"rat"),
            std::u16string_view(u"rat"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F439"),
            std::u16string_view(u"hamster"),
            std::u16string_view(u"hamster"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F430"),
            std::u16string_view(u"rabbit"),
            std::u16string_view(u"rabbit face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F407"),
            std::u16string_view(u"rabbit"),
            std::u16string_view(u"rabbit"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43F\U0000FE0F"),
            std::u16string_view(u"chipmunk"),
            std::u16string_view(u"chipmunk"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9AB"),
            std::u16string_view(u"beaver"),
            std::u16string_view(u"beaver"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F994"),
            std::u16string_view(u"hedgehog"),
            std::u16string_view(u"hedgehog"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F987"),
            std::u16string_view(u"bat"),
            std::u16string_view(u"bat"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43B"),
            std::u16string_view(u"bear"),
            std::u16string_view(u"bear"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43B\U0000200D\U00002744\U0000FE0F"),
            std::u16string_view(u"polar_bear"),
            std::u16string_view(u"polar bear"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F428"),
            std::u16string_view(u"koala"),
            std::u16string_view(u"koala"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43C"),
            std::u16string_view(u"panda"),
            std::u16string_view(u"panda"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A5"),
            std::u16string_view(u"sloth"),
            std::u16string_view(u"sloth"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A6"),
            std::u16string_view(u"otter"),
            std::u16string_view(u"otter"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A8"),
            std::u16string_view(u"skunk"),
            std::u16string_view(u"skunk"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F998"),
            std::u16string_view(u"kangaroo"),
            std::u16string_view(u"kangaroo"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A1"),
            std::u16string_view(u"badger"),
            std::u16string_view(u"badger"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F43E"),
            std::u16string_view(u"paw_prints"),
            std::u16string_view(u"paw prints"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F983"),
            std::u16string_view(u"turkey"),
            std::u16string_view(u"turkey"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F414"),
            std::u16string_view(u"chicken"),
            std::u16string_view(u"chicken"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F413"),
            std::u16string_view(u"rooster"),
            std::u16string_view(u"rooster"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F423"),
            std::u16string_view(u"hatching_chick"),
            std::u16string_view(u"hatching chick"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F424"),
            std::u16string_view(u"baby_chick"),
            std::u16string_view(u"baby chick"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F425"),
            std::u16string_view(u"front_facing_baby_chick"),
            std::u16string_view(u"front-facing baby chick"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F426"),
            std::u16string_view(u"bird"),
            std::u16string_view(u"bird"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F427"),
            std::u16string_view(u"penguin"),
            std::u16string_view(u"penguin"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F54A\U0000FE0F"),
            std::u16string_view(u"dove"),
            std::u16string_view(u"dove"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F985"),
            std::u16string_view(u"eagle"),
            std::u16string_view(u"eagle"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F986"),
            std::u16string_view(u"duck"),
            std::u16string_view(u"duck"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A2"),
            std::u16string_view(u"swan"),
            std::u16string_view(u"swan"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F989"),
            std::u16string_view(u"owl"),
            std::u16string_view(u"owl"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A4"),
            std::u16string_view(u"dodo"),
            std::u16string_view(u"dodo"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB6"),
            std::u16string_view(u"feather"),
            std::u16string_view(u"feather"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A9"),
            std::u16string_view(u"flamingo"),
            std::u16string_view(u"flamingo"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F99A"),
            std::u16string_view(u"peacock"),
            std::u16string_view(u"peacock"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F99C"),
            std::u16string_view(u"parrot"),
            std::u16string_view(u"parrot"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FABD"),
            std::u16string_view(u"wing"),
            std::u16string_view(u"wing"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F426\U0000200D\U00002B1B"),
            std::u16string_view(u"black_bird"),
            std::u16string_view(u"black bird"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FABF"),
            std::u16string_view(u"goose"),
            std::u16string_view(u"goose"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F426\U0000200D\U0001F525"),
            std::u16string_view(u"phoenix"),
            std::u16string_view(u"phoenix"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F438"),
            std::u16string_view(u"frog"),
            std::u16string_view(u"frog"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40A"),
            std::u16string_view(u"crocodile"),
            std::u16string_view(u"crocodile"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F422"),
            std::u16string_view(u"turtle"),
            std::u16string_view(u"turtle"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98E"),
            std::u16string_view(u"lizard"),
            std::u16string_view(u"lizard"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40D"),
            std::u16string_view(u"snake"),
            std::u16string_view(u"snake"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F432"),
            std::u16string_view(u"dragon"),
            std::u16string_view(u"dragon face"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F409"),
            std::u16string_view(u"dragon"),
            std::u16string_view(u"dragon"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F995"),
            std::u16string_view(u"sauropod"),
            std::u16string_view(u"sauropod"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F996"),
            std::u16string_view(u"t_rex"),
            std::u16string_view(u"T-Rex"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F433"),
            std::u16string_view(u"spouting_whale"),
            std::u16string_view(u"spouting whale"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40B"),
            std::u16string_view(u"whale"),
            std::u16string_view(u"whale"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F42C"),
            std::u16string_view(u"dolphin"),
            std::u16string_view(u"dolphin"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9AD"),
            std::u16string_view(u"seal"),
            std::u16string_view(u"seal"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41F"),
            std::u16string_view(u"fish"),
            std::u16string_view(u"fish"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F420"),
            std::u16string_view(u"tropical_fish"),
            std::u16string_view(u"tropical fish"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F421"),
            std::u16string_view(u"blowfish"),
            std::u16string_view(u"blowfish"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F988"),
            std::u16string_view(u"shark"),
            std::u16string_view(u"shark"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F419"),
            std::u16string_view(u"octopus"),
            std::u16string_view(u"octopus"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41A"),
            std::u16string_view(u"spiral_shell"),
            std::u16string_view(u"spiral shell"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB8"),
            std::u16string_view(u"coral"),
            std::u16string_view(u"coral"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FABC"),
            std::u16string_view(u"jellyfish"),
            std::u16string_view(u"jellyfish"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F40C"),
            std::u16string_view(u"snail"),
            std::u16string_view(u"snail"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F98B"),
            std::u16string_view(u"butterfly"),
            std::u16string_view(u"butterfly"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41B"),
            std::u16string_view(u"bug"),
            std::u16string_view(u"bug"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41C"),
            std::u16string_view(u"ant"),
            std::u16string_view(u"ant"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41D"),
            std::u16string_view(u"honeybee"),
            std::u16string_view(u"honeybee"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB2"),
            std::u16string_view(u"beetle"),
            std::u16string_view(u"beetle"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F41E"),
            std::u16string_view(u"lady_beetle"),
            std::u16string_view(u"lady beetle"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F997"),
            std::u16string_view(u"cricket"),
            std::u16string_view(u"cricket"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB3"),
            std::u16string_view(u"cockroach"),
            std::u16string_view(u"cockroach"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F577\U0000FE0F"),
            std::u16string_view(u"spider"),
            std::u16string_view(u"spider"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F578\U0000FE0F"),
            std::u16string_view(u"spider_web"),
            std::u16string_view(u"spider web"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F982"),
            std::u16string_view(u"scorpion"),
            std::u16string_view(u"scorpion"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F99F"),
            std::u16string_view(u"mosquito"),
            std::u16string_view(u"mosquito"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB0"),
            std::u16string_view(u"fly"),
            std::u16string_view(u"fly"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB1"),
            std::u16string_view(u"worm"),
            std::u16string_view(u"worm"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F9A0"),
            std::u16string_view(u"microbe"),
            std::u16string_view(u"microbe"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F490"),
            std::u16string_view(u"bouquet"),
            std::u16string_view(u"bouquet"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F338"),
            std::u16string_view(u"cherry_blossom"),
            std::u16string_view(u"cherry blossom"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F4AE"),
            std::u16string_view(u"white_flower"),
            std::u16string_view(u"white flower"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB7"),
            std::u16string_view(u"lotus"),
            std::u16string_view(u"lotus"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F3F5\U0000FE0F"),
            std::u16string_view(u"rosette"),
            std::u16string_view(u"rosette"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F339"),
            std::u16string_view(u"rose"),
            std::u16string_view(u"rose"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F940"),
            std::u16string_view(u"wilted_flower"),
            std::u16string_view(u"wilted flower"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F33A"),
            std::u16string_view(u"hibiscus"),
            std::u16string_view(u"hibiscus"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F33B"),
            std::u16string_view(u"sunflower"),
            std::u16string_view(u"sunflower"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F33C"),
            std::u16string_view(u"blossom"),
            std::u16string_view(u"blossom"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F337"),
            std::u16string_view(u"tulip"),
            std::u16string_view(u"tulip"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FABB"),
            std::u16string_view(u"hyacinth"),
            std::u16string_view(u"hyacinth"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F331"),
            std::u16string_view(u"seedling"),
            std::u16string_view(u"seedling"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB4"),
            std::u16string_view(u"potted_plant"),
            std::u16string_view(u"potted plant"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F332"),
            std::u16string_view(u"evergreen_tree"),
            std::u16string_view(u"evergreen tree"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F333"),
            std::u16string_view(u"deciduous_tree"),
            std::u16string_view(u"deciduous tree"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F334"),
            std::u16string_view(u"palm_tree"),
            std::u16string_view(u"palm tree"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F335"),
            std::u16string_view(u"cactus"),
            std::u16string_view(u"cactus"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F33E"),
            std::u16string_view(u"sheaf_of_rice"),
            std::u16string_view(u"sheaf of rice"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F33F"),
            std::u16string_view(u"herb"),
            std::u16string_view(u"herb"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U00002618\U0000FE0F"),
            std::u16string_view(u"shamrock"),
            std::u16string_view(u"shamrock"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F340"),
            std::u16string_view(u"four_leaf_clover"),
            std::u16string_view(u"four leaf clover"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F341"),
            std::u16string_view(u"maple_leaf"),
            std::u16string_view(u"maple leaf"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F342"),
            std::u16string_view(u"fallen_leaf"),
            std::u16string_view(u"fallen leaf"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F343"),
            std::u16string_view(u"leaf_fluttering_in_wind"),
            std::u16string_view(u"leaf fluttering in wind"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FAB9"),
            std::u16string_view(u"empty_nest"),
            std::u16string_view(u"empty nest"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001FABA"),
            std::u16string_view(u"nest_with_eggs"),
            std::u16string_view(u"nest with eggs"),
            emoji::Emoji::Category::Nature},
      Emoji{std::u16string_view(u"\U0001F344"),
            std::u16string_view(u"mushroom"),
            std::u16string_view(u"mushroom"),
            emoji::Emoji::Category::Nature},
      // Food
      Emoji{std::u16string_view(u"\U0001F347"),
            std::u16string_view(u"grapes"),
            std::u16string_view(u"grapes"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F348"),
            std::u16string_view(u"melon"),
            std::u16string_view(u"melon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F349"),
            std::u16string_view(u"watermelon"),
            std::u16string_view(u"watermelon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34A"),
            std::u16string_view(u"tangerine"),
            std::u16string_view(u"tangerine"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34B"),
            std::u16string_view(u"lemon"),
            std::u16string_view(u"lemon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34B\U0000200D\U0001F7E9"),
            std::u16string_view(u"lime"),
            std::u16string_view(u"lime"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34C"),
            std::u16string_view(u"banana"),
            std::u16string_view(u"banana"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34D"),
            std::u16string_view(u"pineapple"),
            std::u16string_view(u"pineapple"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96D"),
            std::u16string_view(u"mango"),
            std::u16string_view(u"mango"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34E"),
            std::u16string_view(u"red_apple"),
            std::u16string_view(u"red apple"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F34F"),
            std::u16string_view(u"green_apple"),
            std::u16string_view(u"green apple"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F350"),
            std::u16string_view(u"pear"),
            std::u16string_view(u"pear"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F351"),
            std::u16string_view(u"peach"),
            std::u16string_view(u"peach"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F352"),
            std::u16string_view(u"cherries"),
            std::u16string_view(u"cherries"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F353"),
            std::u16string_view(u"strawberry"),
            std::u16string_view(u"strawberry"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD0"),
            std::u16string_view(u"blueberries"),
            std::u16string_view(u"blueberries"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95D"),
            std::u16string_view(u"kiwi_fruit"),
            std::u16string_view(u"kiwi fruit"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F345"),
            std::u16string_view(u"tomato"),
            std::u16string_view(u"tomato"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD2"),
            std::u16string_view(u"olive"),
            std::u16string_view(u"olive"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F965"),
            std::u16string_view(u"coconut"),
            std::u16string_view(u"coconut"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F951"),
            std::u16string_view(u"avocado"),
            std::u16string_view(u"avocado"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F346"),
            std::u16string_view(u"eggplant"),
            std::u16string_view(u"eggplant"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F954"),
            std::u16string_view(u"potato"),
            std::u16string_view(u"potato"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F955"),
            std::u16string_view(u"carrot"),
            std::u16string_view(u"carrot"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F33D"),
            std::u16string_view(u"ear_of_corn"),
            std::u16string_view(u"ear of corn"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F336\U0000FE0F"),
            std::u16string_view(u"hot_pepper"),
            std::u16string_view(u"hot pepper"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD1"),
            std::u16string_view(u"bell_pepper"),
            std::u16string_view(u"bell pepper"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F952"),
            std::u16string_view(u"cucumber"),
            std::u16string_view(u"cucumber"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96C"),
            std::u16string_view(u"leafy_green"),
            std::u16string_view(u"leafy green"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F966"),
            std::u16string_view(u"broccoli"),
            std::u16string_view(u"broccoli"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C4"),
            std::u16string_view(u"garlic"),
            std::u16string_view(u"garlic"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C5"),
            std::u16string_view(u"onion"),
            std::u16string_view(u"onion"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95C"),
            std::u16string_view(u"peanuts"),
            std::u16string_view(u"peanuts"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD8"),
            std::u16string_view(u"beans"),
            std::u16string_view(u"beans"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F330"),
            std::u16string_view(u"chestnut"),
            std::u16string_view(u"chestnut"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FADA"),
            std::u16string_view(u"ginger_root"),
            std::u16string_view(u"ginger root"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FADB"),
            std::u16string_view(u"pea_pod"),
            std::u16string_view(u"pea pod"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F344\U0000200D\U0001F7EB"),
            std::u16string_view(u"brown_mushroom"),
            std::u16string_view(u"brown mushroom"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35E"),
            std::u16string_view(u"bread"),
            std::u16string_view(u"bread"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F950"),
            std::u16string_view(u"croissant"),
            std::u16string_view(u"croissant"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F956"),
            std::u16string_view(u"baguette_bread"),
            std::u16string_view(u"baguette bread"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD3"),
            std::u16string_view(u"flatbread"),
            std::u16string_view(u"flatbread"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F968"),
            std::u16string_view(u"pretzel"),
            std::u16string_view(u"pretzel"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96F"),
            std::u16string_view(u"bagel"),
            std::u16string_view(u"bagel"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95E"),
            std::u16string_view(u"pancakes"),
            std::u16string_view(u"pancakes"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C7"),
            std::u16string_view(u"waffle"),
            std::u16string_view(u"waffle"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C0"),
            std::u16string_view(u"cheese_wedge"),
            std::u16string_view(u"cheese wedge"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F356"),
            std::u16string_view(u"meat_on_bone"),
            std::u16string_view(u"meat on bone"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F357"),
            std::u16string_view(u"poultry_leg"),
            std::u16string_view(u"poultry leg"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F969"),
            std::u16string_view(u"cut_of_meat"),
            std::u16string_view(u"cut of meat"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F953"),
            std::u16string_view(u"bacon"),
            std::u16string_view(u"bacon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F354"),
            std::u16string_view(u"hamburger"),
            std::u16string_view(u"hamburger"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35F"),
            std::u16string_view(u"french_fries"),
            std::u16string_view(u"french fries"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F355"),
            std::u16string_view(u"pizza"),
            std::u16string_view(u"pizza"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F32D"),
            std::u16string_view(u"hot_dog"),
            std::u16string_view(u"hot dog"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96A"),
            std::u16string_view(u"sandwich"),
            std::u16string_view(u"sandwich"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F32E"),
            std::u16string_view(u"taco"),
            std::u16string_view(u"taco"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F32F"),
            std::u16string_view(u"burrito"),
            std::u16string_view(u"burrito"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD4"),
            std::u16string_view(u"tamale"),
            std::u16string_view(u"tamale"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F959"),
            std::u16string_view(u"stuffed_flatbread"),
            std::u16string_view(u"stuffed flatbread"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C6"),
            std::u16string_view(u"falafel"),
            std::u16string_view(u"falafel"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95A"),
            std::u16string_view(u"egg"),
            std::u16string_view(u"egg"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F373"),
            std::u16string_view(u"cooking"),
            std::u16string_view(u"cooking"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F958"),
            std::u16string_view(u"shallow_pan_of_food"),
            std::u16string_view(u"shallow pan of food"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F372"),
            std::u16string_view(u"pot_of_food"),
            std::u16string_view(u"pot of food"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD5"),
            std::u16string_view(u"fondue"),
            std::u16string_view(u"fondue"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F963"),
            std::u16string_view(u"bowl_with_spoon"),
            std::u16string_view(u"bowl with spoon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F957"),
            std::u16string_view(u"green_salad"),
            std::u16string_view(u"green salad"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37F"),
            std::u16string_view(u"popcorn"),
            std::u16string_view(u"popcorn"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C8"),
            std::u16string_view(u"butter"),
            std::u16string_view(u"butter"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C2"),
            std::u16string_view(u"salt"),
            std::u16string_view(u"salt"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96B"),
            std::u16string_view(u"canned_food"),
            std::u16string_view(u"canned food"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F371"),
            std::u16string_view(u"bento_box"),
            std::u16string_view(u"bento box"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F358"),
            std::u16string_view(u"rice_cracker"),
            std::u16string_view(u"rice cracker"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F359"),
            std::u16string_view(u"rice_ball"),
            std::u16string_view(u"rice ball"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35A"),
            std::u16string_view(u"cooked_rice"),
            std::u16string_view(u"cooked rice"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35B"),
            std::u16string_view(u"curry_rice"),
            std::u16string_view(u"curry rice"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35C"),
            std::u16string_view(u"steaming_bowl"),
            std::u16string_view(u"steaming bowl"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F35D"),
            std::u16string_view(u"spaghetti"),
            std::u16string_view(u"spaghetti"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F360"),
            std::u16string_view(u"roasted_sweet_potato"),
            std::u16string_view(u"roasted sweet potato"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F362"),
            std::u16string_view(u"oden"),
            std::u16string_view(u"oden"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F363"),
            std::u16string_view(u"sushi"),
            std::u16string_view(u"sushi"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F364"),
            std::u16string_view(u"fried_shrimp"),
            std::u16string_view(u"fried shrimp"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F365"),
            std::u16string_view(u"fish_cake_with_swirl"),
            std::u16string_view(u"fish cake with swirl"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F96E"),
            std::u16string_view(u"moon_cake"),
            std::u16string_view(u"moon cake"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F361"),
            std::u16string_view(u"dango"),
            std::u16string_view(u"dango"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95F"),
            std::u16string_view(u"dumpling"),
            std::u16string_view(u"dumpling"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F960"),
            std::u16string_view(u"fortune_cookie"),
            std::u16string_view(u"fortune cookie"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F961"),
            std::u16string_view(u"takeout_box"),
            std::u16string_view(u"takeout box"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F980"),
            std::u16string_view(u"crab"),
            std::u16string_view(u"crab"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F99E"),
            std::u16string_view(u"lobster"),
            std::u16string_view(u"lobster"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F990"),
            std::u16string_view(u"shrimp"),
            std::u16string_view(u"shrimp"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F991"),
            std::u16string_view(u"squid"),
            std::u16string_view(u"squid"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9AA"),
            std::u16string_view(u"oyster"),
            std::u16string_view(u"oyster"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F366"),
            std::u16string_view(u"soft_ice_cream"),
            std::u16string_view(u"soft ice cream"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F367"),
            std::u16string_view(u"shaved_ice"),
            std::u16string_view(u"shaved ice"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F368"),
            std::u16string_view(u"ice_cream"),
            std::u16string_view(u"ice cream"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F369"),
            std::u16string_view(u"doughnut"),
            std::u16string_view(u"doughnut"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36A"),
            std::u16string_view(u"cookie"),
            std::u16string_view(u"cookie"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F382"),
            std::u16string_view(u"birthday_cake"),
            std::u16string_view(u"birthday cake"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F370"),
            std::u16string_view(u"shortcake"),
            std::u16string_view(u"shortcake"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C1"),
            std::u16string_view(u"cupcake"),
            std::u16string_view(u"cupcake"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F967"),
            std::u16string_view(u"pie"),
            std::u16string_view(u"pie"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36B"),
            std::u16string_view(u"chocolate_bar"),
            std::u16string_view(u"chocolate bar"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36C"),
            std::u16string_view(u"candy"),
            std::u16string_view(u"candy"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36D"),
            std::u16string_view(u"lollipop"),
            std::u16string_view(u"lollipop"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36E"),
            std::u16string_view(u"custard"),
            std::u16string_view(u"custard"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F36F"),
            std::u16string_view(u"honey_pot"),
            std::u16string_view(u"honey pot"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37C"),
            std::u16string_view(u"baby_bottle"),
            std::u16string_view(u"baby bottle"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F95B"),
            std::u16string_view(u"glass_of_milk"),
            std::u16string_view(u"glass of milk"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U00002615"),
            std::u16string_view(u"hot_beverage"),
            std::u16string_view(u"hot beverage"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD6"),
            std::u16string_view(u"teapot"),
            std::u16string_view(u"teapot"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F375"),
            std::u16string_view(u"teacup_without_handle"),
            std::u16string_view(u"teacup without handle"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F376"),
            std::u16string_view(u"sake"),
            std::u16string_view(u"sake"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37E"),
            std::u16string_view(u"bottle_with_popping_cork"),
            std::u16string_view(u"bottle with popping cork"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F377"),
            std::u16string_view(u"wine_glass"),
            std::u16string_view(u"wine glass"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F378"),
            std::u16string_view(u"cocktail_glass"),
            std::u16string_view(u"cocktail glass"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F379"),
            std::u16string_view(u"tropical_drink"),
            std::u16string_view(u"tropical drink"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37A"),
            std::u16string_view(u"beer_mug"),
            std::u16string_view(u"beer mug"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37B"),
            std::u16string_view(u"clinking_beer_mugs"),
            std::u16string_view(u"clinking beer mugs"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F942"),
            std::u16string_view(u"clinking_glasses"),
            std::u16string_view(u"clinking glasses"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F943"),
            std::u16string_view(u"tumbler_glass"),
            std::u16string_view(u"tumbler glass"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD7"),
            std::u16string_view(u"pouring_liquid"),
            std::u16string_view(u"pouring liquid"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F964"),
            std::u16string_view(u"cup_with_straw"),
            std::u16string_view(u"cup with straw"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9CB"),
            std::u16string_view(u"bubble_tea"),
            std::u16string_view(u"bubble tea"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C3"),
            std::u16string_view(u"beverage_box"),
            std::u16string_view(u"beverage box"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9C9"),
            std::u16string_view(u"mate"),
            std::u16string_view(u"mate"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F9CA"),
            std::u16string_view(u"ice"),
            std::u16string_view(u"ice"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F962"),
            std::u16string_view(u"chopsticks"),
            std::u16string_view(u"chopsticks"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F37D\U0000FE0F"),
            std::u16string_view(u"fork_and_knife_with_plate"),
            std::u16string_view(u"fork and knife with plate"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F374"),
            std::u16string_view(u"fork_and_knife"),
            std::u16string_view(u"fork and knife"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F944"),
            std::u16string_view(u"spoon"),
            std::u16string_view(u"spoon"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F52A"),
            std::u16string_view(u"kitchen_knife"),
            std::u16string_view(u"kitchen knife"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001FAD9"),
            std::u16string_view(u"jar"),
            std::u16string_view(u"jar"),
            emoji::Emoji::Category::Food},
      Emoji{std::u16string_view(u"\U0001F3FA"),
            std::u16string_view(u"amphora"),
            std::u16string_view(u"amphora"),
            emoji::Emoji::Category::Food},
      // Activity
      Emoji{std::u16string_view(u"\U0001F383"),
            std::u16string_view(u"jack_o_lantern"),
            std::u16string_view(u"jack-o-lantern"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F384"),
            std::u16string_view(u"christmas_tree"),
            std::u16string_view(u"Christmas tree"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F386"),
            std::u16string_view(u"fireworks"),
            std::u16string_view(u"fireworks"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F387"),
            std::u16string_view(u"sparkler"),
            std::u16string_view(u"sparkler"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9E8"),
            std::u16string_view(u"firecracker"),
            std::u16string_view(u"firecracker"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U00002728"),
            std::u16string_view(u"sparkles"),
            std::u16string_view(u"sparkles"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F388"),
            std::u16string_view(u"balloon"),
            std::u16string_view(u"balloon"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F389"),
            std::u16string_view(u"tada"),
            std::u16string_view(u"party popper"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F38A"),
            std::u16string_view(u"confetti_ball"),
            std::u16string_view(u"confetti ball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F38B"),
            std::u16string_view(u"tanabata_tree"),
            std::u16string_view(u"tanabata tree"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F38D"),
            std::u16string_view(u"pine_decoration"),
            std::u16string_view(u"pine decoration"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F38E"),
            std::u16string_view(u"japanese_dolls"),
            std::u16string_view(u"Japanese dolls"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F38F"),
            std::u16string_view(u"carp_streamer"),
            std::u16string_view(u"carp streamer"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F390"),
            std::u16string_view(u"wind_chime"),
            std::u16string_view(u"wind chime"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F391"),
            std::u16string_view(u"moon_viewing_ceremony"),
            std::u16string_view(u"moon viewing ceremony"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9E7"),
            std::u16string_view(u"red_envelope"),
            std::u16string_view(u"red envelope"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F380"),
            std::u16string_view(u"ribbon"),
            std::u16string_view(u"ribbon"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F381"),
            std::u16string_view(u"wrapped_gift"),
            std::u16string_view(u"wrapped gift"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F397\U0000FE0F"),
            std::u16string_view(u"reminder_ribbon"),
            std::u16string_view(u"reminder ribbon"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F39F\U0000FE0F"),
            std::u16string_view(u"admission_tickets"),
            std::u16string_view(u"admission tickets"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3AB"),
            std::u16string_view(u"ticket"),
            std::u16string_view(u"ticket"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F396\U0000FE0F"),
            std::u16string_view(u"military_medal"),
            std::u16string_view(u"military medal"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3C6"),
            std::u16string_view(u"trophy"),
            std::u16string_view(u"trophy"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3C5"),
            std::u16string_view(u"sports_medal"),
            std::u16string_view(u"sports medal"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F947"),
            std::u16string_view(u"1st_place_medal"),
            std::u16string_view(u"1st place medal"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F948"),
            std::u16string_view(u"2nd_place_medal"),
            std::u16string_view(u"2nd place medal"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F949"),
            std::u16string_view(u"3rd_place_medal"),
            std::u16string_view(u"3rd place medal"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U000026BD"),
            std::u16string_view(u"soccer_ball"),
            std::u16string_view(u"soccer ball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U000026BE"),
            std::u16string_view(u"baseball"),
            std::u16string_view(u"baseball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94E"),
            std::u16string_view(u"softball"),
            std::u16string_view(u"softball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3C0"),
            std::u16string_view(u"basketball"),
            std::u16string_view(u"basketball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3D0"),
            std::u16string_view(u"volleyball"),
            std::u16string_view(u"volleyball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3C8"),
            std::u16string_view(u"american_football"),
            std::u16string_view(u"american football"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3C9"),
            std::u16string_view(u"rugby_football"),
            std::u16string_view(u"rugby football"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3BE"),
            std::u16string_view(u"tennis"),
            std::u16string_view(u"tennis"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94F"),
            std::u16string_view(u"flying_disc"),
            std::u16string_view(u"flying disc"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3B3"),
            std::u16string_view(u"bowling"),
            std::u16string_view(u"bowling"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3CF"),
            std::u16string_view(u"cricket_game"),
            std::u16string_view(u"cricket game"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3D1"),
            std::u16string_view(u"field_hockey"),
            std::u16string_view(u"field hockey"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3D2"),
            std::u16string_view(u"ice_hockey"),
            std::u16string_view(u"ice hockey"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94D"),
            std::u16string_view(u"lacrosse"),
            std::u16string_view(u"lacrosse"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3D3"),
            std::u16string_view(u"ping_pong"),
            std::u16string_view(u"ping pong"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3F8"),
            std::u16string_view(u"badminton"),
            std::u16string_view(u"badminton"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94A"),
            std::u16string_view(u"boxing_glove"),
            std::u16string_view(u"boxing glove"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94B"),
            std::u16string_view(u"martial_arts_uniform"),
            std::u16string_view(u"martial arts uniform"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F945"),
            std::u16string_view(u"goal_net"),
            std::u16string_view(u"goal net"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U000026F3"),
            std::u16string_view(u"flag_in_hole"),
            std::u16string_view(u"flag in hole"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U000026F8\U0000FE0F"),
            std::u16string_view(u"ice_skate"),
            std::u16string_view(u"ice skate"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3A3"),
            std::u16string_view(u"fishing_pole"),
            std::u16string_view(u"fishing pole"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F93F"),
            std::u16string_view(u"diving_mask"),
            std::u16string_view(u"diving mask"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3BD"),
            std::u16string_view(u"running_shirt"),
            std::u16string_view(u"running shirt"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3BF"),
            std::u16string_view(u"skis"),
            std::u16string_view(u"skis"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F6F7"),
            std::u16string_view(u"sled"),
            std::u16string_view(u"sled"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F94C"),
            std::u16string_view(u"curling_stone"),
            std::u16string_view(u"curling stone"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3AF"),
            std::u16string_view(u"bullseye"),
            std::u16string_view(u"bullseye"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FA80"),
            std::u16string_view(u"yo_yo"),
            std::u16string_view(u"yo-yo"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FA81"),
            std::u16string_view(u"kite"),
            std::u16string_view(u"kite"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F52B"),
            std::u16string_view(u"water_pistol"),
            std::u16string_view(u"water pistol"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3B1"),
            std::u16string_view(u"pool_8_ball"),
            std::u16string_view(u"pool 8 ball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F52E"),
            std::u16string_view(u"crystal_ball"),
            std::u16string_view(u"crystal ball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FA84"),
            std::u16string_view(u"magic_wand"),
            std::u16string_view(u"magic wand"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3AE"),
            std::u16string_view(u"video_game"),
            std::u16string_view(u"video game"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F579\U0000FE0F"),
            std::u16string_view(u"joystick"),
            std::u16string_view(u"joystick"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3B0"),
            std::u16string_view(u"slot_machine"),
            std::u16string_view(u"slot machine"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3B2"),
            std::u16string_view(u"game_die"),
            std::u16string_view(u"game die"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9E9"),
            std::u16string_view(u"puzzle_piece"),
            std::u16string_view(u"puzzle piece"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9F8"),
            std::u16string_view(u"teddy_bear"),
            std::u16string_view(u"teddy bear"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FA85"),
            std::u16string_view(u"pinata"),
            std::u16string_view(u"piñata"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FAA9"),
            std::u16string_view(u"mirror_ball"),
            std::u16string_view(u"mirror ball"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FA86"),
            std::u16string_view(u"nesting_dolls"),
            std::u16string_view(u"nesting dolls"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U00002660\U0000FE0F"),
            std::u16string_view(u"spade_suit"),
            std::u16string_view(u"spade suit"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U00002665\U0000FE0F"),
            std::u16string_view(u"heart_suit"),
            std::u16string_view(u"heart suit"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U00002666\U0000FE0F"),
            std::u16string_view(u"diamond_suit"),
            std::u16string_view(u"diamond suit"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U00002663\U0000FE0F"),
            std::u16string_view(u"club_suit"),
            std::u16string_view(u"club suit"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0000265F\U0000FE0F"),
            std::u16string_view(u"chess_pawn"),
            std::u16string_view(u"chess pawn"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F0CF"),
            std::u16string_view(u"joker"),
            std::u16string_view(u"joker"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F004"),
            std::u16string_view(u"mahjong_red_dragon"),
            std::u16string_view(u"mahjong red dragon"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3B4"),
            std::u16string_view(u"flower_playing_cards"),
            std::u16string_view(u"flower playing cards"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3AD"),
            std::u16string_view(u"performing_arts"),
            std::u16string_view(u"performing arts"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F5BC\U0000FE0F"),
            std::u16string_view(u"framed_picture"),
            std::u16string_view(u"framed picture"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F3A8"),
            std::u16string_view(u"artist_palette"),
            std::u16string_view(u"artist palette"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9F5"),
            std::u16string_view(u"thread"),
            std::u16string_view(u"thread"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FAA1"),
            std::u16string_view(u"sewing_needle"),
            std::u16string_view(u"sewing needle"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001F9F6"),
            std::u16string_view(u"yarn"),
            std::u16string_view(u"yarn"),
            emoji::Emoji::Category::Activity},
      Emoji{std::u16string_view(u"\U0001FAA2"),
            std::u16string_view(u"knot"),
            std::u16string_view(u"knot"),
            emoji::Emoji::Category::Activity},
      // Travel
      Emoji{std::u16string_view(u"\U0001F30D"),
            std::u16string_view(u"globe_showing_europe_africa"),
            std::u16string_view(u"globe showing Europe-Africa"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F30E"),
            std::u16string_view(u"globe_showing_americas"),
            std::u16string_view(u"globe showing Americas"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F30F"),
            std::u16string_view(u"globe_showing_asia_australia"),
            std::u16string_view(u"globe showing Asia-Australia"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F310"),
            std::u16string_view(u"globe_with_meridians"),
            std::u16string_view(u"globe with meridians"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F5FA\U0000FE0F"),
            std::u16string_view(u"world_map"),
            std::u16string_view(u"world map"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F5FE"),
            std::u16string_view(u"map_of_japan"),
            std::u16string_view(u"map of Japan"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F9ED"),
            std::u16string_view(u"compass"),
            std::u16string_view(u"compass"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D4\U0000FE0F"),
            std::u16string_view(u"snow_capped_mountain"),
            std::u16string_view(u"snow-capped mountain"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026F0\U0000FE0F"),
            std::u16string_view(u"mountain"),
            std::u16string_view(u"mountain"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F30B"),
            std::u16string_view(u"volcano"),
            std::u16string_view(u"volcano"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F5FB"),
            std::u16string_view(u"mount_fuji"),
            std::u16string_view(u"mount fuji"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D5\U0000FE0F"),
            std::u16string_view(u"camping"),
            std::u16string_view(u"camping"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D6\U0000FE0F"),
            std::u16string_view(u"beach_with_umbrella"),
            std::u16string_view(u"beach with umbrella"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DC\U0000FE0F"),
            std::u16string_view(u"desert"),
            std::u16string_view(u"desert"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DD\U0000FE0F"),
            std::u16string_view(u"desert_island"),
            std::u16string_view(u"desert island"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DE\U0000FE0F"),
            std::u16string_view(u"national_park"),
            std::u16string_view(u"national park"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DF\U0000FE0F"),
            std::u16string_view(u"stadium"),
            std::u16string_view(u"stadium"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DB\U0000FE0F"),
            std::u16string_view(u"classical_building"),
            std::u16string_view(u"classical building"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D7\U0000FE0F"),
            std::u16string_view(u"building_construction"),
            std::u16string_view(u"building construction"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F9F1"),
            std::u16string_view(u"brick"),
            std::u16string_view(u"brick"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001FAA8"),
            std::u16string_view(u"rock"),
            std::u16string_view(u"rock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001FAB5"),
            std::u16string_view(u"wood"),
            std::u16string_view(u"wood"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6D6"),
            std::u16string_view(u"hut"),
            std::u16string_view(u"hut"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D8\U0000FE0F"),
            std::u16string_view(u"houses"),
            std::u16string_view(u"houses"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3DA\U0000FE0F"),
            std::u16string_view(u"derelict_house"),
            std::u16string_view(u"derelict house"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E0"),
            std::u16string_view(u"house"),
            std::u16string_view(u"house"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E1"),
            std::u16string_view(u"house_with_garden"),
            std::u16string_view(u"house with garden"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E2"),
            std::u16string_view(u"office_building"),
            std::u16string_view(u"office building"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E3"),
            std::u16string_view(u"japanese_post_office"),
            std::u16string_view(u"Japanese post office"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E4"),
            std::u16string_view(u"post_office"),
            std::u16string_view(u"post office"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E5"),
            std::u16string_view(u"hospital"),
            std::u16string_view(u"hospital"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E6"),
            std::u16string_view(u"bank"),
            std::u16string_view(u"bank"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E8"),
            std::u16string_view(u"hotel"),
            std::u16string_view(u"hotel"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3E9"),
            std::u16string_view(u"love_hotel"),
            std::u16string_view(u"love hotel"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3EA"),
            std::u16string_view(u"convenience_store"),
            std::u16string_view(u"convenience store"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3EB"),
            std::u16string_view(u"school"),
            std::u16string_view(u"school"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3EC"),
            std::u16string_view(u"department_store"),
            std::u16string_view(u"department store"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3ED"),
            std::u16string_view(u"factory"),
            std::u16string_view(u"factory"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3EF"),
            std::u16string_view(u"japanese_castle"),
            std::u16string_view(u"Japanese castle"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3F0"),
            std::u16string_view(u"castle"),
            std::u16string_view(u"castle"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F492"),
            std::u16string_view(u"wedding"),
            std::u16string_view(u"wedding"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F5FC"),
            std::u16string_view(u"tokyo_tower"),
            std::u16string_view(u"Tokyo tower"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F5FD"),
            std::u16string_view(u"statue_of_liberty"),
            std::u16string_view(u"Statue of Liberty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026EA"),
            std::u16string_view(u"church"),
            std::u16string_view(u"church"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F54C"),
            std::u16string_view(u"mosque"),
            std::u16string_view(u"mosque"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6D5"),
            std::u16string_view(u"hindu_temple"),
            std::u16string_view(u"hindu temple"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F54D"),
            std::u16string_view(u"synagogue"),
            std::u16string_view(u"synagogue"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026E9\U0000FE0F"),
            std::u16string_view(u"shinto_shrine"),
            std::u16string_view(u"shinto shrine"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F54B"),
            std::u16string_view(u"kaaba"),
            std::u16string_view(u"kaaba"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026F2"),
            std::u16string_view(u"fountain"),
            std::u16string_view(u"fountain"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026FA"),
            std::u16string_view(u"tent"),
            std::u16string_view(u"tent"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F301"),
            std::u16string_view(u"foggy"),
            std::u16string_view(u"foggy"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F303"),
            std::u16string_view(u"night_with_stars"),
            std::u16string_view(u"night with stars"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3D9\U0000FE0F"),
            std::u16string_view(u"cityscape"),
            std::u16string_view(u"cityscape"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F304"),
            std::u16string_view(u"sunrise_over_mountains"),
            std::u16string_view(u"sunrise over mountains"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F305"),
            std::u16string_view(u"sunrise"),
            std::u16string_view(u"sunrise"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F306"),
            std::u16string_view(u"cityscape_at_dusk"),
            std::u16string_view(u"cityscape at dusk"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F307"),
            std::u16string_view(u"sunset"),
            std::u16string_view(u"sunset"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F309"),
            std::u16string_view(u"bridge_at_night"),
            std::u16string_view(u"bridge at night"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002668\U0000FE0F"),
            std::u16string_view(u"hot_springs"),
            std::u16string_view(u"hot springs"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3A0"),
            std::u16string_view(u"carousel_horse"),
            std::u16string_view(u"carousel horse"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6DD"),
            std::u16string_view(u"playground_slide"),
            std::u16string_view(u"playground slide"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3A1"),
            std::u16string_view(u"ferris_wheel"),
            std::u16string_view(u"ferris wheel"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3A2"),
            std::u16string_view(u"roller_coaster"),
            std::u16string_view(u"roller coaster"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F488"),
            std::u16string_view(u"barber_pole"),
            std::u16string_view(u"barber pole"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3AA"),
            std::u16string_view(u"circus_tent"),
            std::u16string_view(u"circus tent"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F682"),
            std::u16string_view(u"locomotive"),
            std::u16string_view(u"locomotive"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F683"),
            std::u16string_view(u"railway_car"),
            std::u16string_view(u"railway car"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F684"),
            std::u16string_view(u"high_speed_train"),
            std::u16string_view(u"high-speed train"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F685"),
            std::u16string_view(u"bullet_train"),
            std::u16string_view(u"bullet train"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F686"),
            std::u16string_view(u"train"),
            std::u16string_view(u"train"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F687"),
            std::u16string_view(u"metro"),
            std::u16string_view(u"metro"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F688"),
            std::u16string_view(u"light_rail"),
            std::u16string_view(u"light rail"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F689"),
            std::u16string_view(u"station"),
            std::u16string_view(u"station"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68A"),
            std::u16string_view(u"tram"),
            std::u16string_view(u"tram"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69D"),
            std::u16string_view(u"monorail"),
            std::u16string_view(u"monorail"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69E"),
            std::u16string_view(u"mountain_railway"),
            std::u16string_view(u"mountain railway"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68B"),
            std::u16string_view(u"tram_car"),
            std::u16string_view(u"tram car"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68C"),
            std::u16string_view(u"bus"),
            std::u16string_view(u"bus"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68D"),
            std::u16string_view(u"oncoming_bus"),
            std::u16string_view(u"oncoming bus"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68E"),
            std::u16string_view(u"trolleybus"),
            std::u16string_view(u"trolleybus"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F690"),
            std::u16string_view(u"minibus"),
            std::u16string_view(u"minibus"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F691"),
            std::u16string_view(u"ambulance"),
            std::u16string_view(u"ambulance"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F692"),
            std::u16string_view(u"fire_engine"),
            std::u16string_view(u"fire engine"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F693"),
            std::u16string_view(u"police_car"),
            std::u16string_view(u"police car"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F694"),
            std::u16string_view(u"oncoming_police_car"),
            std::u16string_view(u"oncoming police car"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F695"),
            std::u16string_view(u"taxi"),
            std::u16string_view(u"taxi"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F696"),
            std::u16string_view(u"oncoming_taxi"),
            std::u16string_view(u"oncoming taxi"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F697"),
            std::u16string_view(u"automobile"),
            std::u16string_view(u"automobile"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F698"),
            std::u16string_view(u"oncoming_automobile"),
            std::u16string_view(u"oncoming automobile"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F699"),
            std::u16string_view(u"sport_utility_vehicle"),
            std::u16string_view(u"sport utility vehicle"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6FB"),
            std::u16string_view(u"pickup_truck"),
            std::u16string_view(u"pickup truck"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69A"),
            std::u16string_view(u"delivery_truck"),
            std::u16string_view(u"delivery truck"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69B"),
            std::u16string_view(u"articulated_lorry"),
            std::u16string_view(u"articulated lorry"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69C"),
            std::u16string_view(u"tractor"),
            std::u16string_view(u"tractor"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3CE\U0000FE0F"),
            std::u16string_view(u"racing_car"),
            std::u16string_view(u"racing car"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F3CD\U0000FE0F"),
            std::u16string_view(u"motorcycle"),
            std::u16string_view(u"motorcycle"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F5"),
            std::u16string_view(u"motor_scooter"),
            std::u16string_view(u"motor scooter"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F9BD"),
            std::u16string_view(u"manual_wheelchair"),
            std::u16string_view(u"manual wheelchair"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F9BC"),
            std::u16string_view(u"motorized_wheelchair"),
            std::u16string_view(u"motorized wheelchair"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6FA"),
            std::u16string_view(u"auto_rickshaw"),
            std::u16string_view(u"auto rickshaw"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6B2"),
            std::u16string_view(u"bicycle"),
            std::u16string_view(u"bicycle"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F4"),
            std::u16string_view(u"kick_scooter"),
            std::u16string_view(u"kick scooter"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F9"),
            std::u16string_view(u"skateboard"),
            std::u16string_view(u"skateboard"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6FC"),
            std::u16string_view(u"roller_skate"),
            std::u16string_view(u"roller skate"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F68F"),
            std::u16string_view(u"bus_stop"),
            std::u16string_view(u"bus stop"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6E3\U0000FE0F"),
            std::u16string_view(u"motorway"),
            std::u16string_view(u"motorway"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6E4\U0000FE0F"),
            std::u16string_view(u"railway_track"),
            std::u16string_view(u"railway track"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6E2\U0000FE0F"),
            std::u16string_view(u"oil_drum"),
            std::u16string_view(u"oil drum"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026FD"),
            std::u16string_view(u"fuel_pump"),
            std::u16string_view(u"fuel pump"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6DE"),
            std::u16string_view(u"wheel"),
            std::u16string_view(u"wheel"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A8"),
            std::u16string_view(u"police_car_light"),
            std::u16string_view(u"police car light"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A5"),
            std::u16string_view(u"horizontal_traffic_light"),
            std::u16string_view(u"horizontal traffic light"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A6"),
            std::u16string_view(u"vertical_traffic_light"),
            std::u16string_view(u"vertical traffic light"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6D1"),
            std::u16string_view(u"stop_sign"),
            std::u16string_view(u"stop sign"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A7"),
            std::u16string_view(u"construction"),
            std::u16string_view(u"construction"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002693"),
            std::u16string_view(u"anchor"),
            std::u16string_view(u"anchor"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6DF"),
            std::u16string_view(u"ring_buoy"),
            std::u16string_view(u"ring buoy"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026F5"),
            std::u16string_view(u"sailboat"),
            std::u16string_view(u"sailboat"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F6"),
            std::u16string_view(u"canoe"),
            std::u16string_view(u"canoe"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A4"),
            std::u16string_view(u"speedboat"),
            std::u16string_view(u"speedboat"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F3\U0000FE0F"),
            std::u16string_view(u"passenger_ship"),
            std::u16string_view(u"passenger ship"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026F4\U0000FE0F"),
            std::u16string_view(u"ferry"),
            std::u16string_view(u"ferry"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6E5\U0000FE0F"),
            std::u16string_view(u"motor_boat"),
            std::u16string_view(u"motor boat"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A2"),
            std::u16string_view(u"ship"),
            std::u16string_view(u"ship"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002708\U0000FE0F"),
            std::u16string_view(u"airplane"),
            std::u16string_view(u"airplane"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6E9\U0000FE0F"),
            std::u16string_view(u"small_airplane"),
            std::u16string_view(u"small airplane"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6EB"),
            std::u16string_view(u"airplane_departure"),
            std::u16string_view(u"airplane departure"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6EC"),
            std::u16string_view(u"airplane_arrival"),
            std::u16string_view(u"airplane arrival"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001FA82"),
            std::u16string_view(u"parachute"),
            std::u16string_view(u"parachute"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F4BA"),
            std::u16string_view(u"seat"),
            std::u16string_view(u"seat"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F681"),
            std::u16string_view(u"helicopter"),
            std::u16string_view(u"helicopter"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F69F"),
            std::u16string_view(u"suspension_railway"),
            std::u16string_view(u"suspension railway"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A0"),
            std::u16string_view(u"mountain_cableway"),
            std::u16string_view(u"mountain cableway"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6A1"),
            std::u16string_view(u"aerial_tramway"),
            std::u16string_view(u"aerial tramway"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F0\U0000FE0F"),
            std::u16string_view(u"satellite"),
            std::u16string_view(u"satellite"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F680"),
            std::u16string_view(u"rocket"),
            std::u16string_view(u"rocket"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6F8"),
            std::u16string_view(u"flying_saucer"),
            std::u16string_view(u"flying saucer"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F6CE\U0000FE0F"),
            std::u16string_view(u"bellhop_bell"),
            std::u16string_view(u"bellhop bell"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F9F3"),
            std::u16string_view(u"luggage"),
            std::u16string_view(u"luggage"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0000231B"),
            std::u16string_view(u"hourglass_done"),
            std::u16string_view(u"hourglass done"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000023F3"),
            std::u16string_view(u"hourglass_not_done"),
            std::u16string_view(u"hourglass not done"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0000231A"),
            std::u16string_view(u"watch"),
            std::u16string_view(u"watch"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000023F0"),
            std::u16string_view(u"alarm_clock"),
            std::u16string_view(u"alarm clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000023F1\U0000FE0F"),
            std::u16string_view(u"stopwatch"),
            std::u16string_view(u"stopwatch"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000023F2\U0000FE0F"),
            std::u16string_view(u"timer_clock"),
            std::u16string_view(u"timer clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F570\U0000FE0F"),
            std::u16string_view(u"mantelpiece_clock"),
            std::u16string_view(u"mantelpiece clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55B"),
            std::u16string_view(u"twelve_o_clock"),
            std::u16string_view(u"twelve o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F567"),
            std::u16string_view(u"twelve_thirty"),
            std::u16string_view(u"twelve-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F550"),
            std::u16string_view(u"one_o_clock"),
            std::u16string_view(u"one o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55C"),
            std::u16string_view(u"one_thirty"),
            std::u16string_view(u"one-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F551"),
            std::u16string_view(u"two_o_clock"),
            std::u16string_view(u"two o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55D"),
            std::u16string_view(u"two_thirty"),
            std::u16string_view(u"two-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F552"),
            std::u16string_view(u"three_o_clock"),
            std::u16string_view(u"three o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55E"),
            std::u16string_view(u"three_thirty"),
            std::u16string_view(u"three-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F553"),
            std::u16string_view(u"four_o_clock"),
            std::u16string_view(u"four o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55F"),
            std::u16string_view(u"four_thirty"),
            std::u16string_view(u"four-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F554"),
            std::u16string_view(u"five_o_clock"),
            std::u16string_view(u"five o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F560"),
            std::u16string_view(u"five_thirty"),
            std::u16string_view(u"five-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F555"),
            std::u16string_view(u"six_o_clock"),
            std::u16string_view(u"six o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F561"),
            std::u16string_view(u"six_thirty"),
            std::u16string_view(u"six-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F556"),
            std::u16string_view(u"seven_o_clock"),
            std::u16string_view(u"seven o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F562"),
            std::u16string_view(u"seven_thirty"),
            std::u16string_view(u"seven-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F557"),
            std::u16string_view(u"eight_o_clock"),
            std::u16string_view(u"eight o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F563"),
            std::u16string_view(u"eight_thirty"),
            std::u16string_view(u"eight-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F558"),
            std::u16string_view(u"nine_o_clock"),
            std::u16string_view(u"nine o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F564"),
            std::u16string_view(u"nine_thirty"),
            std::u16string_view(u"nine-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F559"),
            std::u16string_view(u"ten_o_clock"),
            std::u16string_view(u"ten o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F565"),
            std::u16string_view(u"ten_thirty"),
            std::u16string_view(u"ten-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F55A"),
            std::u16string_view(u"eleven_o_clock"),
            std::u16string_view(u"eleven o’clock"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F566"),
            std::u16string_view(u"eleven_thirty"),
            std::u16string_view(u"eleven-thirty"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F311"),
            std::u16string_view(u"new_moon"),
            std::u16string_view(u"new moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F312"),
            std::u16string_view(u"waxing_crescent_moon"),
            std::u16string_view(u"waxing crescent moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F313"),
            std::u16string_view(u"first_quarter_moon"),
            std::u16string_view(u"first quarter moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F314"),
            std::u16string_view(u"waxing_gibbous_moon"),
            std::u16string_view(u"waxing gibbous moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F315"),
            std::u16string_view(u"full_moon"),
            std::u16string_view(u"full moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F316"),
            std::u16string_view(u"waning_gibbous_moon"),
            std::u16string_view(u"waning gibbous moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F317"),
            std::u16string_view(u"last_quarter_moon"),
            std::u16string_view(u"last quarter moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F318"),
            std::u16string_view(u"waning_crescent_moon"),
            std::u16string_view(u"waning crescent moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F319"),
            std::u16string_view(u"crescent_moon"),
            std::u16string_view(u"crescent moon"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31A"),
            std::u16string_view(u"new_moon"),
            std::u16string_view(u"new moon face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31B"),
            std::u16string_view(u"first_quarter_moon"),
            std::u16string_view(u"first quarter moon face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31C"),
            std::u16string_view(u"last_quarter_moon"),
            std::u16string_view(u"last quarter moon face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F321\U0000FE0F"),
            std::u16string_view(u"thermometer"),
            std::u16string_view(u"thermometer"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002600\U0000FE0F"),
            std::u16string_view(u"sun"),
            std::u16string_view(u"sun"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31D"),
            std::u16string_view(u"full_moon"),
            std::u16string_view(u"full moon face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31E"),
            std::u16string_view(u"sun_with"),
            std::u16string_view(u"sun with face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001FA90"),
            std::u16string_view(u"ringed_planet"),
            std::u16string_view(u"ringed planet"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002B50"),
            std::u16string_view(u"star"),
            std::u16string_view(u"star"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F31F"),
            std::u16string_view(u"glowing_star"),
            std::u16string_view(u"glowing star"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F320"),
            std::u16string_view(u"shooting_star"),
            std::u16string_view(u"shooting star"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F30C"),
            std::u16string_view(u"milky_way"),
            std::u16string_view(u"milky way"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002601\U0000FE0F"),
            std::u16string_view(u"cloud"),
            std::u16string_view(u"cloud"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026C5"),
            std::u16string_view(u"sun_behind_cloud"),
            std::u16string_view(u"sun behind cloud"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026C8\U0000FE0F"),
            std::u16string_view(u"cloud_with_lightning_and_rain"),
            std::u16string_view(u"cloud with lightning and rain"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F324\U0000FE0F"),
            std::u16string_view(u"sun_behind_small_cloud"),
            std::u16string_view(u"sun behind small cloud"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F325\U0000FE0F"),
            std::u16string_view(u"sun_behind_large_cloud"),
            std::u16string_view(u"sun behind large cloud"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F326\U0000FE0F"),
            std::u16string_view(u"sun_behind_rain_cloud"),
            std::u16string_view(u"sun behind rain cloud"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F327\U0000FE0F"),
            std::u16string_view(u"cloud_with_rain"),
            std::u16string_view(u"cloud with rain"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F328\U0000FE0F"),
            std::u16string_view(u"cloud_with_snow"),
            std::u16string_view(u"cloud with snow"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F329\U0000FE0F"),
            std::u16string_view(u"cloud_with_lightning"),
            std::u16string_view(u"cloud with lightning"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F32A\U0000FE0F"),
            std::u16string_view(u"tornado"),
            std::u16string_view(u"tornado"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F32B\U0000FE0F"),
            std::u16string_view(u"fog"),
            std::u16string_view(u"fog"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F32C\U0000FE0F"),
            std::u16string_view(u"wind"),
            std::u16string_view(u"wind face"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F300"),
            std::u16string_view(u"cyclone"),
            std::u16string_view(u"cyclone"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F308"),
            std::u16string_view(u"rainbow"),
            std::u16string_view(u"rainbow"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F302"),
            std::u16string_view(u"closed_umbrella"),
            std::u16string_view(u"closed umbrella"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002602\U0000FE0F"),
            std::u16string_view(u"umbrella"),
            std::u16string_view(u"umbrella"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002614"),
            std::u16string_view(u"umbrella_with_rain_drops"),
            std::u16string_view(u"umbrella with rain drops"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026F1\U0000FE0F"),
            std::u16string_view(u"umbrella_on_ground"),
            std::u16string_view(u"umbrella on ground"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026A1"),
            std::u16string_view(u"zap"),
            std::u16string_view(u"high voltage"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002744\U0000FE0F"),
            std::u16string_view(u"snowflake"),
            std::u16string_view(u"snowflake"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002603\U0000FE0F"),
            std::u16string_view(u"snowman"),
            std::u16string_view(u"snowman"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U000026C4"),
            std::u16string_view(u"snowman_without_snow"),
            std::u16string_view(u"snowman without snow"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U00002604\U0000FE0F"),
            std::u16string_view(u"comet"),
            std::u16string_view(u"comet"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F525"),
            std::u16string_view(u"fire"),
            std::u16string_view(u"fire"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F4A7"),
            std::u16string_view(u"droplet"),
            std::u16string_view(u"droplet"),
            emoji::Emoji::Category::Travel},
      Emoji{std::u16string_view(u"\U0001F30A"),
            std::u16string_view(u"water_wave"),
            std::u16string_view(u"water wave"),
            emoji::Emoji::Category::Travel},
      // Objects
      Emoji{std::u16string_view(u"\U0001F453"),
            std::u16string_view(u"glasses"),
            std::u16string_view(u"glasses"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F576\U0000FE0F"),
            std::u16string_view(u"sunglasses"),
            std::u16string_view(u"sunglasses"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F97D"),
            std::u16string_view(u"goggles"),
            std::u16string_view(u"goggles"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F97C"),
            std::u16string_view(u"lab_coat"),
            std::u16string_view(u"lab coat"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9BA"),
            std::u16string_view(u"safety_vest"),
            std::u16string_view(u"safety vest"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F454"),
            std::u16string_view(u"necktie"),
            std::u16string_view(u"necktie"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F455"),
            std::u16string_view(u"t_shirt"),
            std::u16string_view(u"t-shirt"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F456"),
            std::u16string_view(u"jeans"),
            std::u16string_view(u"jeans"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9E3"),
            std::u16string_view(u"scarf"),
            std::u16string_view(u"scarf"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9E4"),
            std::u16string_view(u"gloves"),
            std::u16string_view(u"gloves"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9E5"),
            std::u16string_view(u"coat"),
            std::u16string_view(u"coat"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9E6"),
            std::u16string_view(u"socks"),
            std::u16string_view(u"socks"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F457"),
            std::u16string_view(u"dress"),
            std::u16string_view(u"dress"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F458"),
            std::u16string_view(u"kimono"),
            std::u16string_view(u"kimono"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F97B"),
            std::u16string_view(u"sari"),
            std::u16string_view(u"sari"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA71"),
            std::u16string_view(u"one_piece_swimsuit"),
            std::u16string_view(u"one-piece swimsuit"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA72"),
            std::u16string_view(u"briefs"),
            std::u16string_view(u"briefs"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA73"),
            std::u16string_view(u"shorts"),
            std::u16string_view(u"shorts"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F459"),
            std::u16string_view(u"bikini"),
            std::u16string_view(u"bikini"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45A"),
            std::u16string_view(u"woman_s_clothes"),
            std::u16string_view(u"woman’s clothes"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAAD"),
            std::u16string_view(u"folding_hand_fan"),
            std::u16string_view(u"folding hand fan"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45B"),
            std::u16string_view(u"purse"),
            std::u16string_view(u"purse"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45C"),
            std::u16string_view(u"handbag"),
            std::u16string_view(u"handbag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45D"),
            std::u16string_view(u"clutch_bag"),
            std::u16string_view(u"clutch bag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6CD\U0000FE0F"),
            std::u16string_view(u"shopping_bags"),
            std::u16string_view(u"shopping bags"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F392"),
            std::u16string_view(u"backpack"),
            std::u16string_view(u"backpack"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA74"),
            std::u16string_view(u"thong_sandal"),
            std::u16string_view(u"thong sandal"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45E"),
            std::u16string_view(u"man_s_shoe"),
            std::u16string_view(u"man’s shoe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F45F"),
            std::u16string_view(u"running_shoe"),
            std::u16string_view(u"running shoe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F97E"),
            std::u16string_view(u"hiking_boot"),
            std::u16string_view(u"hiking boot"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F97F"),
            std::u16string_view(u"flat_shoe"),
            std::u16string_view(u"flat shoe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F460"),
            std::u16string_view(u"high_heeled_shoe"),
            std::u16string_view(u"high-heeled shoe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F461"),
            std::u16string_view(u"woman_s_sandal"),
            std::u16string_view(u"woman’s sandal"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA70"),
            std::u16string_view(u"ballet_shoes"),
            std::u16string_view(u"ballet shoes"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F462"),
            std::u16string_view(u"woman_s_boot"),
            std::u16string_view(u"woman’s boot"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAAE"),
            std::u16string_view(u"hair_pick"),
            std::u16string_view(u"hair pick"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F451"),
            std::u16string_view(u"crown"),
            std::u16string_view(u"crown"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F452"),
            std::u16string_view(u"woman_s_hat"),
            std::u16string_view(u"woman’s hat"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3A9"),
            std::u16string_view(u"top_hat"),
            std::u16string_view(u"top hat"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F393"),
            std::u16string_view(u"graduation_cap"),
            std::u16string_view(u"graduation cap"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9E2"),
            std::u16string_view(u"billed_cap"),
            std::u16string_view(u"billed cap"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA96"),
            std::u16string_view(u"military_helmet"),
            std::u16string_view(u"military helmet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026D1\U0000FE0F"),
            std::u16string_view(u"rescue_worker_s_helmet"),
            std::u16string_view(u"rescue worker’s helmet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4FF"),
            std::u16string_view(u"prayer_beads"),
            std::u16string_view(u"prayer beads"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F484"),
            std::u16string_view(u"lipstick"),
            std::u16string_view(u"lipstick"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F48D"),
            std::u16string_view(u"ring"),
            std::u16string_view(u"ring"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F48E"),
            std::u16string_view(u"gem_stone"),
            std::u16string_view(u"gem stone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F507"),
            std::u16string_view(u"muted_speaker"),
            std::u16string_view(u"muted speaker"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F508"),
            std::u16string_view(u"speaker_low_volume"),
            std::u16string_view(u"speaker low volume"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F509"),
            std::u16string_view(u"speaker_medium_volume"),
            std::u16string_view(u"speaker medium volume"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50A"),
            std::u16string_view(u"speaker_high_volume"),
            std::u16string_view(u"speaker high volume"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E2"),
            std::u16string_view(u"loudspeaker"),
            std::u16string_view(u"loudspeaker"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E3"),
            std::u16string_view(u"megaphone"),
            std::u16string_view(u"megaphone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4EF"),
            std::u16string_view(u"postal_horn"),
            std::u16string_view(u"postal horn"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F514"),
            std::u16string_view(u"bell"),
            std::u16string_view(u"bell"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F515"),
            std::u16string_view(u"bell_with_slash"),
            std::u16string_view(u"bell with slash"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3BC"),
            std::u16string_view(u"musical_score"),
            std::u16string_view(u"musical score"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3B5"),
            std::u16string_view(u"musical_note"),
            std::u16string_view(u"musical note"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3B6"),
            std::u16string_view(u"musical_notes"),
            std::u16string_view(u"musical notes"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F399\U0000FE0F"),
            std::u16string_view(u"studio_microphone"),
            std::u16string_view(u"studio microphone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F39A\U0000FE0F"),
            std::u16string_view(u"level_slider"),
            std::u16string_view(u"level slider"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F39B\U0000FE0F"),
            std::u16string_view(u"control_knobs"),
            std::u16string_view(u"control knobs"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3A4"),
            std::u16string_view(u"microphone"),
            std::u16string_view(u"microphone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3A7"),
            std::u16string_view(u"headphone"),
            std::u16string_view(u"headphone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4FB"),
            std::u16string_view(u"radio"),
            std::u16string_view(u"radio"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3B7"),
            std::u16string_view(u"saxophone"),
            std::u16string_view(u"saxophone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA97"),
            std::u16string_view(u"accordion"),
            std::u16string_view(u"accordion"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3B8"),
            std::u16string_view(u"guitar"),
            std::u16string_view(u"guitar"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3B9"),
            std::u16string_view(u"musical_keyboard"),
            std::u16string_view(u"musical keyboard"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3BA"),
            std::u16string_view(u"trumpet"),
            std::u16string_view(u"trumpet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3BB"),
            std::u16string_view(u"violin"),
            std::u16string_view(u"violin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA95"),
            std::u16string_view(u"banjo"),
            std::u16string_view(u"banjo"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F941"),
            std::u16string_view(u"drum"),
            std::u16string_view(u"drum"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA98"),
            std::u16string_view(u"long_drum"),
            std::u16string_view(u"long drum"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA87"),
            std::u16string_view(u"maracas"),
            std::u16string_view(u"maracas"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA88"),
            std::u16string_view(u"flute"),
            std::u16string_view(u"flute"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F1"),
            std::u16string_view(u"mobile_phone"),
            std::u16string_view(u"mobile phone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F2"),
            std::u16string_view(u"mobile_phone_with_arrow"),
            std::u16string_view(u"mobile phone with arrow"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0000260E\U0000FE0F"),
            std::u16string_view(u"telephone"),
            std::u16string_view(u"telephone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4DE"),
            std::u16string_view(u"telephone_receiver"),
            std::u16string_view(u"telephone receiver"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4DF"),
            std::u16string_view(u"pager"),
            std::u16string_view(u"pager"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E0"),
            std::u16string_view(u"fax_machine"),
            std::u16string_view(u"fax machine"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50B"),
            std::u16string_view(u"battery"),
            std::u16string_view(u"battery"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAAB"),
            std::u16string_view(u"low_battery"),
            std::u16string_view(u"low battery"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50C"),
            std::u16string_view(u"electric_plug"),
            std::u16string_view(u"electric plug"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4BB"),
            std::u16string_view(u"laptop"),
            std::u16string_view(u"laptop"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5A5\U0000FE0F"),
            std::u16string_view(u"desktop_computer"),
            std::u16string_view(u"desktop computer"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5A8\U0000FE0F"),
            std::u16string_view(u"printer"),
            std::u16string_view(u"printer"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002328\U0000FE0F"),
            std::u16string_view(u"keyboard"),
            std::u16string_view(u"keyboard"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5B1\U0000FE0F"),
            std::u16string_view(u"computer_mouse"),
            std::u16string_view(u"computer mouse"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5B2\U0000FE0F"),
            std::u16string_view(u"trackball"),
            std::u16string_view(u"trackball"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4BD"),
            std::u16string_view(u"computer_disk"),
            std::u16string_view(u"computer disk"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4BE"),
            std::u16string_view(u"floppy_disk"),
            std::u16string_view(u"floppy disk"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4BF"),
            std::u16string_view(u"optical_disk"),
            std::u16string_view(u"optical disk"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C0"),
            std::u16string_view(u"dvd"),
            std::u16string_view(u"dvd"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9EE"),
            std::u16string_view(u"abacus"),
            std::u16string_view(u"abacus"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3A5"),
            std::u16string_view(u"movie_camera"),
            std::u16string_view(u"movie camera"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F39E\U0000FE0F"),
            std::u16string_view(u"film_frames"),
            std::u16string_view(u"film frames"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4FD\U0000FE0F"),
            std::u16string_view(u"film_projector"),
            std::u16string_view(u"film projector"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3AC"),
            std::u16string_view(u"clapper_board"),
            std::u16string_view(u"clapper board"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4FA"),
            std::u16string_view(u"television"),
            std::u16string_view(u"television"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F7"),
            std::u16string_view(u"camera"),
            std::u16string_view(u"camera"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F8"),
            std::u16string_view(u"camera_with_flash"),
            std::u16string_view(u"camera with flash"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F9"),
            std::u16string_view(u"video_camera"),
            std::u16string_view(u"video camera"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4FC"),
            std::u16string_view(u"videocassette"),
            std::u16string_view(u"videocassette"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50D"),
            std::u16string_view(u"magnifying_glass_tilted_left"),
            std::u16string_view(u"magnifying glass tilted left"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50E"),
            std::u16string_view(u"magnifying_glass_tilted_right"),
            std::u16string_view(u"magnifying glass tilted right"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F56F\U0000FE0F"),
            std::u16string_view(u"candle"),
            std::u16string_view(u"candle"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4A1"),
            std::u16string_view(u"light_bulb"),
            std::u16string_view(u"light bulb"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F526"),
            std::u16string_view(u"flashlight"),
            std::u16string_view(u"flashlight"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3EE"),
            std::u16string_view(u"red_paper_lantern"),
            std::u16string_view(u"red paper lantern"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA94"),
            std::u16string_view(u"diya_lamp"),
            std::u16string_view(u"diya lamp"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D4"),
            std::u16string_view(u"notebook_with_decorative_cover"),
            std::u16string_view(u"notebook with decorative cover"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D5"),
            std::u16string_view(u"closed_book"),
            std::u16string_view(u"closed book"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D6"),
            std::u16string_view(u"open_book"),
            std::u16string_view(u"open book"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D7"),
            std::u16string_view(u"green_book"),
            std::u16string_view(u"green book"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D8"),
            std::u16string_view(u"blue_book"),
            std::u16string_view(u"blue book"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D9"),
            std::u16string_view(u"orange_book"),
            std::u16string_view(u"orange book"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4DA"),
            std::u16string_view(u"books"),
            std::u16string_view(u"books"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D3"),
            std::u16string_view(u"notebook"),
            std::u16string_view(u"notebook"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D2"),
            std::u16string_view(u"ledger"),
            std::u16string_view(u"ledger"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C3"),
            std::u16string_view(u"page_with_curl"),
            std::u16string_view(u"page with curl"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4DC"),
            std::u16string_view(u"scroll"),
            std::u16string_view(u"scroll"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C4"),
            std::u16string_view(u"page_facing_up"),
            std::u16string_view(u"page facing up"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4F0"),
            std::u16string_view(u"newspaper"),
            std::u16string_view(u"newspaper"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5DE\U0000FE0F"),
            std::u16string_view(u"rolled_up_newspaper"),
            std::u16string_view(u"rolled-up newspaper"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D1"),
            std::u16string_view(u"bookmark_tabs"),
            std::u16string_view(u"bookmark tabs"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F516"),
            std::u16string_view(u"bookmark"),
            std::u16string_view(u"bookmark"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3F7\U0000FE0F"),
            std::u16string_view(u"label"),
            std::u16string_view(u"label"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B0"),
            std::u16string_view(u"money_bag"),
            std::u16string_view(u"money bag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA99"),
            std::u16string_view(u"coin"),
            std::u16string_view(u"coin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B4"),
            std::u16string_view(u"yen"),
            std::u16string_view(u"yen banknote"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B5"),
            std::u16string_view(u"dollar"),
            std::u16string_view(u"dollar banknote"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B6"),
            std::u16string_view(u"euro"),
            std::u16string_view(u"euro banknote"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B7"),
            std::u16string_view(u"pound"),
            std::u16string_view(u"pound banknote"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B8"),
            std::u16string_view(u"money_with_wings"),
            std::u16string_view(u"money with wings"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B3"),
            std::u16string_view(u"credit_card"),
            std::u16string_view(u"credit card"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FE"),
            std::u16string_view(u"receipt"),
            std::u16string_view(u"receipt"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4B9"),
            std::u16string_view(u"chart_increasing_with_yen"),
            std::u16string_view(u"chart increasing with yen"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002709\U0000FE0F"),
            std::u16string_view(u"envelope"),
            std::u16string_view(u"envelope"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E7"),
            std::u16string_view(u"e_mail"),
            std::u16string_view(u"e-mail"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E8"),
            std::u16string_view(u"incoming_envelope"),
            std::u16string_view(u"incoming envelope"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E9"),
            std::u16string_view(u"envelope_with_arrow"),
            std::u16string_view(u"envelope with arrow"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E4"),
            std::u16string_view(u"outbox_tray"),
            std::u16string_view(u"outbox tray"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E5"),
            std::u16string_view(u"inbox_tray"),
            std::u16string_view(u"inbox tray"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E6"),
            std::u16string_view(u"package"),
            std::u16string_view(u"package"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4EB"),
            std::u16string_view(u"closed_mailbox_with_raised_flag"),
            std::u16string_view(u"closed mailbox with raised flag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4EA"),
            std::u16string_view(u"closed_mailbox_with_lowered_flag"),
            std::u16string_view(u"closed mailbox with lowered flag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4EC"),
            std::u16string_view(u"open_mailbox_with_raised_flag"),
            std::u16string_view(u"open mailbox with raised flag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4ED"),
            std::u16string_view(u"open_mailbox_with_lowered_flag"),
            std::u16string_view(u"open mailbox with lowered flag"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4EE"),
            std::u16string_view(u"postbox"),
            std::u16string_view(u"postbox"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5F3\U0000FE0F"),
            std::u16string_view(u"ballot_box_with_ballot"),
            std::u16string_view(u"ballot box with ballot"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0000270F\U0000FE0F"),
            std::u16string_view(u"pencil"),
            std::u16string_view(u"pencil"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002712\U0000FE0F"),
            std::u16string_view(u"black_nib"),
            std::u16string_view(u"black nib"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F58B\U0000FE0F"),
            std::u16string_view(u"fountain_pen"),
            std::u16string_view(u"fountain pen"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F58A\U0000FE0F"),
            std::u16string_view(u"pen"),
            std::u16string_view(u"pen"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F58C\U0000FE0F"),
            std::u16string_view(u"paintbrush"),
            std::u16string_view(u"paintbrush"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F58D\U0000FE0F"),
            std::u16string_view(u"crayon"),
            std::u16string_view(u"crayon"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4DD"),
            std::u16string_view(u"memo"),
            std::u16string_view(u"memo"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4BC"),
            std::u16string_view(u"briefcase"),
            std::u16string_view(u"briefcase"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C1"),
            std::u16string_view(u"file_folder"),
            std::u16string_view(u"file folder"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C2"),
            std::u16string_view(u"open_file_folder"),
            std::u16string_view(u"open file folder"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5C2\U0000FE0F"),
            std::u16string_view(u"card_index_dividers"),
            std::u16string_view(u"card index dividers"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C5"),
            std::u16string_view(u"calendar"),
            std::u16string_view(u"calendar"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C6"),
            std::u16string_view(u"tear_off_calendar"),
            std::u16string_view(u"tear-off calendar"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5D2\U0000FE0F"),
            std::u16string_view(u"spiral_notepad"),
            std::u16string_view(u"spiral notepad"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5D3\U0000FE0F"),
            std::u16string_view(u"spiral_calendar"),
            std::u16string_view(u"spiral calendar"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C7"),
            std::u16string_view(u"card_index"),
            std::u16string_view(u"card index"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C8"),
            std::u16string_view(u"chart_increasing"),
            std::u16string_view(u"chart increasing"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4C9"),
            std::u16string_view(u"chart_decreasing"),
            std::u16string_view(u"chart decreasing"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CA"),
            std::u16string_view(u"bar_chart"),
            std::u16string_view(u"bar chart"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CB"),
            std::u16string_view(u"clipboard"),
            std::u16string_view(u"clipboard"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CC"),
            std::u16string_view(u"pushpin"),
            std::u16string_view(u"pushpin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CD"),
            std::u16string_view(u"round_pushpin"),
            std::u16string_view(u"round pushpin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CE"),
            std::u16string_view(u"paperclip"),
            std::u16string_view(u"paperclip"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F587\U0000FE0F"),
            std::u16string_view(u"linked_paperclips"),
            std::u16string_view(u"linked paperclips"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4CF"),
            std::u16string_view(u"straight_ruler"),
            std::u16string_view(u"straight ruler"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4D0"),
            std::u16string_view(u"triangular_ruler"),
            std::u16string_view(u"triangular ruler"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002702\U0000FE0F"),
            std::u16string_view(u"scissors"),
            std::u16string_view(u"scissors"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5C3\U0000FE0F"),
            std::u16string_view(u"card_file_box"),
            std::u16string_view(u"card file box"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5C4\U0000FE0F"),
            std::u16string_view(u"file_cabinet"),
            std::u16string_view(u"file cabinet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5D1\U0000FE0F"),
            std::u16string_view(u"wastebasket"),
            std::u16string_view(u"wastebasket"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F512"),
            std::u16string_view(u"locked"),
            std::u16string_view(u"locked"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F513"),
            std::u16string_view(u"unlocked"),
            std::u16string_view(u"unlocked"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F50F"),
            std::u16string_view(u"locked_with_pen"),
            std::u16string_view(u"locked with pen"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F510"),
            std::u16string_view(u"locked_with_key"),
            std::u16string_view(u"locked with key"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F511"),
            std::u16string_view(u"key"),
            std::u16string_view(u"key"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5DD\U0000FE0F"),
            std::u16string_view(u"old_key"),
            std::u16string_view(u"old key"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F528"),
            std::u16string_view(u"hammer"),
            std::u16string_view(u"hammer"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA93"),
            std::u16string_view(u"axe"),
            std::u16string_view(u"axe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026CF\U0000FE0F"),
            std::u16string_view(u"pick"),
            std::u16string_view(u"pick"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002692\U0000FE0F"),
            std::u16string_view(u"hammer_and_pick"),
            std::u16string_view(u"hammer and pick"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6E0\U0000FE0F"),
            std::u16string_view(u"hammer_and_wrench"),
            std::u16string_view(u"hammer and wrench"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5E1\U0000FE0F"),
            std::u16string_view(u"dagger"),
            std::u16string_view(u"dagger"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002694\U0000FE0F"),
            std::u16string_view(u"crossed_swords"),
            std::u16string_view(u"crossed swords"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4A3"),
            std::u16string_view(u"bomb"),
            std::u16string_view(u"bomb"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA83"),
            std::u16string_view(u"boomerang"),
            std::u16string_view(u"boomerang"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F3F9"),
            std::u16string_view(u"bow_and_arrow"),
            std::u16string_view(u"bow and arrow"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6E1\U0000FE0F"),
            std::u16string_view(u"shield"),
            std::u16string_view(u"shield"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9A"),
            std::u16string_view(u"carpentry_saw"),
            std::u16string_view(u"carpentry saw"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F527"),
            std::u16string_view(u"wrench"),
            std::u16string_view(u"wrench"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9B"),
            std::u16string_view(u"screwdriver"),
            std::u16string_view(u"screwdriver"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F529"),
            std::u16string_view(u"nut_and_bolt"),
            std::u16string_view(u"nut and bolt"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002699\U0000FE0F"),
            std::u16string_view(u"gear"),
            std::u16string_view(u"gear"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5DC\U0000FE0F"),
            std::u16string_view(u"clamp"),
            std::u16string_view(u"clamp"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002696\U0000FE0F"),
            std::u16string_view(u"balance_scale"),
            std::u16string_view(u"balance scale"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9AF"),
            std::u16string_view(u"white_cane"),
            std::u16string_view(u"white cane"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F517"),
            std::u16string_view(u"link"),
            std::u16string_view(u"link"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026D3\U0000FE0F\U0000200D\U0001F4A5"),
            std::u16string_view(u"broken_chain"),
            std::u16string_view(u"broken chain"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026D3\U0000FE0F"),
            std::u16string_view(u"chains"),
            std::u16string_view(u"chains"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9D"),
            std::u16string_view(u"hook"),
            std::u16string_view(u"hook"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9F0"),
            std::u16string_view(u"toolbox"),
            std::u16string_view(u"toolbox"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9F2"),
            std::u16string_view(u"magnet"),
            std::u16string_view(u"magnet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9C"),
            std::u16string_view(u"ladder"),
            std::u16string_view(u"ladder"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U00002697\U0000FE0F"),
            std::u16string_view(u"alembic"),
            std::u16string_view(u"alembic"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9EA"),
            std::u16string_view(u"test_tube"),
            std::u16string_view(u"test tube"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9EB"),
            std::u16string_view(u"petri_dish"),
            std::u16string_view(u"petri dish"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9EC"),
            std::u16string_view(u"dna"),
            std::u16string_view(u"dna"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F52C"),
            std::u16string_view(u"microscope"),
            std::u16string_view(u"microscope"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F52D"),
            std::u16string_view(u"telescope"),
            std::u16string_view(u"telescope"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F4E1"),
            std::u16string_view(u"satellite_antenna"),
            std::u16string_view(u"satellite antenna"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F489"),
            std::u16string_view(u"syringe"),
            std::u16string_view(u"syringe"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA78"),
            std::u16string_view(u"drop_of_blood"),
            std::u16string_view(u"drop of blood"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F48A"),
            std::u16string_view(u"pill"),
            std::u16string_view(u"pill"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA79"),
            std::u16string_view(u"adhesive_bandage"),
            std::u16string_view(u"adhesive bandage"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA7C"),
            std::u16string_view(u"crutch"),
            std::u16string_view(u"crutch"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA7A"),
            std::u16string_view(u"stethoscope"),
            std::u16string_view(u"stethoscope"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA7B"),
            std::u16string_view(u"x_ray"),
            std::u16string_view(u"x-ray"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6AA"),
            std::u16string_view(u"door"),
            std::u16string_view(u"door"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6D7"),
            std::u16string_view(u"elevator"),
            std::u16string_view(u"elevator"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9E"),
            std::u16string_view(u"mirror"),
            std::u16string_view(u"mirror"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA9F"),
            std::u16string_view(u"window"),
            std::u16string_view(u"window"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6CF\U0000FE0F"),
            std::u16string_view(u"bed"),
            std::u16string_view(u"bed"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6CB\U0000FE0F"),
            std::u16string_view(u"couch_and_lamp"),
            std::u16string_view(u"couch and lamp"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA91"),
            std::u16string_view(u"chair"),
            std::u16string_view(u"chair"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6BD"),
            std::u16string_view(u"toilet"),
            std::u16string_view(u"toilet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA0"),
            std::u16string_view(u"plunger"),
            std::u16string_view(u"plunger"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6BF"),
            std::u16string_view(u"shower"),
            std::u16string_view(u"shower"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6C1"),
            std::u16string_view(u"bathtub"),
            std::u16string_view(u"bathtub"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA4"),
            std::u16string_view(u"mouse_trap"),
            std::u16string_view(u"mouse trap"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FA92"),
            std::u16string_view(u"razor"),
            std::u16string_view(u"razor"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9F4"),
            std::u16string_view(u"lotion_bottle"),
            std::u16string_view(u"lotion bottle"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9F7"),
            std::u16string_view(u"safety_pin"),
            std::u16string_view(u"safety pin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9F9"),
            std::u16string_view(u"broom"),
            std::u16string_view(u"broom"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FA"),
            std::u16string_view(u"basket"),
            std::u16string_view(u"basket"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FB"),
            std::u16string_view(u"roll_of_paper"),
            std::u16string_view(u"roll of paper"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA3"),
            std::u16string_view(u"bucket"),
            std::u16string_view(u"bucket"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FC"),
            std::u16string_view(u"soap"),
            std::u16string_view(u"soap"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAE7"),
            std::u16string_view(u"bubbles"),
            std::u16string_view(u"bubbles"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA5"),
            std::u16string_view(u"toothbrush"),
            std::u16string_view(u"toothbrush"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FD"),
            std::u16string_view(u"sponge"),
            std::u16string_view(u"sponge"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9EF"),
            std::u16string_view(u"fire_extinguisher"),
            std::u16string_view(u"fire extinguisher"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6D2"),
            std::u16string_view(u"shopping_cart"),
            std::u16string_view(u"shopping cart"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F6AC"),
            std::u16string_view(u"cigarette"),
            std::u16string_view(u"cigarette"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026B0\U0000FE0F"),
            std::u16string_view(u"coffin"),
            std::u16string_view(u"coffin"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA6"),
            std::u16string_view(u"headstone"),
            std::u16string_view(u"headstone"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U000026B1\U0000FE0F"),
            std::u16string_view(u"funeral_urn"),
            std::u16string_view(u"funeral urn"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F9FF"),
            std::u16string_view(u"nazar_amulet"),
            std::u16string_view(u"nazar amulet"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAAC"),
            std::u16string_view(u"hamsa"),
            std::u16string_view(u"hamsa"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001F5FF"),
            std::u16string_view(u"moai"),
            std::u16string_view(u"moai"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAA7"),
            std::u16string_view(u"placard"),
            std::u16string_view(u"placard"),
            emoji::Emoji::Category::Objects},
      Emoji{std::u16string_view(u"\U0001FAAA"),
            std::u16string_view(u"identification_card"),
            std::u16string_view(u"identification card"),
            emoji::Emoji::Category::Objects},
      // Symbols
      Emoji{std::u16string_view(u"\U0000200C\U0001F1E6\U0000200C"),
            std::u16string_view(u"regional_indicator_a"),
            std::u16string_view(u"regional indicator a"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1E7\U0000200C"),
            std::u16string_view(u"regional_indicator_b"),
            std::u16string_view(u"regional indicator b"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1E8\U0000200C"),
            std::u16string_view(u"regional_indicator_c"),
            std::u16string_view(u"regional indicator c"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1E9\U0000200C"),
            std::u16string_view(u"regional_indicator_d"),
            std::u16string_view(u"regional indicator d"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1EA\U0000200C"),
            std::u16string_view(u"regional_indicator_e"),
            std::u16string_view(u"regional indicator e"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1EB\U0000200C"),
            std::u16string_view(u"regional_indicator_f"),
            std::u16string_view(u"regional indicator f"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1EC\U0000200C"),
            std::u16string_view(u"regional_indicator_g"),
            std::u16string_view(u"regional indicator g"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1ED\U0000200C"),
            std::u16string_view(u"regional_indicator_h"),
            std::u16string_view(u"regional indicator h"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1EE\U0000200C"),
            std::u16string_view(u"regional_indicator_i"),
            std::u16string_view(u"regional indicator i"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1EF\U0000200C"),
            std::u16string_view(u"regional_indicator_j"),
            std::u16string_view(u"regional indicator j"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F0\U0000200C"),
            std::u16string_view(u"regional_indicator_k"),
            std::u16string_view(u"regional indicator k"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F1\U0000200C"),
            std::u16string_view(u"regional_indicator_l"),
            std::u16string_view(u"regional indicator l"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F2\U0000200C"),
            std::u16string_view(u"regional_indicator_m"),
            std::u16string_view(u"regional indicator m"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F3\U0000200C"),
            std::u16string_view(u"regional_indicator_n"),
            std::u16string_view(u"regional indicator n"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F4\U0000200C"),
            std::u16string_view(u"regional_indicator_o"),
            std::u16string_view(u"regional indicator o"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F5\U0000200C"),
            std::u16string_view(u"regional_indicator_p"),
            std::u16string_view(u"regional indicator p"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F6\U0000200C"),
            std::u16string_view(u"regional_indicator_q"),
            std::u16string_view(u"regional indicator q"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F7\U0000200C"),
            std::u16string_view(u"regional_indicator_r"),
            std::u16string_view(u"regional indicator r"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F8\U0000200C"),
            std::u16string_view(u"regional_indicator_s"),
            std::u16string_view(u"regional indicator s"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1F9\U0000200C"),
            std::u16string_view(u"regional_indicator_t"),
            std::u16string_view(u"regional indicator t"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FA\U0000200C"),
            std::u16string_view(u"regional_indicator_u"),
            std::u16string_view(u"regional indicator u"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FB\U0000200C"),
            std::u16string_view(u"regional_indicator_v"),
            std::u16string_view(u"regional indicator v"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FC\U0000200C"),
            std::u16string_view(u"regional_indicator_w"),
            std::u16string_view(u"regional indicator w"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FD\U0000200C"),
            std::u16string_view(u"regional_indicator_x"),
            std::u16string_view(u"regional indicator x"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FE\U0000200C"),
            std::u16string_view(u"regional_indicator_y"),
            std::u16string_view(u"regional indicator y"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000200C\U0001F1FF\U0000200C"),
            std::u16string_view(u"regional_indicator_z"),
            std::u16string_view(u"regional indicator z"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F3E7"),
            std::u16string_view(u"atm_sign"),
            std::u16string_view(u"ATM sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6AE"),
            std::u16string_view(u"litter_in_bin_sign"),
            std::u16string_view(u"litter in bin sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B0"),
            std::u16string_view(u"potable_water"),
            std::u16string_view(u"potable water"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000267F"),
            std::u16string_view(u"wheelchair_symbol"),
            std::u16string_view(u"wheelchair symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B9"),
            std::u16string_view(u"men_s_room"),
            std::u16string_view(u"men’s room"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6BA"),
            std::u16string_view(u"women_s_room"),
            std::u16string_view(u"women’s room"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6BB"),
            std::u16string_view(u"restroom"),
            std::u16string_view(u"restroom"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6BC"),
            std::u16string_view(u"baby_symbol"),
            std::u16string_view(u"baby symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6BE"),
            std::u16string_view(u"water_closet"),
            std::u16string_view(u"water closet"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6C2"),
            std::u16string_view(u"passport_control"),
            std::u16string_view(u"passport control"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6C3"),
            std::u16string_view(u"customs"),
            std::u16string_view(u"customs"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6C4"),
            std::u16string_view(u"baggage_claim"),
            std::u16string_view(u"baggage claim"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6C5"),
            std::u16string_view(u"left_luggage"),
            std::u16string_view(u"left luggage"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026A0\U0000FE0F"),
            std::u16string_view(u"warning"),
            std::u16string_view(u"warning"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B8"),
            std::u16string_view(u"children_crossing"),
            std::u16string_view(u"children crossing"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026D4"),
            std::u16string_view(u"no_entry"),
            std::u16string_view(u"no entry"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6AB"),
            std::u16string_view(u"prohibited"),
            std::u16string_view(u"prohibited"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B3"),
            std::u16string_view(u"no_bicycles"),
            std::u16string_view(u"no bicycles"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6AD"),
            std::u16string_view(u"no_smoking"),
            std::u16string_view(u"no smoking"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6AF"),
            std::u16string_view(u"no_littering"),
            std::u16string_view(u"no littering"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B1"),
            std::u16string_view(u"non_potable_water"),
            std::u16string_view(u"non-potable water"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6B7"),
            std::u16string_view(u"no_pedestrians"),
            std::u16string_view(u"no pedestrians"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4F5"),
            std::u16string_view(u"no_mobile_phones"),
            std::u16string_view(u"no mobile phones"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51E"),
            std::u16string_view(u"no_one_under_eighteen"),
            std::u16string_view(u"no one under eighteen"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002622\U0000FE0F"),
            std::u16string_view(u"radioactive"),
            std::u16string_view(u"radioactive"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002623\U0000FE0F"),
            std::u16string_view(u"biohazard"),
            std::u16string_view(u"biohazard"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B06\U0000FE0F"),
            std::u16string_view(u"up_arrow"),
            std::u16string_view(u"up arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002197\U0000FE0F"),
            std::u16string_view(u"up_right_arrow"),
            std::u16string_view(u"up-right arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000027A1\U0000FE0F"),
            std::u16string_view(u"right_arrow"),
            std::u16string_view(u"right arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002198\U0000FE0F"),
            std::u16string_view(u"down_right_arrow"),
            std::u16string_view(u"down-right arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B07\U0000FE0F"),
            std::u16string_view(u"down_arrow"),
            std::u16string_view(u"down arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002199\U0000FE0F"),
            std::u16string_view(u"down_left_arrow"),
            std::u16string_view(u"down-left arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B05\U0000FE0F"),
            std::u16string_view(u"left_arrow"),
            std::u16string_view(u"left arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002196\U0000FE0F"),
            std::u16string_view(u"up_left_arrow"),
            std::u16string_view(u"up-left arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002195\U0000FE0F"),
            std::u16string_view(u"up_down_arrow"),
            std::u16string_view(u"up-down arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002194\U0000FE0F"),
            std::u16string_view(u"left_right_arrow"),
            std::u16string_view(u"left-right arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000021A9\U0000FE0F"),
            std::u16string_view(u"right_arrow_curving_left"),
            std::u16string_view(u"right arrow curving left"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000021AA\U0000FE0F"),
            std::u16string_view(u"left_arrow_curving_right"),
            std::u16string_view(u"left arrow curving right"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002934\U0000FE0F"),
            std::u16string_view(u"right_arrow_curving_up"),
            std::u16string_view(u"right arrow curving up"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002935\U0000FE0F"),
            std::u16string_view(u"right_arrow_curving_down"),
            std::u16string_view(u"right arrow curving down"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F503"),
            std::u16string_view(u"clockwise_vertical_arrows"),
            std::u16string_view(u"clockwise vertical arrows"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F504"),
            std::u16string_view(u"counterclockwise_arrows"),
            std::u16string_view(u"counterclockwise arrows button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F519"),
            std::u16string_view(u"back_arrow"),
            std::u16string_view(u"BACK arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51A"),
            std::u16string_view(u"end_arrow"),
            std::u16string_view(u"END arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51B"),
            std::u16string_view(u"on_arrow"),
            std::u16string_view(u"ON! arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51C"),
            std::u16string_view(u"soon_arrow"),
            std::u16string_view(u"SOON arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51D"),
            std::u16string_view(u"top_arrow"),
            std::u16string_view(u"TOP arrow"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6D0"),
            std::u16string_view(u"place_of_worship"),
            std::u16string_view(u"place of worship"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000269B\U0000FE0F"),
            std::u16string_view(u"atom_symbol"),
            std::u16string_view(u"atom symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F549\U0000FE0F"),
            std::u16string_view(u"om"),
            std::u16string_view(u"om"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002721\U0000FE0F"),
            std::u16string_view(u"star_of_david"),
            std::u16string_view(u"star of David"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002638\U0000FE0F"),
            std::u16string_view(u"wheel_of_dharma"),
            std::u16string_view(u"wheel of dharma"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000262F\U0000FE0F"),
            std::u16string_view(u"yin_yang"),
            std::u16string_view(u"yin yang"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000271D\U0000FE0F"),
            std::u16string_view(u"latin_cross"),
            std::u16string_view(u"latin cross"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002626\U0000FE0F"),
            std::u16string_view(u"orthodox_cross"),
            std::u16string_view(u"orthodox cross"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000262A\U0000FE0F"),
            std::u16string_view(u"star_and_crescent"),
            std::u16string_view(u"star and crescent"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000262E\U0000FE0F"),
            std::u16string_view(u"peace_symbol"),
            std::u16string_view(u"peace symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F54E"),
            std::u16string_view(u"menorah"),
            std::u16string_view(u"menorah"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F52F"),
            std::u16string_view(u"dotted_six_pointed_star"),
            std::u16string_view(u"dotted six-pointed star"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001FAAF"),
            std::u16string_view(u"khanda"),
            std::u16string_view(u"khanda"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002648"),
            std::u16string_view(u"aries"),
            std::u16string_view(u"Aries"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002649"),
            std::u16string_view(u"taurus"),
            std::u16string_view(u"Taurus"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264A"),
            std::u16string_view(u"gemini"),
            std::u16string_view(u"Gemini"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264B"),
            std::u16string_view(u"cancer"),
            std::u16string_view(u"Cancer"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264C"),
            std::u16string_view(u"leo"),
            std::u16string_view(u"Leo"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264D"),
            std::u16string_view(u"virgo"),
            std::u16string_view(u"Virgo"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264E"),
            std::u16string_view(u"libra"),
            std::u16string_view(u"Libra"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000264F"),
            std::u16string_view(u"scorpio"),
            std::u16string_view(u"Scorpio"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002650"),
            std::u16string_view(u"sagittarius"),
            std::u16string_view(u"Sagittarius"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002651"),
            std::u16string_view(u"capricorn"),
            std::u16string_view(u"Capricorn"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002652"),
            std::u16string_view(u"aquarius"),
            std::u16string_view(u"Aquarius"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002653"),
            std::u16string_view(u"pisces"),
            std::u16string_view(u"Pisces"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026CE"),
            std::u16string_view(u"ophiuchus"),
            std::u16string_view(u"Ophiuchus"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F500"),
            std::u16string_view(u"shuffle_tracks"),
            std::u16string_view(u"shuffle tracks button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F501"),
            std::u16string_view(u"repeat"),
            std::u16string_view(u"repeat button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F502"),
            std::u16string_view(u"repeat_single"),
            std::u16string_view(u"repeat single button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025B6\U0000FE0F"),
            std::u16string_view(u"play"),
            std::u16string_view(u"play button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023E9"),
            std::u16string_view(u"fast_forward"),
            std::u16string_view(u"fast-forward button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023ED\U0000FE0F"),
            std::u16string_view(u"next_track"),
            std::u16string_view(u"next track button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023EF\U0000FE0F"),
            std::u16string_view(u"play_or_pause"),
            std::u16string_view(u"play or pause button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025C0\U0000FE0F"),
            std::u16string_view(u"reverse"),
            std::u16string_view(u"reverse button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023EA"),
            std::u16string_view(u"fast_reverse"),
            std::u16string_view(u"fast reverse button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023EE\U0000FE0F"),
            std::u16string_view(u"last_track"),
            std::u16string_view(u"last track button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F53C"),
            std::u16string_view(u"upwards"),
            std::u16string_view(u"upwards button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023EB"),
            std::u16string_view(u"fast_up"),
            std::u16string_view(u"fast up button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F53D"),
            std::u16string_view(u"downwards"),
            std::u16string_view(u"downwards button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023EC"),
            std::u16string_view(u"fast_down"),
            std::u16string_view(u"fast down button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023F8\U0000FE0F"),
            std::u16string_view(u"pause"),
            std::u16string_view(u"pause button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023F9\U0000FE0F"),
            std::u16string_view(u"stop"),
            std::u16string_view(u"stop button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023FA\U0000FE0F"),
            std::u16string_view(u"record"),
            std::u16string_view(u"record button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000023CF\U0000FE0F"),
            std::u16string_view(u"eject"),
            std::u16string_view(u"eject button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F3A6"),
            std::u16string_view(u"cinema"),
            std::u16string_view(u"cinema"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F505"),
            std::u16string_view(u"dim"),
            std::u16string_view(u"dim button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F506"),
            std::u16string_view(u"bright"),
            std::u16string_view(u"bright button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4F6"),
            std::u16string_view(u"antenna_bars"),
            std::u16string_view(u"antenna bars"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F6DC"),
            std::u16string_view(u"wireless"),
            std::u16string_view(u"wireless"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4F3"),
            std::u16string_view(u"vibration_mode"),
            std::u16string_view(u"vibration mode"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4F4"),
            std::u16string_view(u"mobile_phone_off"),
            std::u16string_view(u"mobile phone off"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002640\U0000FE0F"),
            std::u16string_view(u"female_sign"),
            std::u16string_view(u"female sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002642\U0000FE0F"),
            std::u16string_view(u"male_sign"),
            std::u16string_view(u"male sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026A7\U0000FE0F"),
            std::u16string_view(u"transgender_symbol"),
            std::u16string_view(u"transgender symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002716\U0000FE0F"),
            std::u16string_view(u"multiply"),
            std::u16string_view(u"multiply"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002795"),
            std::u16string_view(u"plus"),
            std::u16string_view(u"plus"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002796"),
            std::u16string_view(u"minus"),
            std::u16string_view(u"minus"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002797"),
            std::u16string_view(u"divide"),
            std::u16string_view(u"divide"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7F0"),
            std::u16string_view(u"heavy_equals_sign"),
            std::u16string_view(u"heavy equals sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000267E\U0000FE0F"),
            std::u16string_view(u"infinity"),
            std::u16string_view(u"infinity"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000203C\U0000FE0F"),
            std::u16string_view(u"bangbang"),
            std::u16string_view(u"double exclamation mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002049\U0000FE0F"),
            std::u16string_view(u"interrobang"),
            std::u16string_view(u"exclamation question mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002753"),
            std::u16string_view(u"question"),
            std::u16string_view(u"red question mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002754"),
            std::u16string_view(u"white_question"),
            std::u16string_view(u"white question mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002755"),
            std::u16string_view(u"white_exclamation"),
            std::u16string_view(u"white exclamation mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002757"),
            std::u16string_view(u"exclamation"),
            std::u16string_view(u"red exclamation mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00003030\U0000FE0F"),
            std::u16string_view(u"wavy_dash"),
            std::u16string_view(u"wavy dash"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4B1"),
            std::u16string_view(u"currency_exchange"),
            std::u16string_view(u"currency exchange"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4B2"),
            std::u16string_view(u"heavy_dollar_sign"),
            std::u16string_view(u"heavy dollar sign"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002695\U0000FE0F"),
            std::u16string_view(u"medical_symbol"),
            std::u16string_view(u"medical symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000267B\U0000FE0F"),
            std::u16string_view(u"recycling_symbol"),
            std::u16string_view(u"recycling symbol"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000269C\U0000FE0F"),
            std::u16string_view(u"fleur_de_lis"),
            std::u16string_view(u"fleur-de-lis"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F531"),
            std::u16string_view(u"trident_emblem"),
            std::u16string_view(u"trident emblem"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4DB"),
            std::u16string_view(u"name_badge"),
            std::u16string_view(u"name badge"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F530"),
            std::u16string_view(u"japanese_symbol_for_beginner"),
            std::u16string_view(u"Japanese symbol for beginner"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B55"),
            std::u16string_view(u"hollow_red_circle"),
            std::u16string_view(u"hollow red circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002705"),
            std::u16string_view(u"check_mark"),
            std::u16string_view(u"check mark button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002611\U0000FE0F"),
            std::u16string_view(u"check_box_with_check"),
            std::u16string_view(u"check box with check"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002714\U0000FE0F"),
            std::u16string_view(u"check_mark"),
            std::u16string_view(u"check mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000274C"),
            std::u16string_view(u"cross_mark"),
            std::u16string_view(u"cross mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000274E"),
            std::u16string_view(u"cross_mark"),
            std::u16string_view(u"cross mark button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000027B0"),
            std::u16string_view(u"curly_loop"),
            std::u16string_view(u"curly loop"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000027BF"),
            std::u16string_view(u"double_curly_loop"),
            std::u16string_view(u"double curly loop"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000303D\U0000FE0F"),
            std::u16string_view(u"part_alternation_mark"),
            std::u16string_view(u"part alternation mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002733\U0000FE0F"),
            std::u16string_view(u"eight_spoked_asterisk"),
            std::u16string_view(u"eight-spoked asterisk"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002734\U0000FE0F"),
            std::u16string_view(u"eight_pointed_star"),
            std::u16string_view(u"eight-pointed star"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002747\U0000FE0F"),
            std::u16string_view(u"sparkle"),
            std::u16string_view(u"sparkle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000000A9\U0000FE0F"),
            std::u16string_view(u"copyright"),
            std::u16string_view(u"copyright"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000000AE\U0000FE0F"),
            std::u16string_view(u"registered"),
            std::u16string_view(u"registered"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002122\U0000FE0F"),
            std::u16string_view(u"trade_mark"),
            std::u16string_view(u"trade mark"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000023\U0000FE0F\U000020E3"),
            std::u16string_view(u"hash"),
            std::u16string_view(u"keycap: #"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0000002A\U0000FE0F\U000020E3"),
            std::u16string_view(u"asterisk"),
            std::u16string_view(u"keycap: *"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000030\U0000FE0F\U000020E3"),
            std::u16string_view(u"zero"),
            std::u16string_view(u"keycap: 0"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000031\U0000FE0F\U000020E3"),
            std::u16string_view(u"one"),
            std::u16string_view(u"keycap: 1"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000032\U0000FE0F\U000020E3"),
            std::u16string_view(u"two"),
            std::u16string_view(u"keycap: 2"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000033\U0000FE0F\U000020E3"),
            std::u16string_view(u"three"),
            std::u16string_view(u"keycap: 3"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000034\U0000FE0F\U000020E3"),
            std::u16string_view(u"four"),
            std::u16string_view(u"keycap: 4"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000035\U0000FE0F\U000020E3"),
            std::u16string_view(u"five"),
            std::u16string_view(u"keycap: 5"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000036\U0000FE0F\U000020E3"),
            std::u16string_view(u"six"),
            std::u16string_view(u"keycap: 6"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000037\U0000FE0F\U000020E3"),
            std::u16string_view(u"seven"),
            std::u16string_view(u"keycap: 7"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000038\U0000FE0F\U000020E3"),
            std::u16string_view(u"eight"),
            std::u16string_view(u"keycap: 8"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00000039\U0000FE0F\U000020E3"),
            std::u16string_view(u"nine"),
            std::u16string_view(u"keycap: 9"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F51F"),
            std::u16string_view(u"ten"),
            std::u16string_view(u"keycap: 10"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F520"),
            std::u16string_view(u"big_abcd"),
            std::u16string_view(u"input latin uppercase"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F521"),
            std::u16string_view(u"abcd"),
            std::u16string_view(u"input latin lowercase"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F522"),
            std::u16string_view(u"1234"),
            std::u16string_view(u"input numbers"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F523"),
            std::u16string_view(u"symbols"),
            std::u16string_view(u"input symbols"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F524"),
            std::u16string_view(u"abc"),
            std::u16string_view(u"input latin letters"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F170\U0000FE0F"),
            std::u16string_view(u"a"),
            std::u16string_view(u"A button (blood type)"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F18E"),
            std::u16string_view(u"ab"),
            std::u16string_view(u"AB button (blood type)"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F171\U0000FE0F"),
            std::u16string_view(u"b"),
            std::u16string_view(u"B button (blood type)"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F191"),
            std::u16string_view(u"cl"),
            std::u16string_view(u"CL button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F192"),
            std::u16string_view(u"cool"),
            std::u16string_view(u"COOL button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F193"),
            std::u16string_view(u"free"),
            std::u16string_view(u"FREE button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002139\U0000FE0F"),
            std::u16string_view(u"information"),
            std::u16string_view(u"information"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F194"),
            std::u16string_view(u"id"),
            std::u16string_view(u"ID button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000024C2\U0000FE0F"),
            std::u16string_view(u"circled_m"),
            std::u16string_view(u"circled M"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F195"),
            std::u16string_view(u"new"),
            std::u16string_view(u"NEW button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F196"),
            std::u16string_view(u"ng"),
            std::u16string_view(u"NG button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F17E\U0000FE0F"),
            std::u16string_view(u"o"),
            std::u16string_view(u"O button (blood type)"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F197"),
            std::u16string_view(u"ok"),
            std::u16string_view(u"OK button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F17F\U0000FE0F"),
            std::u16string_view(u"p"),
            std::u16string_view(u"P button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F198"),
            std::u16string_view(u"sos"),
            std::u16string_view(u"SOS button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F199"),
            std::u16string_view(u"up_"),
            std::u16string_view(u"UP! button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F19A"),
            std::u16string_view(u"vs"),
            std::u16string_view(u"VS button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F201"),
            std::u16string_view(u"japanese_here_"),
            std::u16string_view(u"Japanese “here” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F202\U0000FE0F"),
            std::u16string_view(u"japanese_service_charge_"),
            std::u16string_view(u"Japanese “service charge” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F237\U0000FE0F"),
            std::u16string_view(u"japanese_monthly_amount_"),
            std::u16string_view(u"Japanese “monthly amount” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F236"),
            std::u16string_view(u"japanese_not_free_of_charge_"),
            std::u16string_view(u"Japanese “not free of charge” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F22F"),
            std::u16string_view(u"japanese_reserved_"),
            std::u16string_view(u"Japanese “reserved” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F250"),
            std::u16string_view(u"japanese_bargain_"),
            std::u16string_view(u"Japanese “bargain” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F239"),
            std::u16string_view(u"japanese_discount_"),
            std::u16string_view(u"Japanese “discount” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F21A"),
            std::u16string_view(u"japanese_free_of_charge_"),
            std::u16string_view(u"Japanese “free of charge” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F232"),
            std::u16string_view(u"japanese_prohibited_"),
            std::u16string_view(u"Japanese “prohibited” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F251"),
            std::u16string_view(u"japanese_acceptable_"),
            std::u16string_view(u"Japanese “acceptable” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F238"),
            std::u16string_view(u"japanese_application_"),
            std::u16string_view(u"Japanese “application” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F234"),
            std::u16string_view(u"japanese_passing_grade_"),
            std::u16string_view(u"Japanese “passing grade” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F233"),
            std::u16string_view(u"japanese_vacancy_"),
            std::u16string_view(u"Japanese “vacancy” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00003297\U0000FE0F"),
            std::u16string_view(u"japanese_congratulations_"),
            std::u16string_view(u"Japanese “congratulations” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00003299\U0000FE0F"),
            std::u16string_view(u"japanese_secret_"),
            std::u16string_view(u"Japanese “secret” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F23A"),
            std::u16string_view(u"japanese_open_for_business_"),
            std::u16string_view(u"Japanese “open for business” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F235"),
            std::u16string_view(u"japanese_no_vacancy_"),
            std::u16string_view(u"Japanese “no vacancy” button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F534"),
            std::u16string_view(u"red_circle"),
            std::u16string_view(u"red circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E0"),
            std::u16string_view(u"orange_circle"),
            std::u16string_view(u"orange circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E1"),
            std::u16string_view(u"yellow_circle"),
            std::u16string_view(u"yellow circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E2"),
            std::u16string_view(u"green_circle"),
            std::u16string_view(u"green circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F535"),
            std::u16string_view(u"blue_circle"),
            std::u16string_view(u"blue circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E3"),
            std::u16string_view(u"purple_circle"),
            std::u16string_view(u"purple circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E4"),
            std::u16string_view(u"brown_circle"),
            std::u16string_view(u"brown circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026AB"),
            std::u16string_view(u"black_circle"),
            std::u16string_view(u"black circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000026AA"),
            std::u16string_view(u"white_circle"),
            std::u16string_view(u"white circle"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E5"),
            std::u16string_view(u"red_square"),
            std::u16string_view(u"red square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E7"),
            std::u16string_view(u"orange_square"),
            std::u16string_view(u"orange square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E8"),
            std::u16string_view(u"yellow_square"),
            std::u16string_view(u"yellow square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E9"),
            std::u16string_view(u"green_square"),
            std::u16string_view(u"green square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7E6"),
            std::u16string_view(u"blue_square"),
            std::u16string_view(u"blue square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7EA"),
            std::u16string_view(u"purple_square"),
            std::u16string_view(u"purple square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F7EB"),
            std::u16string_view(u"brown_square"),
            std::u16string_view(u"brown square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B1B"),
            std::u16string_view(u"black_large_square"),
            std::u16string_view(u"black large square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U00002B1C"),
            std::u16string_view(u"white_large_square"),
            std::u16string_view(u"white large square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025FC\U0000FE0F"),
            std::u16string_view(u"black_medium_square"),
            std::u16string_view(u"black medium square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025FB\U0000FE0F"),
            std::u16string_view(u"white_medium_square"),
            std::u16string_view(u"white medium square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025FE"),
            std::u16string_view(u"black_medium_small_square"),
            std::u16string_view(u"black medium-small square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025FD"),
            std::u16string_view(u"white_medium_small_square"),
            std::u16string_view(u"white medium-small square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025AA\U0000FE0F"),
            std::u16string_view(u"black_small_square"),
            std::u16string_view(u"black small square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U000025AB\U0000FE0F"),
            std::u16string_view(u"white_small_square"),
            std::u16string_view(u"white small square"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F536"),
            std::u16string_view(u"large_orange_diamond"),
            std::u16string_view(u"large orange diamond"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F537"),
            std::u16string_view(u"large_blue_diamond"),
            std::u16string_view(u"large blue diamond"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F538"),
            std::u16string_view(u"small_orange_diamond"),
            std::u16string_view(u"small orange diamond"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F539"),
            std::u16string_view(u"small_blue_diamond"),
            std::u16string_view(u"small blue diamond"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F53A"),
            std::u16string_view(u"red_triangle_pointed_up"),
            std::u16string_view(u"red triangle pointed up"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F53B"),
            std::u16string_view(u"red_triangle_pointed_down"),
            std::u16string_view(u"red triangle pointed down"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F4A0"),
            std::u16string_view(u"diamond_with_a_dot"),
            std::u16string_view(u"diamond with a dot"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F518"),
            std::u16string_view(u"radio"),
            std::u16string_view(u"radio button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F533"),
            std::u16string_view(u"white_square"),
            std::u16string_view(u"white square button"),
            emoji::Emoji::Category::Symbols},
      Emoji{std::u16string_view(u"\U0001F532"),
            std::u16string_view(u"black_square"),
            std::u16string_view(u"black square button"),
            emoji::Emoji::Category::Symbols},
      // Flags
      Emoji{std::u16string_view(u"\U0001F3C1"),
            std::u16string_view(u"chequered_flag"),
            std::u16string_view(u"chequered flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F6A9"),
            std::u16string_view(u"triangular_flag"),
            std::u16string_view(u"triangular flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F38C"),
            std::u16string_view(u"crossed_flags"),
            std::u16string_view(u"crossed flags"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F3F4"),
            std::u16string_view(u"black_flag"),
            std::u16string_view(u"black flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F3F3\U0000FE0F"),
            std::u16string_view(u"white_flag"),
            std::u16string_view(u"white flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F3F3\U0000FE0F\U0000200D\U0001F308"),
            std::u16string_view(u"rainbow_flag"),
            std::u16string_view(u"rainbow flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F3F3\U0000FE0F\U0000200D\U000026A7\U0000FE0F"),
            std::u16string_view(u"transgender_flag"),
            std::u16string_view(u"transgender flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F3F4\U0000200D\U00002620\U0000FE0F"),
            std::u16string_view(u"pirate_flag"),
            std::u16string_view(u"pirate flag"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1E8"),
            std::u16string_view(u"ascension_island_flag"),
            std::u16string_view(u"flag: Ascension Island"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1E9"),
            std::u16string_view(u"andorra_flag"),
            std::u16string_view(u"flag: Andorra"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1EA"),
            std::u16string_view(u"united_arab_emirates_flag"),
            std::u16string_view(u"flag: United Arab Emirates"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1EB"),
            std::u16string_view(u"afghanistan_flag"),
            std::u16string_view(u"flag: Afghanistan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1EC"),
            std::u16string_view(u"antigua_and_barbuda_flag"),
            std::u16string_view(u"flag: Antigua & Barbuda"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1EE"),
            std::u16string_view(u"anguilla_flag"),
            std::u16string_view(u"flag: Anguilla"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F1"),
            std::u16string_view(u"albania_flag"),
            std::u16string_view(u"flag: Albania"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F2"),
            std::u16string_view(u"armenia_flag"),
            std::u16string_view(u"flag: Armenia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F4"),
            std::u16string_view(u"angola_flag"),
            std::u16string_view(u"flag: Angola"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F6"),
            std::u16string_view(u"antarctica_flag"),
            std::u16string_view(u"flag: Antarctica"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F7"),
            std::u16string_view(u"argentina_flag"),
            std::u16string_view(u"flag: Argentina"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F8"),
            std::u16string_view(u"american_samoa_flag"),
            std::u16string_view(u"flag: American Samoa"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1F9"),
            std::u16string_view(u"austria_flag"),
            std::u16string_view(u"flag: Austria"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1FA"),
            std::u16string_view(u"australia_flag"),
            std::u16string_view(u"flag: Australia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1FC"),
            std::u16string_view(u"aruba_flag"),
            std::u16string_view(u"flag: Aruba"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1FD"),
            std::u16string_view(u"aland_islands_flag"),
            std::u16string_view(u"flag: Åland Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E6\U0001F1FF"),
            std::u16string_view(u"azerbaijan_flag"),
            std::u16string_view(u"flag: Azerbaijan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1E6"),
            std::u16string_view(u"bosnia_and_herzegovina_flag"),
            std::u16string_view(u"flag: Bosnia & Herzegovina"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1E7"),
            std::u16string_view(u"barbados_flag"),
            std::u16string_view(u"flag: Barbados"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1E9"),
            std::u16string_view(u"bangladesh_flag"),
            std::u16string_view(u"flag: Bangladesh"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1EA"),
            std::u16string_view(u"belgium_flag"),
            std::u16string_view(u"flag: Belgium"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1EB"),
            std::u16string_view(u"burkina_faso_flag"),
            std::u16string_view(u"flag: Burkina Faso"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1EC"),
            std::u16string_view(u"bulgaria_flag"),
            std::u16string_view(u"flag: Bulgaria"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1ED"),
            std::u16string_view(u"bahrain_flag"),
            std::u16string_view(u"flag: Bahrain"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1EE"),
            std::u16string_view(u"burundi_flag"),
            std::u16string_view(u"flag: Burundi"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1EF"),
            std::u16string_view(u"benin_flag"),
            std::u16string_view(u"flag: Benin"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F1"),
            std::u16string_view(u"st_barthelemy_flag"),
            std::u16string_view(u"flag: St. Barthélemy"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F2"),
            std::u16string_view(u"bermuda_flag"),
            std::u16string_view(u"flag: Bermuda"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F3"),
            std::u16string_view(u"brunei_flag"),
            std::u16string_view(u"flag: Brunei"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F4"),
            std::u16string_view(u"bolivia_flag"),
            std::u16string_view(u"flag: Bolivia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F6"),
            std::u16string_view(u"caribbean_netherlands_flag"),
            std::u16string_view(u"flag: Caribbean Netherlands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F7"),
            std::u16string_view(u"brazil_flag"),
            std::u16string_view(u"flag: Brazil"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F8"),
            std::u16string_view(u"bahamas_flag"),
            std::u16string_view(u"flag: Bahamas"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1F9"),
            std::u16string_view(u"bhutan_flag"),
            std::u16string_view(u"flag: Bhutan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1FB"),
            std::u16string_view(u"bouvet_island_flag"),
            std::u16string_view(u"flag: Bouvet Island"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1FC"),
            std::u16string_view(u"botswana_flag"),
            std::u16string_view(u"flag: Botswana"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1FE"),
            std::u16string_view(u"belarus_flag"),
            std::u16string_view(u"flag: Belarus"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E7\U0001F1FF"),
            std::u16string_view(u"belize_flag"),
            std::u16string_view(u"flag: Belize"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1E6"),
            std::u16string_view(u"canada_flag"),
            std::u16string_view(u"flag: Canada"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1E8"),
            std::u16string_view(u"cocos_keeling_islands_flag"),
            std::u16string_view(u"flag: Cocos (Keeling) Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1E9"),
            std::u16string_view(u"congo_kinshasa_flag"),
            std::u16string_view(u"flag: Congo - Kinshasa"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1EB"),
            std::u16string_view(u"central_african_republic_flag"),
            std::u16string_view(u"flag: Central African Republic"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1EC"),
            std::u16string_view(u"congo_brazzaville_flag"),
            std::u16string_view(u"flag: Congo - Brazzaville"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1ED"),
            std::u16string_view(u"switzerland_flag"),
            std::u16string_view(u"flag: Switzerland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1EE"),
            std::u16string_view(u"cote_d_ivoire_flag"),
            std::u16string_view(u"flag: Côte d’Ivoire"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F0"),
            std::u16string_view(u"cook_islands_flag"),
            std::u16string_view(u"flag: Cook Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F1"),
            std::u16string_view(u"chile_flag"),
            std::u16string_view(u"flag: Chile"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F2"),
            std::u16string_view(u"cameroon_flag"),
            std::u16string_view(u"flag: Cameroon"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F3"),
            std::u16string_view(u"china_flag"),
            std::u16string_view(u"flag: China"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F4"),
            std::u16string_view(u"colombia_flag"),
            std::u16string_view(u"flag: Colombia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F5"),
            std::u16string_view(u"clipperton_island_flag"),
            std::u16string_view(u"flag: Clipperton Island"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1F7"),
            std::u16string_view(u"costa_rica_flag"),
            std::u16string_view(u"flag: Costa Rica"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FA"),
            std::u16string_view(u"cuba_flag"),
            std::u16string_view(u"flag: Cuba"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FB"),
            std::u16string_view(u"cape_verde_flag"),
            std::u16string_view(u"flag: Cape Verde"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FC"),
            std::u16string_view(u"curacao_flag"),
            std::u16string_view(u"flag: Curaçao"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FD"),
            std::u16string_view(u"christmas_island_flag"),
            std::u16string_view(u"flag: Christmas Island"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FE"),
            std::u16string_view(u"cyprus_flag"),
            std::u16string_view(u"flag: Cyprus"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E8\U0001F1FF"),
            std::u16string_view(u"czechia_flag"),
            std::u16string_view(u"flag: Czechia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1EA"),
            std::u16string_view(u"germany_flag"),
            std::u16string_view(u"flag: Germany"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1EC"),
            std::u16string_view(u"diego_garcia_flag"),
            std::u16string_view(u"flag: Diego Garcia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1EF"),
            std::u16string_view(u"djibouti_flag"),
            std::u16string_view(u"flag: Djibouti"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1F0"),
            std::u16string_view(u"denmark_flag"),
            std::u16string_view(u"flag: Denmark"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1F2"),
            std::u16string_view(u"dominica_flag"),
            std::u16string_view(u"flag: Dominica"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1F4"),
            std::u16string_view(u"dominican_republic_flag"),
            std::u16string_view(u"flag: Dominican Republic"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1E9\U0001F1FF"),
            std::u16string_view(u"algeria_flag"),
            std::u16string_view(u"flag: Algeria"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1E6"),
            std::u16string_view(u"ceuta_and_melilla_flag"),
            std::u16string_view(u"flag: Ceuta & Melilla"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1E8"),
            std::u16string_view(u"ecuador_flag"),
            std::u16string_view(u"flag: Ecuador"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1EA"),
            std::u16string_view(u"estonia_flag"),
            std::u16string_view(u"flag: Estonia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1EC"),
            std::u16string_view(u"egypt_flag"),
            std::u16string_view(u"flag: Egypt"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1ED"),
            std::u16string_view(u"western_sahara_flag"),
            std::u16string_view(u"flag: Western Sahara"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1F7"),
            std::u16string_view(u"eritrea_flag"),
            std::u16string_view(u"flag: Eritrea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1F8"),
            std::u16string_view(u"spain_flag"),
            std::u16string_view(u"flag: Spain"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1F9"),
            std::u16string_view(u"ethiopia_flag"),
            std::u16string_view(u"flag: Ethiopia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EA\U0001F1FA"),
            std::u16string_view(u"european_union_flag"),
            std::u16string_view(u"flag: European Union"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1EE"),
            std::u16string_view(u"finland_flag"),
            std::u16string_view(u"flag: Finland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1EF"),
            std::u16string_view(u"fiji_flag"),
            std::u16string_view(u"flag: Fiji"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1F0"),
            std::u16string_view(u"falkland_islands_flag"),
            std::u16string_view(u"flag: Falkland Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1F2"),
            std::u16string_view(u"micronesia_flag"),
            std::u16string_view(u"flag: Micronesia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1F4"),
            std::u16string_view(u"faroe_islands_flag"),
            std::u16string_view(u"flag: Faroe Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EB\U0001F1F7"),
            std::u16string_view(u"france_flag"),
            std::u16string_view(u"flag: France"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1E6"),
            std::u16string_view(u"gabon_flag"),
            std::u16string_view(u"flag: Gabon"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1E7"),
            std::u16string_view(u"united_kingdom_flag"),
            std::u16string_view(u"flag: United Kingdom"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1E9"),
            std::u16string_view(u"grenada_flag"),
            std::u16string_view(u"flag: Grenada"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1EA"),
            std::u16string_view(u"georgia_flag"),
            std::u16string_view(u"flag: Georgia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1EB"),
            std::u16string_view(u"french_guiana_flag"),
            std::u16string_view(u"flag: French Guiana"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1EC"),
            std::u16string_view(u"guernsey_flag"),
            std::u16string_view(u"flag: Guernsey"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1ED"),
            std::u16string_view(u"ghana_flag"),
            std::u16string_view(u"flag: Ghana"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1EE"),
            std::u16string_view(u"gibraltar_flag"),
            std::u16string_view(u"flag: Gibraltar"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F1"),
            std::u16string_view(u"greenland_flag"),
            std::u16string_view(u"flag: Greenland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F2"),
            std::u16string_view(u"gambia_flag"),
            std::u16string_view(u"flag: Gambia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F3"),
            std::u16string_view(u"guinea_flag"),
            std::u16string_view(u"flag: Guinea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F5"),
            std::u16string_view(u"guadeloupe_flag"),
            std::u16string_view(u"flag: Guadeloupe"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F6"),
            std::u16string_view(u"equatorial_guinea_flag"),
            std::u16string_view(u"flag: Equatorial Guinea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F7"),
            std::u16string_view(u"greece_flag"),
            std::u16string_view(u"flag: Greece"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F8"),
            std::u16string_view(u"south_georgia_and_south_sandwich_islands_flag"),
            std::u16string_view(u"flag: South Georgia & South Sandwich Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1F9"),
            std::u16string_view(u"guatemala_flag"),
            std::u16string_view(u"flag: Guatemala"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1FA"),
            std::u16string_view(u"guam_flag"),
            std::u16string_view(u"flag: Guam"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1FC"),
            std::u16string_view(u"guinea_bissau_flag"),
            std::u16string_view(u"flag: Guinea-Bissau"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EC\U0001F1FE"),
            std::u16string_view(u"guyana_flag"),
            std::u16string_view(u"flag: Guyana"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1F0"),
            std::u16string_view(u"hong_kong_sar_china_flag"),
            std::u16string_view(u"flag: Hong Kong SAR China"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1F2"),
            std::u16string_view(u"heard_and_mcdonald_islands_flag"),
            std::u16string_view(u"flag: Heard & McDonald Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1F3"),
            std::u16string_view(u"honduras_flag"),
            std::u16string_view(u"flag: Honduras"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1F7"),
            std::u16string_view(u"croatia_flag"),
            std::u16string_view(u"flag: Croatia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1F9"),
            std::u16string_view(u"haiti_flag"),
            std::u16string_view(u"flag: Haiti"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1ED\U0001F1FA"),
            std::u16string_view(u"hungary_flag"),
            std::u16string_view(u"flag: Hungary"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1E8"),
            std::u16string_view(u"canary_islands_flag"),
            std::u16string_view(u"flag: Canary Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1E9"),
            std::u16string_view(u"indonesia_flag"),
            std::u16string_view(u"flag: Indonesia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1EA"),
            std::u16string_view(u"ireland_flag"),
            std::u16string_view(u"flag: Ireland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F1"),
            std::u16string_view(u"israel_flag"),
            std::u16string_view(u"flag: Israel"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F2"),
            std::u16string_view(u"isle_of_man_flag"),
            std::u16string_view(u"flag: Isle of Man"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F3"),
            std::u16string_view(u"india_flag"),
            std::u16string_view(u"flag: India"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F4"),
            std::u16string_view(u"british_indian_ocean_territory_flag"),
            std::u16string_view(u"flag: British Indian Ocean Territory"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F6"),
            std::u16string_view(u"iraq_flag"),
            std::u16string_view(u"flag: Iraq"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F7"),
            std::u16string_view(u"iran_flag"),
            std::u16string_view(u"flag: Iran"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F8"),
            std::u16string_view(u"iceland_flag"),
            std::u16string_view(u"flag: Iceland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EE\U0001F1F9"),
            std::u16string_view(u"italy_flag"),
            std::u16string_view(u"flag: Italy"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EF\U0001F1EA"),
            std::u16string_view(u"jersey_flag"),
            std::u16string_view(u"flag: Jersey"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EF\U0001F1F2"),
            std::u16string_view(u"jamaica_flag"),
            std::u16string_view(u"flag: Jamaica"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EF\U0001F1F4"),
            std::u16string_view(u"jordan_flag"),
            std::u16string_view(u"flag: Jordan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1EF\U0001F1F5"),
            std::u16string_view(u"japan_flag"),
            std::u16string_view(u"flag: Japan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1EA"),
            std::u16string_view(u"kenya_flag"),
            std::u16string_view(u"flag: Kenya"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1EC"),
            std::u16string_view(u"kyrgyzstan_flag"),
            std::u16string_view(u"flag: Kyrgyzstan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1ED"),
            std::u16string_view(u"cambodia_flag"),
            std::u16string_view(u"flag: Cambodia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1EE"),
            std::u16string_view(u"kiribati_flag"),
            std::u16string_view(u"flag: Kiribati"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1F2"),
            std::u16string_view(u"comoros_flag"),
            std::u16string_view(u"flag: Comoros"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1F3"),
            std::u16string_view(u"st_kitts_and_nevis_flag"),
            std::u16string_view(u"flag: St. Kitts & Nevis"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1F5"),
            std::u16string_view(u"north_korea_flag"),
            std::u16string_view(u"flag: North Korea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1F7"),
            std::u16string_view(u"south_korea_flag"),
            std::u16string_view(u"flag: South Korea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1FC"),
            std::u16string_view(u"kuwait_flag"),
            std::u16string_view(u"flag: Kuwait"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1FE"),
            std::u16string_view(u"cayman_islands_flag"),
            std::u16string_view(u"flag: Cayman Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F0\U0001F1FF"),
            std::u16string_view(u"kazakhstan_flag"),
            std::u16string_view(u"flag: Kazakhstan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1E6"),
            std::u16string_view(u"laos_flag"),
            std::u16string_view(u"flag: Laos"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1E7"),
            std::u16string_view(u"lebanon_flag"),
            std::u16string_view(u"flag: Lebanon"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1E8"),
            std::u16string_view(u"st_lucia_flag"),
            std::u16string_view(u"flag: St. Lucia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1EE"),
            std::u16string_view(u"liechtenstein_flag"),
            std::u16string_view(u"flag: Liechtenstein"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1F0"),
            std::u16string_view(u"sri_lanka_flag"),
            std::u16string_view(u"flag: Sri Lanka"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1F7"),
            std::u16string_view(u"liberia_flag"),
            std::u16string_view(u"flag: Liberia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1F8"),
            std::u16string_view(u"lesotho_flag"),
            std::u16string_view(u"flag: Lesotho"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1F9"),
            std::u16string_view(u"lithuania_flag"),
            std::u16string_view(u"flag: Lithuania"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1FA"),
            std::u16string_view(u"luxembourg_flag"),
            std::u16string_view(u"flag: Luxembourg"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1FB"),
            std::u16string_view(u"latvia_flag"),
            std::u16string_view(u"flag: Latvia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F1\U0001F1FE"),
            std::u16string_view(u"libya_flag"),
            std::u16string_view(u"flag: Libya"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1E6"),
            std::u16string_view(u"morocco_flag"),
            std::u16string_view(u"flag: Morocco"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1E8"),
            std::u16string_view(u"monaco_flag"),
            std::u16string_view(u"flag: Monaco"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1E9"),
            std::u16string_view(u"moldova_flag"),
            std::u16string_view(u"flag: Moldova"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1EA"),
            std::u16string_view(u"montenegro_flag"),
            std::u16string_view(u"flag: Montenegro"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1EB"),
            std::u16string_view(u"st_martin_flag"),
            std::u16string_view(u"flag: St. Martin"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1EC"),
            std::u16string_view(u"madagascar_flag"),
            std::u16string_view(u"flag: Madagascar"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1ED"),
            std::u16string_view(u"marshall_islands_flag"),
            std::u16string_view(u"flag: Marshall Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F0"),
            std::u16string_view(u"north_macedonia_flag"),
            std::u16string_view(u"flag: North Macedonia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F1"),
            std::u16string_view(u"mali_flag"),
            std::u16string_view(u"flag: Mali"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F2"),
            std::u16string_view(u"myanmar_burma_flag"),
            std::u16string_view(u"flag: Myanmar (Burma)"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F3"),
            std::u16string_view(u"mongolia_flag"),
            std::u16string_view(u"flag: Mongolia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F4"),
            std::u16string_view(u"macao_sar_china_flag"),
            std::u16string_view(u"flag: Macao SAR China"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F5"),
            std::u16string_view(u"northern_mariana_islands_flag"),
            std::u16string_view(u"flag: Northern Mariana Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F6"),
            std::u16string_view(u"martinique_flag"),
            std::u16string_view(u"flag: Martinique"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F7"),
            std::u16string_view(u"mauritania_flag"),
            std::u16string_view(u"flag: Mauritania"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F8"),
            std::u16string_view(u"montserrat_flag"),
            std::u16string_view(u"flag: Montserrat"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1F9"),
            std::u16string_view(u"malta_flag"),
            std::u16string_view(u"flag: Malta"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FA"),
            std::u16string_view(u"mauritius_flag"),
            std::u16string_view(u"flag: Mauritius"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FB"),
            std::u16string_view(u"maldives_flag"),
            std::u16string_view(u"flag: Maldives"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FC"),
            std::u16string_view(u"malawi_flag"),
            std::u16string_view(u"flag: Malawi"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FD"),
            std::u16string_view(u"mexico_flag"),
            std::u16string_view(u"flag: Mexico"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FE"),
            std::u16string_view(u"malaysia_flag"),
            std::u16string_view(u"flag: Malaysia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F2\U0001F1FF"),
            std::u16string_view(u"mozambique_flag"),
            std::u16string_view(u"flag: Mozambique"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1E6"),
            std::u16string_view(u"namibia_flag"),
            std::u16string_view(u"flag: Namibia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1E8"),
            std::u16string_view(u"new_caledonia_flag"),
            std::u16string_view(u"flag: New Caledonia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1EA"),
            std::u16string_view(u"niger_flag"),
            std::u16string_view(u"flag: Niger"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1EB"),
            std::u16string_view(u"norfolk_island_flag"),
            std::u16string_view(u"flag: Norfolk Island"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1EC"),
            std::u16string_view(u"nigeria_flag"),
            std::u16string_view(u"flag: Nigeria"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1EE"),
            std::u16string_view(u"nicaragua_flag"),
            std::u16string_view(u"flag: Nicaragua"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1F1"),
            std::u16string_view(u"netherlands_flag"),
            std::u16string_view(u"flag: Netherlands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1F4"),
            std::u16string_view(u"norway_flag"),
            std::u16string_view(u"flag: Norway"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1F5"),
            std::u16string_view(u"nepal_flag"),
            std::u16string_view(u"flag: Nepal"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1F7"),
            std::u16string_view(u"nauru_flag"),
            std::u16string_view(u"flag: Nauru"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1FA"),
            std::u16string_view(u"niue_flag"),
            std::u16string_view(u"flag: Niue"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F3\U0001F1FF"),
            std::u16string_view(u"new_zealand_flag"),
            std::u16string_view(u"flag: New Zealand"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F4\U0001F1F2"),
            std::u16string_view(u"oman_flag"),
            std::u16string_view(u"flag: Oman"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1E6"),
            std::u16string_view(u"panama_flag"),
            std::u16string_view(u"flag: Panama"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1EA"),
            std::u16string_view(u"peru_flag"),
            std::u16string_view(u"flag: Peru"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1EB"),
            std::u16string_view(u"french_polynesia_flag"),
            std::u16string_view(u"flag: French Polynesia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1EC"),
            std::u16string_view(u"papua_new_guinea_flag"),
            std::u16string_view(u"flag: Papua New Guinea"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1ED"),
            std::u16string_view(u"philippines_flag"),
            std::u16string_view(u"flag: Philippines"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F0"),
            std::u16string_view(u"pakistan_flag"),
            std::u16string_view(u"flag: Pakistan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F1"),
            std::u16string_view(u"poland_flag"),
            std::u16string_view(u"flag: Poland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F2"),
            std::u16string_view(u"st_pierre_and_miquelon_flag"),
            std::u16string_view(u"flag: St. Pierre & Miquelon"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F3"),
            std::u16string_view(u"pitcairn_islands_flag"),
            std::u16string_view(u"flag: Pitcairn Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F7"),
            std::u16string_view(u"puerto_rico_flag"),
            std::u16string_view(u"flag: Puerto Rico"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F8"),
            std::u16string_view(u"palestinian_territories_flag"),
            std::u16string_view(u"flag: Palestinian Territories"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1F9"),
            std::u16string_view(u"portugal_flag"),
            std::u16string_view(u"flag: Portugal"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1FC"),
            std::u16string_view(u"palau_flag"),
            std::u16string_view(u"flag: Palau"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F5\U0001F1FE"),
            std::u16string_view(u"paraguay_flag"),
            std::u16string_view(u"flag: Paraguay"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F6\U0001F1E6"),
            std::u16string_view(u"qatar_flag"),
            std::u16string_view(u"flag: Qatar"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F7\U0001F1EA"),
            std::u16string_view(u"reunion_flag"),
            std::u16string_view(u"flag: Réunion"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F7\U0001F1F4"),
            std::u16string_view(u"romania_flag"),
            std::u16string_view(u"flag: Romania"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F7\U0001F1F8"),
            std::u16string_view(u"serbia_flag"),
            std::u16string_view(u"flag: Serbia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F7\U0001F1FA"),
            std::u16string_view(u"russia_flag"),
            std::u16string_view(u"flag: Russia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F7\U0001F1FC"),
            std::u16string_view(u"rwanda_flag"),
            std::u16string_view(u"flag: Rwanda"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1E6"),
            std::u16string_view(u"saudi_arabia_flag"),
            std::u16string_view(u"flag: Saudi Arabia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1E7"),
            std::u16string_view(u"solomon_islands_flag"),
            std::u16string_view(u"flag: Solomon Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1E8"),
            std::u16string_view(u"seychelles_flag"),
            std::u16string_view(u"flag: Seychelles"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1E9"),
            std::u16string_view(u"sudan_flag"),
            std::u16string_view(u"flag: Sudan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1EA"),
            std::u16string_view(u"sweden_flag"),
            std::u16string_view(u"flag: Sweden"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1EC"),
            std::u16string_view(u"singapore_flag"),
            std::u16string_view(u"flag: Singapore"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1ED"),
            std::u16string_view(u"st_helena_flag"),
            std::u16string_view(u"flag: St. Helena"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1EE"),
            std::u16string_view(u"slovenia_flag"),
            std::u16string_view(u"flag: Slovenia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1EF"),
            std::u16string_view(u"svalbard_and_jan_mayen_flag"),
            std::u16string_view(u"flag: Svalbard & Jan Mayen"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F0"),
            std::u16string_view(u"slovakia_flag"),
            std::u16string_view(u"flag: Slovakia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F1"),
            std::u16string_view(u"sierra_leone_flag"),
            std::u16string_view(u"flag: Sierra Leone"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F2"),
            std::u16string_view(u"san_marino_flag"),
            std::u16string_view(u"flag: San Marino"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F3"),
            std::u16string_view(u"senegal_flag"),
            std::u16string_view(u"flag: Senegal"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F4"),
            std::u16string_view(u"somalia_flag"),
            std::u16string_view(u"flag: Somalia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F7"),
            std::u16string_view(u"suriname_flag"),
            std::u16string_view(u"flag: Suriname"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F8"),
            std::u16string_view(u"south_sudan_flag"),
            std::u16string_view(u"flag: South Sudan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1F9"),
            std::u16string_view(u"sao_tome_and_principe_flag"),
            std::u16string_view(u"flag: São Tomé & Príncipe"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1FB"),
            std::u16string_view(u"el_salvador_flag"),
            std::u16string_view(u"flag: El Salvador"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1FD"),
            std::u16string_view(u"sint_maarten_flag"),
            std::u16string_view(u"flag: Sint Maarten"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1FE"),
            std::u16string_view(u"syria_flag"),
            std::u16string_view(u"flag: Syria"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F8\U0001F1FF"),
            std::u16string_view(u"eswatini_flag"),
            std::u16string_view(u"flag: Eswatini"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1E6"),
            std::u16string_view(u"tristan_da_cunha_flag"),
            std::u16string_view(u"flag: Tristan da Cunha"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1E8"),
            std::u16string_view(u"turks_and_caicos_islands_flag"),
            std::u16string_view(u"flag: Turks & Caicos Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1E9"),
            std::u16string_view(u"chad_flag"),
            std::u16string_view(u"flag: Chad"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1EB"),
            std::u16string_view(u"french_southern_territories_flag"),
            std::u16string_view(u"flag: French Southern Territories"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1EC"),
            std::u16string_view(u"togo_flag"),
            std::u16string_view(u"flag: Togo"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1ED"),
            std::u16string_view(u"thailand_flag"),
            std::u16string_view(u"flag: Thailand"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1EF"),
            std::u16string_view(u"tajikistan_flag"),
            std::u16string_view(u"flag: Tajikistan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F0"),
            std::u16string_view(u"tokelau_flag"),
            std::u16string_view(u"flag: Tokelau"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F1"),
            std::u16string_view(u"timor_leste_flag"),
            std::u16string_view(u"flag: Timor-Leste"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F2"),
            std::u16string_view(u"turkmenistan_flag"),
            std::u16string_view(u"flag: Turkmenistan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F3"),
            std::u16string_view(u"tunisia_flag"),
            std::u16string_view(u"flag: Tunisia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F4"),
            std::u16string_view(u"tonga_flag"),
            std::u16string_view(u"flag: Tonga"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F7"),
            std::u16string_view(u"turkiye_flag"),
            std::u16string_view(u"flag: Türkiye"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1F9"),
            std::u16string_view(u"trinidad_and_tobago_flag"),
            std::u16string_view(u"flag: Trinidad & Tobago"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1FB"),
            std::u16string_view(u"tuvalu_flag"),
            std::u16string_view(u"flag: Tuvalu"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1FC"),
            std::u16string_view(u"taiwan_flag"),
            std::u16string_view(u"flag: Taiwan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1F9\U0001F1FF"),
            std::u16string_view(u"tanzania_flag"),
            std::u16string_view(u"flag: Tanzania"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1E6"),
            std::u16string_view(u"ukraine_flag"),
            std::u16string_view(u"flag: Ukraine"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1EC"),
            std::u16string_view(u"uganda_flag"),
            std::u16string_view(u"flag: Uganda"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1F2"),
            std::u16string_view(u"us_outlying_islands_flag"),
            std::u16string_view(u"flag: U.S. Outlying Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1F3"),
            std::u16string_view(u"united_nations_flag"),
            std::u16string_view(u"flag: United Nations"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1F8"),
            std::u16string_view(u"united_states_flag"),
            std::u16string_view(u"flag: United States"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1FE"),
            std::u16string_view(u"uruguay_flag"),
            std::u16string_view(u"flag: Uruguay"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FA\U0001F1FF"),
            std::u16string_view(u"uzbekistan_flag"),
            std::u16string_view(u"flag: Uzbekistan"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1E6"),
            std::u16string_view(u"vatican_city_flag"),
            std::u16string_view(u"flag: Vatican City"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1E8"),
            std::u16string_view(u"st_vincent_and_grenadines_flag"),
            std::u16string_view(u"flag: St. Vincent & Grenadines"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1EA"),
            std::u16string_view(u"venezuela_flag"),
            std::u16string_view(u"flag: Venezuela"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1EC"),
            std::u16string_view(u"british_virgin_islands_flag"),
            std::u16string_view(u"flag: British Virgin Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1EE"),
            std::u16string_view(u"us_virgin_islands_flag"),
            std::u16string_view(u"flag: U.S. Virgin Islands"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1F3"),
            std::u16string_view(u"vietnam_flag"),
            std::u16string_view(u"flag: Vietnam"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FB\U0001F1FA"),
            std::u16string_view(u"vanuatu_flag"),
            std::u16string_view(u"flag: Vanuatu"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FC\U0001F1EB"),
            std::u16string_view(u"wallis_and_futuna_flag"),
            std::u16string_view(u"flag: Wallis & Futuna"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FC\U0001F1F8"),
            std::u16string_view(u"samoa_flag"),
            std::u16string_view(u"flag: Samoa"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FD\U0001F1F0"),
            std::u16string_view(u"kosovo_flag"),
            std::u16string_view(u"flag: Kosovo"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FE\U0001F1EA"),
            std::u16string_view(u"yemen_flag"),
            std::u16string_view(u"flag: Yemen"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FE\U0001F1F9"),
            std::u16string_view(u"mayotte_flag"),
            std::u16string_view(u"flag: Mayotte"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FF\U0001F1E6"),
            std::u16string_view(u"south_africa_flag"),
            std::u16string_view(u"flag: South Africa"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FF\U0001F1F2"),
            std::u16string_view(u"zambia_flag"),
            std::u16string_view(u"flag: Zambia"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(u"\U0001F1FF\U0001F1FC"),
            std::u16string_view(u"zimbabwe_flag"),
            std::u16string_view(u"flag: Zimbabwe"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(
              u"\U0001F3F4\U000E0067\U000E0062\U000E0065\U000E006E\U000E0067\U000E007F"),
            std::u16string_view(u"england_flag"),
            std::u16string_view(u"flag: England"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(
              u"\U0001F3F4\U000E0067\U000E0062\U000E0073\U000E0063\U000E0074\U000E007F"),
            std::u16string_view(u"scotland_flag"),
            std::u16string_view(u"flag: Scotland"),
            emoji::Emoji::Category::Flags},
      Emoji{std::u16string_view(
              u"\U0001F3F4\U000E0067\U000E0062\U000E0077\U000E006C\U000E0073\U000E007F"),
            std::u16string_view(u"wales_flag"),
            std::u16string_view(u"flag: Wales"),
            emoji::Emoji::Category::Flags},
};
